# General auxiliary functions & classes
#
# Stilianos Louca
# Copyright 2024
#
# LICENSE AGREEMENT
# - - - - - - - - -
# All rights reserved.
# Use and redistributions of this code is permitted for commercial and non-commercial purposes,
# under the following conditions:
#
#	* Redistributions must retain the above copyright notice, this list of 
#	  conditions and the following disclaimer in the code itself, as well 
#	  as in documentation and/or other materials provided with the code.
#	* Neither the name of the original author (Stilianos Louca), nor the names 
#	  of its contributors may be used to endorse or promote products derived 
#	  from this code without specific prior written permission.
#	* Proper attribution must be given to the original author, including a 
#     reference to the peer-reviewed publication through which the code was published.
#
# THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# - - - - - - - - -

import string
import sys, os
import numpy
import math
import scipy
from joblib import Parallel, delayed
import matplotlib
import matplotlib.pyplot as pyplot
import mpl_toolkits.axes_grid1
import warnings
import time
import re
import fnmatch
import pickle
import hashlib
import pandas
from numpy import nan, inf, pi

FILE_EXTENSION_TO_COLUMN_DELIMITER 	= {"tsv":"\t", "tab":"\t", "csv":",", "dnl":"\t"}
DEFAULT_COLORMAP_FOR_CORRELATIONS  	= "RdBu"
GENERIC_BAR_COLOR_PALETTE			= numpy.array(["#7fcbfa", "#fc8b8b", "#a2f2ae", "#fcd58b", "#fa73e6", "#9ee6e2", "#f792f1", "#8dcc93", "#e6be91", "#fa8c8e", "#bf99e8", "#9cd3f7", "#837bdb", "#f77f43", "#9891e6", "#f5aa5f", "#c76363", "#c98dc9", "#98d5fa", "#ebb191", "#f5b576", "#89cccc", "#ebeb91", "#d4a6ed", "#e07979", "#aaa4ed", "#d4b3e8", "#8bc98f", "#dba7db"], dtype=object)



##################################
# MARK Misc

def float_or_nan(s):
    try:
        f = float(s)
        return f
    except ValueError:
        return nan

numpy_float_or_nan = numpy.vectorize(float_or_nan,otypes=[float])


def is_constant(A):
	if(numpy.isscalar(A) or (A.size==0)): return True
	elif(numpy.all(A==A.item(0))): return True
	else: return False


def is_list_or_tuple(obj):
	return (isinstance(obj,list) or isinstance(obj,tuple))


# print a 1D numeric vector in standard layout, as a string
def sprint_vector(A, separator="  "):
	return separator.join(['%g'%(a) for a in A])

# print the shape of a numpy array in the format N1 x N2 x .. Nn
def sprint_shape(array):
	return sprint_vector(array.shape, separator=" x ")


def float_or_nan(s):
    try:
        f = float(s)
        return f
    except ValueError:
        return nan
     
     
def pad_string_right(x, min_len, char=" "):
	return x + char*(max(0,min_len-len(x)))


# join a list/tuple of strings using a separator, omitting any empty ones
def join_nonempty(strings, sep):
	return sep.join(filter(None, strings))


# compute logarithms without warnings in the case of zero/negative values
def log_nansafe(X):
	with warnings.catch_warnings():
		warnings.filterwarnings(action='ignore', message='invalid value encountered in log')
		warnings.filterwarnings(action='ignore', message='divide by zero encountered in log')
		return numpy.log(X)


# divide two numpy arrays or scalars A/B, handling division by zero more gracefully:
#	x/0 yields +inf if x>0
#	x/0 yields -inf if x<0
#	0/0 yields nan
# this function can account for broadcasting between A & B. Scalar A & B are also supported.
def divide_nansafe(A, B):
	if(numpy.isscalar(B)):
		if(B==0): return (inf*numpy.sign(A) if numpy.isscalar(A) else inf*numpy.sign(A))
		else: return A/B
	else:
		with warnings.catch_warnings():
			warnings.filterwarnings(action='ignore', message='divide by zero encountered in divide')
			warnings.filterwarnings(action='ignore', message='invalid value encountered in divide')
			warnings.filterwarnings(action='ignore', message='invalid value encountered in multiply')
			ratio = A/B
			if(A.shape==B.shape):
				ratio[B==0] = inf*numpy.sign(A[B==0])
			else:
				Bzeros = numpy.tile(B==0, numpy.divide(B.shape,A.shape).astype(int))
				ratio[Bzeros] = inf*numpy.sign(A[Bzeros]) # account for broadcasting that was applied to A/B
	return ratio


def invert_permutation(perm, check_validity=False):
	if(check_validity):
		sorted_perm = numpy.sort(perm)
		if((sorted_perm[-1]!=len(perm)-1) or (sorted_perm[0]!=0)): raise ValueError("Not a valid permutation: %s"%(str(perm)))
		if(numpy.any(numpy.diff(sorted_perm)!=1)): raise ValueError("Not a valid permutation: %s"%(str(perm)))
	inv = numpy.empty_like(perm, dtype=int)
	inv[perm] = numpy.arange(len(perm))
	return inv
   
   

# return the set {0,..,N-1} \ subset
# subset may be a slice object or a 1D array/list of integers or a single integer
def integer_complement(N, subset):
	keep = numpy.ones(N, dtype=bool)
	keep[subset,...] = False # use ellipsis at the end to accommodate situations where subset is a tuple
	return numpy.nonzero(keep)[0]



# choose a good format for formatting values of a grid, with enough precision so that adjacent grid points can be distinguished, but not too much superfluous precision
def best_grid_number_format(grid):
	diff_order = int(math.floor(numpy.log10(numpy.min(numpy.abs(numpy.diff(grid))))))
	max_order  = int(math.ceil(numpy.log10(numpy.max(numpy.abs(grid)))))
	if(diff_order<=0): return "%."+str(1+abs(diff_order))+"f"
	else: return "%"+str(max_order-diff_order+1)+"g"



def get_date_time():
	return time.strftime("%Y.%m.%d %H:%M:%S")


def get_shell_command():
	arguments = sys.argv
	arguments = [("'"+a+"'" if (a=="" or re.search(r"[\s*#:;]", a)) else ("'"+'\\\\t'+"'" if (a=='\\t') else a)).replace("\n","") for a in arguments];
	return ' '.join(arguments)


# compute a hash value for an arbitrary object
# This function first pickles the object and then computes an SHA256 hash from the pickle
def hash_object(obj):
	return hashlib.sha256(pickle.dumps(obj)).hexdigest()


# amend an object in place by adding custom attributes
# For example:
#	model = amend_object(model, label="favorite model")
def amend_object(obj, **kwargs):
	for key,value in kwargs.items():
		obj.__dict__[key] = value
	return obj


# Convert a specification of Nthreads to an actual integer number of threads to use
# Nthreads is either interpreted as the number of threads to use (if strictly positive), or as the number of threads to use below the number of available cores (if non-positive). If negative or zero, use the number of available cores minus this value. If None, the total number of available cores is returned.
def parse_Nthreads(Nthreads):
	if(Nthreads is None):
		return os.cpu_count()
	elif(Nthreads>0): 
		return Nthreads
	else:
		return max(1, os.cpu_count()+Nthreads)


# execute a target function multiple times in parallel (if reasonable)
# For fast functions jobs are first grouped into batches to reduce the overhead of dispatching lots of parallel jobs
def run_parallel(function, 			# a callable (e.g., a function), to be executed multiple times. This function should take a single positional argument. If you want to run a function that takes multiple argument, use lambda as a wrapper.
				arguments, 			# a sequence (e.g., list) of length N, where N is the number of times the function should be executed. Each element of arguments[] is the single input argument for a single function call.
				Nthreads = None,	# integer, number of parallel threads to use. Interpreted as specified in parse_Nthreads(). If 1, a simple loop is used without parallelization; a loop i used on purpose to facilitate debugging.
				fast 	 = False,	# whether the function is fast, i.e., it takes a short time to execute a single call compared to the overhead associated with dispatching jobs to processes.
				prefer   = None,	# either None, "processes" or "threads"
				randomize= False):	# randomize the order in which tasks are executed, if parallelized. This may help with load balancing across threads.
	Nthreads = parse_Nthreads(Nthreads)
	N 		 = len(arguments)
	order 	 = list(range(N))
	if(randomize): random.shuffle(order)
	if((Nthreads>1) and (N>1)):
		if(fast):
			# group jobs into batches
			# for fast functions this can speed up computation by ~ 25%.
			batch_size = max(1,int(0.2*N/Nthreads))
			Nbatches   = math.ceil(N/batch_size)
			def aux_run_batch(b):
				return [function(arguments[order[a]]) for a in range(b*batch_size,min((b+1)*batch_size,N))]
			results = Parallel(n_jobs=min(N,Nthreads), prefer=prefer, verbose=0)(delayed(aux_run_batch)(b) for b in range(Nbatches))
			results = [r for rr in results for r in rr] # flatten results, i.e. unwrap batches
		else:
			# run standard parallelized
			results = Parallel(n_jobs=min(N,Nthreads), prefer=prefer, verbose=0)(delayed(function)(arguments[order[a]]) for a in range(N))
		if(randomize):
			inverse_order = invert_permutation(order)
			results = [results[k] for k in inverse_order]
		return results
	else:
		results = [None]*len(arguments)
		for a,argument in enumerate(arguments):
			results[a] = function(argument)
		return results
		
		
# given 2 sequences/1D arrays (needles[] & haystack[]), find the integer locations of needles in haystack.
# This is similar to the match() function in R
# Internally, this function first creates a dictionary lookup table for haystack, so it is most efficient when the number of needles is much larger than the haystack.
def find_values(needles,	# sequence (e.g., list or 1D array) of values to find
				haystack, 	# sequence of possible values to search in
				missing	= "error"): # either "error" or "skip" or "flag" (return -1 as location). If skip, the returned integer sequence may be shorter than the input needles.
	value2index = {value:v for v,value in enumerate(haystack)}
	if(missing=="error"):
		return [value2index[value] for value in needles]
	elif(missing=="skip"):
		return [value2index[value] for value in needles if (value in value2index)]
	elif(missing=="flag"):
		return [value2index.get(value,-1) for value in needles]

	
# given a sequence (e.g. list or tuple or 1D numpy array) or values, group the values based on identity
# The function returns a new list of length NG, where each element is a 1D integer array containing the indices in the original values[] of identical elements.
# For example, [3,4,8,3,4,15] will return [[0,3], [1,4], [2], [5]].
# Optionally, one can limit the size of each returned group.
def group_identicals(values, max_group_size=inf):
	values  = numpy.asarray(values)
	uniques = numpy.unique(values)
	value2unique = {value:u for u,value in enumerate(uniques)}
	groups  = [[] for g in range(len(uniques))]
	for v,value in enumerate(values):
		groups[value2unique[value]].append(v)
	if(numpy.isfinite(max_group_size)):
		groups = [group[:min(max_group_size,len(group))] for group in groups]
	return groups



# convert a sequence of integer indices into a corresponding boolean array, e.g. for purposes of array subsetting
def indices2booleans(N, indices):
	booleans = numpy.zeros(N, dtype=bool)
	booleans[indices] = True
	return booleans


# compute the CO2, HCO3^{-} & CO3^{2-} concentrations in aquatic solution based on pH and DIC.
# temperature & salinity may be provided to adjust the equilibrium constants to non-standard conditions.
# pH, DIC, temperature and salinity may be scalars or numpy arrays of arbitrary shape, but they must all be broadcastable to each other.
# The returned CO2, HCO3 & CO3 concentrations will be in the same units as the provided DIC.
def carbonate_species_concentrations(pH, DIC, temperature=25, salinity=0):
	# determine equilibrium constants of carbonic acid, based on temperature & salinity
	# Reference: Millero et al. (2002). Dissociation constants for carbonic acid determined from field measurements. Equations 5 & 6.
	temperature	+= 273.15
	pK1			= -8.712 - 9.460e-3 * salinity + 8.56e-5*(salinity**2) + 1355.1/temperature + 1.7976*numpy.log(temperature)
	pK2			= 17.0001 - 0.01259*salinity - 7.9334e-5*(salinity**2) + 936.291/temperature - 1.87354*numpy.log(temperature) - 2.61471 * salinity/temperature + 0.07479 * (salinity**2)/temperature
	K1, K2		= 10**(-pK1), 10**(-pK2)
	H   		= numpy.power(10,-pH)
	denominator = (H**2 + K1*H+K1*K2)
	CO2  		= (H**2)*DIC/denominator
	HCO3 		= K1*H*DIC/denominator
	CO3  		= K1*K2*DIC/denominator
	return IDict(CO2=CO2, HCO3=HCO3, CO3=CO3)


###############################################
# MARK Filtering


# return a list of indices corresponding to items to keep
# this function preserves the order in start_pool[], or of names[] if start_pool is None.
def filter_names(names,
				start_pool	 	= None,		# starting pool to further filter. This will typically be range(len(names)), but can also be a subset thereof in case of multiple filtering. Can also be None (equivalent to range(len(names)))
				only_names 		= None, 	# optional, either a string specifying a comma-separated list of names (or wildcards or regular expressions) to keep, or a list/set of such names/wildcards, or a list of integer indices. If None or has len==0, this filter is ignored.
				omit_names 		= None,		# optional, either a string specifying a comma-separated list of names (or wildcards or regular expressions) to omit, or a list/set of such names/wildcards, or a list of integer indices. If None or has len==0, this filter is ignored.
				case_sensitive 	= False,
				pattern_type	= "name",	# either "index" (i.e., integers) or "name" or "wildcard" or "regex", specifying the type of patterns listed in only_names & omit_names. By far the most efficient is "index", followed by "name".
				as_indices		= True):	# if True, the indices in names[] are returned. If False, the corresponding names are returned.
	N = len(names)
	if((start_pool is None) and ((only_names is None) or (len(only_names)==0)) and ((omit_names is None) or (len(omit_names)==0))):
		items_to_keep = numpy.arange(N, dtype=int)
	elif(pattern_type=="index"):
		# filter based on integer indices
		subkeep = numpy.empty(N, dtype=bool)
		keep = numpy.ones(N, dtype=bool)
		if(start_pool is not None):
			subkeep[:] = False
			subkeep[start_pool] = True
			keep = keep & subkeep		
		if((only_names is not None) and (len(only_names)>0)):
			subkeep[:] = False
			subkeep[only_names] = True
			keep = keep & subkeep
		if((omit_names is not None) and (len(omit_names)>0)):
			keep[omit_names] = False
		items_to_keep = numpy.where(keep)[0]
	else:
		# filter based on string matching	
		items_to_keep = (list(range(len(names))) if (start_pool is None) else start_pool)
		if(not case_sensitive): 
			local_names = [name.lower() for name in names]
		else:
			local_names = names
		if((only_names is not None) and (len(only_names)>0)):
			if(isinstance(only_names,str)): only_names = list(filter(len, only_names.split(',')))
			if(not case_sensitive): only_names = [name.lower() for name in only_names]
			if(pattern_type=="wildcard"):
				only_names = [re.compile(fnmatch.translate(w)) for w in only_names] # convert wildcards to compiled regex objects
				items_to_keep = [i for i in items_to_keep if any((w.match(local_names[i]) is not None) for w in only_names)]
				# alternative: equivalent, but slower
				#items_to_keep = [i for i in items_to_keep if any(fnmatch.fnmatchcase(local_names[i],w) for w in only_names)]
			elif(pattern_type=="regex"):
				only_names = [re.compile(w) for w in only_names] # compile regex objects
				items_to_keep = [i for i in items_to_keep if any((w.match(local_names[i]) is not None) for w in only_names)]
			elif(pattern_type=="name"):
				only_names = set(only_names)
				items_to_keep = [i for i in items_to_keep if (local_names[i] in only_names)]
		if((omit_names is not None) and (len(omit_names)>0)):
			if(isinstance(omit_names,str)): omit_names = list(filter(len, omit_names.split(',')))
			if(not case_sensitive): omit_names = [name.lower() for name in omit_names];
			if(pattern_type=="wildcard"):
				omit_names = [re.compile(fnmatch.translate(w)) for w in omit_names] # convert wildcards to compiled regex objects
				items_to_keep = [i for i in items_to_keep if all((w.match(local_names[i]) is None) for w in omit_names)]
				# alternative: equivalent, but slower
				#items_to_keep = [i for i in items_to_keep if (not any(fnmatch.fnmatchcase(local_names[i],w) for w in omit_names))]
			elif(pattern_type=="regex"):
				omit_names = [re.compile(w) for w in omit_names] # compile regex objects
				items_to_keep = [i for i in items_to_keep if all((w.match(local_names[i]) is None) for w in omit_names)]		
			elif(pattern_type=="name"):
				omit_names = set(omit_names)
				items_to_keep = [i for i in items_to_keep if (local_names[i] not in omit_names)]
	if(as_indices):
		return items_to_keep
	else:
		return [names[k] for k in items_to_keep]
	
	


# remove nan and inf values from one or more synchronized scalar or multidimensional series, by keeping only the records (time points) where all series have finite values.
# *value_arrays should be one or more 1D or multidimensional numpy arrays, of sizes NT x NV1 x ..., NT x NV2 x ... etc, but may also include None.
# Arrays need not have the same shapes except for their first axis, which should have length NT in all cases. The arrays are assumed to be synchronized along their first axis.
# For example, one array may be 1D (containing times) and another array may be 2D (containing multidimensional values at those times).
# For example, you may call this function as follows:
#	new_times, new_observables, new_drivers = filter_series_nans(times, observables, drivers)
def filter_series_nans(*value_arrays):
	valids = True
	for array in value_arrays:
		if(array is not None): valids = valids & numpy.all(numpy.isfinite(array), axis=tuple(range(1,array.ndim)))
	NT = len(valids)
	valids = numpy.where(valids)[0]
	if(len(valids)==NT):
		new_value_arrays = value_arrays
	else:
		new_value_arrays = [None]*len(value_arrays)
		for a,array in enumerate(value_arrays):
			new_value_arrays[a] = (None if (array is None) else array[valids,...])
	return (*new_value_arrays,)



##############################################################
# MARK Calculus


# compute sign(x)*|x|^exponent
def power_transform(value, exponent):
	if(numpy.isscalar(value)):
		return (-1 if (value<0) else 1)*(abs(value)**exponent)
	else:
		return numpy.sign(value)*numpy.power(numpy.abs(value),exponent)

# inverse function of power_transform()
def power_transform_inv(value, exponent):
	if(numpy.isscalar(value)):
		return (-1 if (value<0) else 1)*(abs(value)**(1/exponent))
	else:
		return numpy.sign(value)*numpy.power(numpy.abs(value),(1/exponent))


# linearly interpolate a multivariate time series
# The input arguments are similar to those in numpy.interp (but some default values are different).
# One difference is that this function supports multivariate time series (i.e., with fp having more than 1 column and in fact an arbitrary number of dimensions).
# Another difference is that the function can deal with nans (if check_nan is True)
# The returned array will have the same number of dimensions as fp, with all axes except the interpolated one having the same size as fp.
# Hence, if fp is of size NP x N1 ... x Nn and axis=0, then the returned array will have size NX x N1 x .. x Nn, where NX:=len(x).
def interpolate_time_series(x, 			# 1D numpy array, the x-coordinates at which to evaluate the interpolated values.
							xp,			# 1D numpy array of size NP, the x-coordinates of the data points, must be increasing
							fp,			# (n+1)-dimensional numpy array of size NP x N1 x N2 x .. x Nn, the y-coordinates of the data points. By default interpolation will occur along the first axis (but see argument axis).
							axis		= 0,	# integer, specifying the axis of fp along which axis to interpolate. 
							left		= nan,
							right		= nan,
							period		= None,
							degree		= 1,	# interpolation degree. Either 1 ("linear") or -1 ("const", use last available value, i.e. at or below each requested x)
							check_nan	= True,
							Nthreads	= None):
	NX = len(x)
	if(axis!=0): fp = numpy.moveaxis(fp, source=axis, destination=0) # temporarily move the interpolated axis to the beginning
	NV = int(numpy.prod(fp.shape[1:]))
	input_shape = fp.shape
	if(fp.ndim!=2):
		# flatten axes 1,2,.. for easier processing below
		fp = numpy.reshape(fp,(fp.shape[0],NV))
	def aux_interpolate_scalar(d):
		if(check_nan):
			valids = numpy.where(numpy.isfinite(xp) & numpy.isfinite(fp[:,d]))[0]
			if(len(valids)>=2):
				sub_xp = xp[valids]
				sub_fp = fp[valids,d]
			else:
				return numpy.full(NX,nan)
		else:
			sub_xp = xp
			sub_fp = fp[:,d]
		if(degree==1):
			return numpy.interp(x=x, xp=sub_xp, fp=sub_fp, left=left, right=right, period=period)
		elif(degree==-1):
			lasts = sub_xp.searchsorted(x, side="right")-1
			if(numpy.all(lasts)>=0):
				return sub_fp[lasts]
			else:
				f = numpy.full(NX, nan)
				f[lasts>=0] = sub_fp[lasts[lasts>=0]]
				return f
	interpolated = numpy.asarray(run_parallel(function=aux_interpolate_scalar, arguments=range(NV), Nthreads=Nthreads, fast=True, prefer="threads")).transpose()
	if(len(input_shape)!=2): 
		# revert interpolated array to proper shape
		interpolated = numpy.reshape(interpolated, (NX,)+input_shape[1:])
	if(axis!=0):
		interpolated = numpy.moveaxis(interpolated, source=0, destination=axis)
	return interpolated


# populate a 2D array of size N x M with M Gaussian distributions, one Gaussian per column
# this may be used e.g. to construct simple initial conditions for multivariate PDEs
# mean, std and mass may each be a scalar or a 1D array or sequence of size M
def Gaussians_on_grid(M, grid, mean, std, mass=1):
	N = len(grid)
	value = numpy.empty((N,M))
	for m in range(M):
		this_std  = (std if numpy.isscalar(std) else std[m])
		this_mean = (mean if numpy.isscalar(mean) else mean[m])
		this_mass = (mass if numpy.isscalar(mass) else mass[m])
		value[:,m] = (this_mass/numpy.sqrt(2*numpy.pi*this_std**2))*numpy.exp(-(grid-this_mean)**2/(2*this_std**2))
	return value



# Estimate the Jacobian of some multivariate-multidimensional function F:R^NX->R^NY at some point X0 via central finite differences.
# The step in each direction is determined iteratively.
# The estimated Jacobian will be returned as a 2D array of size NY x NX.
# If dX0 includes zeros, or if dX0 is None and some X0 are zero, then those X will not be varied and the derivatives w.r.t. those X will simply be set to nan.
def estimate_Jacobian_FD(F, 					# differentiable function that takes a 1D numeric array of size NX as argument and returns a 1D numeric array of size NY. If verbose_F==True, then F should also take arguments verbosity & verbose_prefix.
						 X0,					# 1D numeric array of size NX	
						 dX0	 		= None,	# optional, either a scalar float or a 1D numeric array of size NX, specifying the initial X steps for finite differences. If None, this is automatically chosen based on the input X0. Any X whose dX0 is zero will not be varied, i.e. the derivative w.r.t. to that X will not be computed.
						 Y0				= None,	# optional, the result of F(X0). If None, this is computed here. The caller may provide this for computational efficiency, for example if Y0 has already been computed as part of previous steps.
						 max_refinements= 10,	# integer, maximum number of iterative refinements of the finite differences computed per X component. This may be used to limit the computation done, at the expense of potentially prematurely stopping the refinements. If 0, no refinements are done, i.e., the derivatives are estimated based on the initial dX0.
						 rtol			= 0.001,# relative error tolerance in the estimation of the derivatives. Iterative refinements will stop once the derivative changes by less than this threshold, relative to the previous iteration.
						 normalize		= False,# if True, the derivatives are normalized by multiplying with |Y|/|X|, i.e., instead of returning dY_i/dX_j the function returns |X_j/Y_i|*dY_i/dX_j. These are sometimes called normalized local sensitivity coefficients.
						 Xnames			= None,	# optional string sequence of length NX, specifying names for the various X components. Used only for message printing. Ignored if verbosity==0.
						 Nthreads		= 0,	# number of parallel threads to use for computing derivatives of distinct X components.
						 verbose_F		= True,	# whether F accepts the two additional arguments verbosity & verbose_prefix
						 verbosity 		= 0,
						 verbose_prefix = ""):
	# basic input checking
	if((verbosity>0) and (Xnames is None)): Xnames = ["X#%d"%(k) for k in range(len(X0))]
	
	def aux_F(X,vstep):
		if(verbose_F): return F(X,verbosity=max(0,verbosity-vstep),verbose_prefix=verbose_prefix+("  "*vstep))
		else: return F(X) 
		
	# determine and evaluate Y0
	if(Y0 is None):
		if(verbosity>=1): print("%sEvaluating at X0.."%(verbose_prefix))
		Y0 = aux_F(X0, vstep=1)
	scalar 	= numpy.isscalar(Y0)
	NX, NY 	= len(X0), (1 if scalar else len(Y0))

	# determine the initial step for each X component
	if(dX0 is None): dX0 = numpy.abs(X0)/10
	elif(numpy.isscalar(dX0)): dX0 = numpy.full(dX0, NX, dtype=float)
	elif(dX0.shape!=(NX,)): raise ValueError("dX0 has invalid shape %s: Expected a scalar or a 1D array of length %d"%(sprint_shape(dX0),NX))
	
	# define auxiliary function for computing derivatives w.r.t. a specific X component, e.g. for parallelization purposes
	def aux_derive_component(k):
		if(dX0[k]==0): return numpy.full(NY,nan) # skip this X
		if(verbosity>=1): print("%sEstimating derivatives w.r.t. to %s.."%(verbose_prefix,Xnames[k]))
		step  = dX0[k]
		Xl,Xr = X0.copy(), X0.copy()
		Xl[k] = X0[k] - step
		Xr[k] = X0[k] + step
		if(verbosity>=2): print("%s  Examining %s values %g & %g.."%(verbose_prefix,Xnames[k],Xl[k],Xr[k]))
		Jk = (aux_F(Xr,vstep=2) - aux_F(Xl,vstep=2))/(2*step) # estimated derivative along the k-th axis via central finite difference
		if(max_refinements==0): return Jk
		# iteratively refine the step until convergence
		for refinement in range(max_refinements):
			step /= 2
			Xl[k] = X0[k] - step
			Xr[k] = X0[k] + step
			if(verbosity>=2): print("%s  Refining %s (round #%d out of max %d), examining values %g & %g.."%(verbose_prefix,Xnames[k],refinement+1,max_refinements,Xl[k],Xr[k]))
			Jklatest = (aux_F(Xr,vstep=2) - aux_F(Xl,vstep=2))/(2*step)
			if(numpy.all((numpy.abs(Jklatest-Jk)<=rtol*numpy.abs(Jk)))):
				# change from previous estimate is sufficiently small, so finish
				if(verbosity>=3): print("%s    Note: Finite differences converged for %s derivatives.."%(verbose_prefix,Xnames[k]))
				break
			Jk = Jklatest
		return Jklatest
	
	J 			= numpy.full((NY,NX),nan) # preallocate space
	focalX 		= numpy.where(dX0!=0)[0] # X components w.r.t. which to estimate derivatives
	J[:,focalX] = numpy.asarray(run_parallel(aux_derive_component, arguments=focalX, Nthreads=Nthreads), dtype=float).T
	if(normalize): J = numpy.expand_dims(1/numpy.abs(Y0), axis=1) * numpy.expand_dims(numpy.abs(X0), axis=0) * J
	return J


# Integrate a scalar-valued or array-valued function over a 1D grid
# The grid may be irregular, and the integrated function is assumed to either vary linearly between grid points or be piecewise constant (value(x) := value at nearest grid point at or below x)
# Optionally, the average rather than area-under-the-curve may be computed (set average=True). In that case this function is conceptually analogous to numpy.mean(values,axis=axis), but taking grid steps and interpolation between grid points into account.
def integrate_on_grid(grid,					# 1D numeric array of size N, listing grid points in ascending order
					  values, 				# n-dimensional numeric array with one axis synchronized with grid[], listing values to integrate along the specified axis. May contain nans if check_nan==True.
					  axis		= 0,		# axis along which to integrate. For example, if values[] has shape N1 x N2 x N3 and axis=1 then the result will have shape N1 x N3.
					  grid_min	= -inf,		# optional minimum grid value to integrate over
					  grid_max	= inf,		# optional maximum grid value to integrate over
					  average 	= False,	# if True, compute the average rather than the sum.
					  degree	= 1,		# polynomial degree for interpolating the function between successive grid points. Either 1 ("linear") or -1 ("const", use last available value, i.e. at or below each requested grid point).
					  check_nan = True):
	NG = len(grid)
	if(NG!=values.shape[axis]): raise ValueError("values has invalid shape %s: Expected size %d for axis %d"%(sprint_shape(values),NG,axis))	
	n = values.ndim
	non_integrated_axes = tuple(range(axis))+tuple(range(axis+1,n))
	if(check_nan): 
		valids 	= numpy.where(numpy.all(numpy.isfinite(values),axis=non_integrated_axes))[0]
		grid 	= grid[valids]
		values 	= numpy.take(values, indices=valids, axis=axis)
	if(numpy.isfinite(grid_min) or numpy.isfinite(grid_max)):
		valids 	= numpy.where((grid>=grid_min) & (grid<=grid_max))[0]
		grid 	= grid[valids]
		values 	= numpy.take(values, indices=valids, axis=axis)
	output_shape = values.shape[:axis]+values.shape[(axis+1):]
	grid = numpy.expand_dims(grid,axis=non_integrated_axes)
	if(NG==1): 
		return (numpy.take(values,indices=0,axis=axis) if average else numpy.zeros(output_shape, dtype=values.dtype))
	elif(NG==0):
		return numpy.full(output_shape,(nan if average else 0))
	if(degree==1): 
		# trapezoid integration scheme, exact for piecewise linear functions
		I = numpy.sum(0.5*(slice_along_axis(values,indices=slice(1,None),axis=axis)+slice_along_axis(values,indices=slice(0,-1),axis=axis))*numpy.diff(grid,axis=axis), axis=axis)
	elif(degree==-1):
		# integrate piecewise constant function
		I = numpy.sum(slice_along_axis(values,indices=slice(0,-1),axis=axis)*numpy.diff(grid,axis=axis), axis=axis)
	if(average): I /= (slice_along_axis(grid,indices=-1,axis=axis)-slice_along_axis(grid,indices=0,axis=axis))
	return I


##############################################################
# MARK Arrays and linear algebra


# create a 2D banded matrix based on a provided sequence of bands
# bands must be a sequence of length NB, each element of which is either a scalar or a 1D array or a list, representing the contents of one band
# offsets must be a sequence of length NB, each element of which is an integer specifying the location of the band relative to the diagonal (>0 for above-diagonals, <0 for below-diagonals)
# This function returns a fill 2D array, i.e. not in sparse format
def get_band_matrix(N, bands, offsets, dtype=float):
	A = numpy.zeros((N,N), dtype=dtype)
	for band,offset in zip(bands, offsets):
		A[range(0,M-offset), range(offset,M)] = band
	return A


# slice a 1D or multidimensional array based on a sequence of indices, properly handling broadcasted axes
# indices must be a sequence, each element of which is some type of valid slicing object, for example a range or slice object, or a single integer, or a list of integers etc.
# If the array has N axes and len(indices)=M, with M<=N, then the last (N-M) axes of the array are not sliced, i.e., they are kept in their entirety
# For example, if array is of shape N1 x N2 x 1 x 2, then the last two axes are broadcasted during indexing; if indices[2] and indices[3] are sequences, then the returned array's last 2 axes will again be of size 1; if indices[2] & indices[3] are scalars, the returned array will lack those axes
# If the array is in fact a scalar, then that scalar is returned as-is.
def slice_broadcasted(array, *indices):
	if(numpy.isscalar(array)): return array
	indices = list(indices)
	for axis in range(len(indices)):
		if(array.shape[axis]==1): indices[axis] = (0 if numpy.isscalar(indices[axis]) else [0])
	return array[*indices,...]


def array_if_scalar(value, shape, dtype=float):
	if(numpy.isscalar(value)): return numpy.full(shape, value, dtype=dtype)
	else: return numpy.asarray(value)


def array_insert(array, indices, values, inplace=False):
	if(inplace):
		new_array = array	
	else:
		new_array = array.copy()
	new_array[indices] = values
	return new_array


# efficiently slice a multidimensional numpy array along a specific axis
# This is a faster alternative to numpy.take(), as the latter always creates a copy of the data
# In contrast to numpy.take(), this function also supports slice objects as indices.
# If broadcasted==True and array.shape[axis]==1, then this is the same as requesting only the 0'th index of that axis.
def slice_along_axis(array, axis, indices, broadcasted=False):
	if(broadcasted and numpy.isscalar(array)): return array
	axis = numpy.core.multiarray.normalize_axis_index(axis, ndim=array.ndim)
	if(broadcasted and (array.shape[axis]==1)):
		indices = (0 if numpy.isscalar(indices) else [0])
	slices 	= (slice(None),)*axis + (indices,)
	return array[slices]


# Find the maximum value in an array and its location
# This function is equivalent to (numpy.argmax(array,axis=axis), numpy.max(array,axis=axis)), but avoiding duplicate computation
# Note that if axis is not None and array.ndim>=2, the returned location & value will be arrays, each with array.ndim-1 axes.
def argmax2(array, axis=None, check_nan=False):
	if((axis is None) or (array.ndim==1)):
		if(check_nan):
			location = numpy.nanargmax(array)
		else:
			location = numpy.argmax(array)
		maxvalue = array.flat[location]
		return location, maxvalue
	else:
		if(check_nan):
			locations = numpy.nanargmax(array, axis=axis, keepdims=True)
		else:
			locations = numpy.argmax(array, axis=axis, keepdims=True)
		maxvalues = numpy.take_along_axis(array, indices=locations, axis=axis)
		return numpy.squeeze(locations, axis=axis), maxvalues
	

##############################################################
# MARK Plotting


def char2hex(c):
	a = ord(c);
	if((a>=48) and (a<=57)): return (a-48);
	elif((a>=65) and (a<=90)): return 10+(a-65);
	elif((a>=97) and (a<=122)): return 10+(a-97);
	else: return numpy.nan;
	

# takes a matplotlib.cm tuple or hex-color spec (e.g. '#E0A3FF') and returns a grey-scale version of the color in the same format as the original
def color_to_grey_scale(C):
	if(isinstance(C, str)):
		# hex string
		return '#'+("%02x"%(int((char2hex(C[1])*16+char2hex(C[2]))*0.21 + (char2hex(C[3])*16+char2hex(C[4]))*0.72 + (char2hex(C[5])*16+char2hex(C[6]))*0.07)))*3
	else:
		# tuple
		return 	(C[0]*0.21 + C[1]*0.72 + C[2]*0.07);


# given an axis range (e.g., for plotting), determine by what factor the values should be rescaled to make the numeric values more meaningful (i.e., around order of magnitude 1-100).
# The function returns the best factor as well as the correspondingly modified units
# for example, if max_abs_value==4e-7 and the original units are "M", this will return the factor 1e9 with new units "nM"
order_of_magnitude_prefixes_up   = ["k", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q"]
order_of_magnitude_prefixes_down = ["m", "μ", "n", "p", "f", "a", "z", "y", "r", "q"]
def best_axis_factor(max_abs_value, unit):
	# first remove any existing common multipliers if already present
	# For example, if the original unit is km then it will become "m" and the range is multiplied by 1000
	if(max_abs_value==0): return 1, unit
	if(unit[0].isnumeric()): return 1, unit # maybe a unit of the form 1/..., which are not supported yet by this function
	factor = 1
	if((p:=next((p for p,prefix in enumerate(order_of_magnitude_prefixes_up) for u in ["M", "s", "m"] if (unit==prefix+u)), -1))>=0):
		unit 	= unit[1]
		factor 	= 10**(3*(p+1))
	elif((p:=next((p for p,prefix in enumerate(order_of_magnitude_prefixes_down) for u in ["M", "s", "m"] if (unit==prefix+u)), -1))>=0):
		unit 	= unit[1]
		factor	= 10**(-3*(p+1))
	max_abs_value *= factor
	# determine order of magnitude, in multiples of 1000 (e.g., 1 for 1000, 2 for 1000,000 etc)
	oom = int(numpy.log10(max_abs_value)//3)
	if(oom==0):
		return factor, unit
	elif(oom<0):
		return factor*(10**(-oom*3)), order_of_magnitude_prefixes_down[-oom-1]+unit
	elif(oom>0):
		return factor*(10**(-oom*3)), order_of_magnitude_prefixes_up[oom-1]+unit
	
		

# plot one or more scalar fields over time & 1D-space as contour plots
def plot_contour(Tgrid, 						# 1D numeric array of size NT, listing times in ascending order
				 Zgrid,							# 1D numeric array of size NZ, listing spatial coordinates (Z) in ascending order
				 value,							# either a 2D numeric array of size NT x NZ or a 3D numeric array of size NT x NZ x M, where M is the number of components to plot (one contour plot per component)
				 plot_file,						# string, path to the output plot file. The file extension determines the file format (e.g., "pdf" or "jpg").
				 Tlabel			= "time", 		# string, label for time-axis
				 Zlabel			= "location",	# string, label for the spatial (Z) axis
				 cbar_label		= "",			# either a single string, or a string sequence of length M, optional labels for the value axes for the color bar. Only relevant if show_cbar=True.
				 cmap			= "inferno",	# string, color map name to use. Must be one of Matplotlib's valie color map names, such as "inferno" or "hot"
				 value_log		= False,		# whether to map values to colors on a logarithmic scale
				 reverseZ		= False,		# whether to reverse the Z axis (Z increases downwards)
				 verticalZ		= True,			# whether the Z axis should be vertical (and thus multiple components are plotted below each other). If False, multiple components are plotted side by side (left to right). Hence, in either case the various plots share the same time axis.
				 plot_title		= "",			# string, optional plot super title, i.e., for the entire collection of subplots
				 subplot_titles	= None,			# optional string sequence of length M, titles to show above each individual subplot
				 show_cbar		= True,			# show color bar next to each contour plot
				 Nlevels		= 30,			# integer, number of color levels
				 plot_sizeZ		= 1.5,			# float, plot size in inches corresponding to the Z axis (e.g. plot height if verticalZ==True)
				 plot_sizeT		= 3,			# float, plot size in inches corresponding to the time axis (e.g. plot width if verticalZ==True)
				 dpi			= 300):			# dots per inch. Only relevant for non-PDF outputs
	# basic input checking
	if(value.ndim==2): value = value.reshape(value.shape+(1,))
	Zgrid = numpy.asarray(Zgrid)
	Tgrid = numpy.asarray(Tgrid)
	NT, NZ, M = value.shape
	if((NZ,)!=Zgrid.shape): raise ValueError("Zgrid has shape %s, but expected a 1D array of length %d"%(sprint_shape(Zgrid),NZ))
	if((NT,)!=Tgrid.shape): raise ValueError("Tgrid has shape %s, but expected a 1D array of length %d"%(sprint_shape(Tgrid),NT))
	
	Nrows 		= (M if verticalZ else 1)
	Ncols 		= (1 if verticalZ else M)
	plot_width  = (plot_sizeT if verticalZ else plot_sizeZ)
	plot_height = (plot_sizeZ if verticalZ else plot_sizeT)
	fig, axes = pyplot.subplots(nrows	= Nrows, 
								ncols	= Ncols, 
								figsize	= (plot_width*Ncols+2, plot_height*Nrows+2), 
								sharex	= verticalZ, 
								sharey	= (not verticalZ),
								squeeze	= False)
	if(verticalZ): pyplot.subplots_adjust(hspace=0.4)
	else: pyplot.subplots_adjust(wspace=0.4)
	if(plot_title!=""): fig.suptitle(plot_title)
	axes = axes.ravel()
	for m,ax in enumerate(axes):
		component_value = value[...,m]
		if(value_log):
			min_nonzero = numpy.min(component_value[component_value>0])
			color_levels = numpy.concatenate(([0],numpy.geomspace(min_nonzero, numpy.max(component_value), num=Nlevels)))
		else:
			color_levels = Nlevels
		cplot = ax.contourf(	(Tgrid if verticalZ else Zgrid),
								(Zgrid if verticalZ else Tgrid),
								(component_value.T if verticalZ else component_value),
								norm 	= ("log" if value_log else "linear"),
								levels	= color_levels,
								cmap	= cmap)
		cplot.set_edgecolor("face")
		if((subplot_titles is not None) and (subplot_titles[m]!="")): ax.set_title(subplot_titles[m])
		if((m==M-1) and verticalZ):
			ax.set_xlabel(Tlabel)
		elif((m==0) and (not verticalZ)):
			ax.set_ylabel(Tlabel)
		if(verticalZ):
			ax.set_ylabel(Zlabel)
			if(reverseZ): ax.invert_yaxis()
		else:
			ax.set_xlabel(Zlabel)
			if(reverseZ): ax.invert_xaxis()
		if(show_cbar): 
			cbar = pyplot.colorbar(cplot)
			cbar.ax.locator_params(nbins=5) # reduce the number of colorbar ticks
			this_cbar_label = (cbar_label if isinstance(cbar_label,str) else cbar_label[m])
			if(this_cbar_label!=""): cbar.set_label(this_cbar_label, rotation=270, labelpad=15)
	if((parent:=os.path.dirname(plot_file))!=""): os.makedirs(parent, exist_ok=True)
	fig.savefig(plot_file, dpi=dpi, format=os.path.splitext(plot_file)[1][1:], bbox_inches='tight')
	pyplot.close(fig)



# plot one or more scalar fields over 1D-space as curves
def plot_profile(Zgrid,							# 1D numeric array of size NZ, listing spatial coordinates in ascending order
				 value,							# either a 1D numeric array of size NZ or a 2D numeric array of size NZ x M, where M is the number of components to plot (one contour plot per component)
				 plot_file,						# string, path to the output plot file. The file extension determines the file format (e.g., "pdf" or "jpg").
				 Zlabel			= "location",	# string, label for the spatial (Z) axis
				 Vlabel			= "value",		# either a single string, or a string sequence of length M, labels for the value axes
				 color			= "#206ebd",	# string or sequence of strings, curve color(s) to use
				 value_scale	= "linear",		# axis scale for the values. Options are "linear", "log", "symlog:"<linthresh> or "power:"<exponent>
				 reverseZ		= False,		# whether to reverse the Z axis (Z increases downwards or leftwards)
				 verticalZ		= True,			# whether the Z axis should be vertical (and thus multiple components are plotted next to each other). If False, multiple components are plotted below one another (from top to bottom). Hence, in either case the various plots share the same Z axis.
				 value_ref		= None,			# optional value at which to draw a reference line. Either a scalar (same value_ref for all plots) or a numeric sequence of length M.
				 plot_title		= "",			# string, optional plot super title, i.e., for the entire collection of subplots
				 subplot_titles	= None,			# optional string sequence of length M, titles to show above each individual subplot
				 plot_sizeZ		= 2,			# float, plot size in inches corresponding to the Z axis (e.g. plot height if verticalZ==True)
				 plot_sizeV		= 1.5,			# float, plot size in inches corresponding to the value axis (e.g. plot width if verticalZ==True)
				 dpi			= 300):			# dots per inch. Only relevant for non-PDF outputs
	# basic input checking
	if(value.ndim==1): value = value.reshape(value.shape+(1,))
	elif(value.ndim>2): raise ValueError("Input value array must be 1D or 2D; instead, got %dD"%(value.ndim))
	NZ, M = value.shape
	Zgrid = numpy.asarray(Zgrid)
	if((NZ,)!=Zgrid.shape): raise ValueError("Zgrid has shape %s, but expected a 1D array of length %d"%(sprint_shape(Zgrid),NZ))
	
	if(value_scale.startswith("power:")):
		exponent = float_or_nan(value_scale[len("power:"):])
		value_scale_specs = {"value":"function", "functions":(lambda y: power_transform(y,exponent), lambda y: power_transform_inv(y,exponent))}
	elif(value_scale.startswith("symlog:")):
		value_scale_specs = {"value":"symlog", "linthresh":float_or_nan(value_scale[len("symlog:"):])}
	else:
		value_scale_specs = {"value":value_scale}
	
	Nrows 		= (1 if verticalZ else M)
	Ncols 		= (M if verticalZ else 1)
	plot_width  = (plot_sizeV if verticalZ else plot_sizeZ)
	plot_height = (plot_sizeZ if verticalZ else plot_sizeV)
	fig, axes = pyplot.subplots(nrows	= Nrows, 
								ncols	= Ncols, 
								figsize	= (plot_width*Ncols, plot_height*Nrows),
								sharex	= (not verticalZ), 
								sharey	= verticalZ,
								squeeze	= False)
	if(not verticalZ): pyplot.subplots_adjust(hspace=0.4)
	if(plot_title!=""): fig.suptitle(plot_title)
	axes = axes.ravel()
	for m,ax in enumerate(axes):
		component_value = value[...,m]
		ax.plot((value[:,m] if verticalZ else Zgrid),
				(Zgrid if verticalZ else value[:,m]), 
				c 		  = (color if isinstance(color,str) else color[m]),
				linewidth = 1.5,
				linestyle = "solid")
		if(value_ref is not None):
			if(verticalZ): ax.axvline(slice_broadcasted(value_ref,m), color="#505050")
			else: ax.axhline(slice_broadcasted(value_ref,m), color="#505050")
		if((subplot_titles is not None) and (subplot_titles[m]!="")): ax.set_title(subplot_titles[m])
		this_Vlabel = (Vlabel if isinstance(Vlabel,str) else Vlabel[m])
		if((m==0) and verticalZ):
			ax.set_ylabel(Zlabel)
			if(reverseZ): ax.invert_yaxis()
		elif((m==M-1) and (not verticalZ)):
			ax.set_xlabel(Zlabel)
			if(reverseZ): ax.invert_xaxis()
		if(verticalZ):
			ax.set_xlabel(this_Vlabel)
			ax.set_xscale(**value_scale_specs)
		else:
			ax.set_ylabel(this_Vlabel)
			ax.set_yscale(**value_scale_specs)
	if((parent:=os.path.dirname(plot_file))!=""): os.makedirs(parent, exist_ok=True)
	fig.savefig(plot_file, dpi=dpi, format=os.path.splitext(plot_file)[1][1:], bbox_inches='tight')
	pyplot.close(fig)



# plot a heatmap of regularly spaced grid data
# The xvalues and yvalues (corresponding to columns & rows) may be discrete labels or numerical, however in the latter case they must be regularly spaced (since heatmap cells will be regularly sized along each axis)
def plot_heatmap(plot_file,					# string, path to the output plot file. The file extension determines the file format (e.g., "pdf" or "jpg").
				 table,						# 2D numeric numpy array of size NR x NC, listing the values determining the heatmap colors. Columns with greater index are plotted further to the right, rows with greater index are plotted further down.
				 row_values		= None,		# optional list or 1D numpy array of strings or numbers, of length NC, listing labels or x-values for the various table columns. If of type float, these will be interpreted as axis values, otherwise as discrete labels.
				 column_values	= None,		# optional list or 1D numpy array of strings or numbers, of length NR, listing labels or y-values for the various table rows. If of type float, these will be interpreted as axis values, otherwise as discrete labels.
				 show_colorbar	= True,
				 colormap		= "hot",	# name of the colormap to use, or a colormap object. For a list of options see: https://matplotlib.org/stable/tutorials/colors/colormaps.html
				 colormap_min	= None,		# optional minimum value covered by the colormap
				 colormap_max	= None,		# optional maximum value covered by the colormap
				 xlabel			= "", 		# optional label for x-axis, e.g. 'time (seconds)'
				 ylabel			= "",		# optional label for y-axis, e.g. 'gene'
				 clabel			= "",		# label for the colorbar axis, e.g. 'activity'
				 show_values	= False,	# if True, the numerical values are shown in each heatmap cell. If False, no values are shown. If a 2D boolean numpy array (of size NR x NC), values will be displayed in those cells where show_values[r,c] is True.
				 value_format	= "%.2g",	# format specifier for converting table values to strings, when displaying them inside cells. Only relevant if show_values!=False
				 show_grid		= "auto",	# either "auto", "vertical", "horizontal", "all" or "none"
				 cluster		= "none",	# whether and how to reorder rows & columns via hierarchically clustering. Options are "none", "rows", "columns", "rows_and_columns_joint" and "rows_and_columns_separate".
				 plot_title		= "",
				 plot_width		= None,		# numeric, plot width in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of columns.
				 plot_height	= None,		# numeric, plot height in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of rows.
				 dpi			= 300):		# resolution of the image, in dots per inch. Note that fonts scale with the inch-size of the figure, so a higher DPI means more smooth text for bitmap images. For PDFs this parameter is irrelevant. 
	# basic options checking
	(NR, NC) = table.shape
	if(plot_width is None):  plot_width = 0.5 * NC
	if(plot_height is None): plot_height = 0.5 * NR
	if(cluster not in ["none", "rows", "columns", "rows_and_columns_joint", "rows_and_columns_separate"]): raise ValueError("Invalid option '%s' for rows_and_columns_separate"%(rows_and_columns_separate))
	if((column_values is not None) and (len(column_values)!=NC)): raise ValueError("Invalid number of column_values (%d): Expected %d, based on the number of table columns"%(len(column_values),NC))
	if((row_values is not None) and (len(row_values)!=NR)): raise ValueError("Invalid number of row_values (%d): Expected %d, based on the number of table rows"%(len(row_values),NR))
	
	# determine the nature and limits of the axes
	xnumeric 	= (column_values is not None) and ((isinstance(column_values, numpy.ndarray) and (column_values.dtype==float)) or (isinstance(column_values, list) and isinstance(column_values[0], float)))
	ynumeric	= (row_values is not None) and ((isinstance(row_values, numpy.ndarray) and (row_values.dtype==float)) or (isinstance(row_values, list) and isinstance(row_values[0], float)))
	xlim		= ((column_values[0]-0.5*(column_values[min(NC-1,1)]-column_values[0]) if xnumeric else -0.5), (column_values[-1]+0.5*(column_values[-1]-column_values[-min(NC,2)]) if xnumeric else NC-0.5))
	ylim	 	= ((row_values[0]-0.5*(row_values[min(NR-1,1)]-row_values[0]) if ynumeric else -0.5), (row_values[-1]+0.5*(row_values[-1]-row_values[-min(NR,2)]) if ynumeric else NR-0.5))
	data_min 	= numpy.nanmin(table)
	data_max	= numpy.nanmax(table)
	if(colormap_min is None): colormap_min = data_min
	if(colormap_max is None): colormap_max = data_max
	
	# initialize figure
	bottom_margin	= 0.5 	# inches
	top_margin		= 0.5	# inches
	left_margin		= 1		# inches
	right_margin	= (1 if show_colorbar else 0.3) 	# inches
	total_width		= plot_width+left_margin+right_margin
	total_height	= plot_height+top_margin+bottom_margin
	fig 			= pyplot.figure(figsize=(total_width, total_height))
	ax				= pyplot.gca()
	
	# cluster rows & columns if needed
	if(("rows" in cluster) and (NR>1)):
		row_order = scipy.cluster.hierarchy.leaves_list(scipy.cluster.hierarchy.optimal_leaf_ordering(scipy.cluster.hierarchy.ward(table),table))
		if(row_values is not None): row_values = row_values[row_order]
		table = table[row_order,:]
		if(isinstance(show_values,numpy.ndarray)): show_values = show_values[row_order,:]
	if(("columns" in cluster) and (NC>1)):
		if(cluster=="rows_and_columns_joint"):
			column_order = row_order
		else:
			column_order = scipy.cluster.hierarchy.leaves_list(scipy.cluster.hierarchy.optimal_leaf_ordering(scipy.cluster.hierarchy.ward(table.T),table.T))
		if(column_values is not None): column_values = [column_values[k] for k in column_order]
		table = table[:,column_order]
		if(isinstance(show_values,numpy.ndarray)): show_values = show_values[:,column_order]

	# plot heatmap
	colormap = matplotlib.cm.get_cmap(colormap)
	im = pyplot.imshow(	table, 
						cmap			= colormap,
						aspect			= (plot_height/(1 if (NR==1) else (ylim[1]-ylim[0])))/(plot_width/(1 if (NC==1) else (xlim[1]-xlim[0]))),
						interpolation	= "nearest", 
						vmin			= colormap_min, 
						vmax			= colormap_max,
						extent 			= (xlim[0], xlim[1], ylim[1], ylim[0]))
						
	# set axis ticks
	if(not xnumeric): 
		if(column_values is None):
			ax.set_xticks([], minor=False)
		else:
			rotate_xticks = (next((1 for x in column_values if (len(x)>2)), -1)>0)
			ax.set_xticks(numpy.arange(0,NC,1), labels=column_values, minor=False, rotation=(45 if rotate_xticks else 0), horizontalalignment=("right" if rotate_xticks else "center"))
		ax.set_xticks(numpy.arange(xlim[0],xlim[1],numpy.sign(xlim[1]-xlim[0])), minor=True)
	if(not ynumeric): 
		if(row_values is None):
			ax.set_yticks([], minor=False)
		else:
			ax.set_yticks(numpy.arange(0,NR,1), labels=row_values, minor=False)
		ax.set_yticks(numpy.arange(ylim[0],ylim[1],numpy.sign(ylim[1]-ylim[0])), minor=True)
	if(not (xnumeric and ynumeric)): 
		ax.tick_params(which='minor', length=0, color='black') # hide minor ticks
		
	# show grid lines if needed
	if(show_grid=="auto"): 
		show_grid = ("all" if (xnumeric and ynumeric) else ("vertical" if ynumeric else ("horizontal" if xnumeric else "none")))
	if(show_grid!="none"):
		ax.grid(visible=True, axis={"all":"both", "vertical":"x", "horizontal":"y"}[show_grid], color='w', linestyle='-', linewidth=1, which='minor') # show white grid lines at minor tick locations (to separate heatmap cells)

	# plot labels and titles
	pyplot.title(plot_title+"\n",fontsize="medium", fontname="Helvetica") # this should come before adding the color bar
	if(xlabel!=""): pyplot.xlabel(xlabel, fontsize="medium", fontname="Helvetica")
	horizontal_ylabel = (len(ylabel)<=2)
	if(ylabel!=""): pyplot.ylabel(ylabel, fontsize="medium", fontname="Helvetica", rotation=(0 if horizontal_ylabel else 90), horizontalalignment=('right' if horizontal_ylabel else "center"))
				
	# add colorbar
	if(show_colorbar):
		divider  = mpl_toolkits.axes_grid1.make_axes_locatable(ax)
		cax 	 = divider.append_axes("right", size=0.15, pad=0.05)
		colorbar = pyplot.colorbar(im, cax=cax)
		colorbar.ax.set_ylabel(clabel, rotation=-90, va="bottom")

	# show values in grid cells
	if(show_values is not False):
		relative_values = (table - colormap_min)/(colormap_max-colormap_min)
		for r in range(NR):
			for c in range(NC):
				if((show_values is True) or show_values[r,c]):
					ax.text((column_values[c] if xnumeric else c), 
							(row_values[r] if ynumeric else r), 
							value_format%(table[r,c]), 
							ha="center", 
							va="center", 
							color=("white" if (color_to_grey_scale(colormap(relative_values[r,c]))<0.5) else "black"), 
							fontsize="small")

	if((parent:=os.path.dirname(plot_file))!=""): os.makedirs(parent, exist_ok=True)
	fig.savefig(plot_file, dpi=dpi, format=os.path.splitext(plot_file)[1][1:], bbox_inches='tight')
	pyplot.close(fig)



# add marks on the horizontal or vertical axis of a plot, with optional labels
def plot_axis_marks(marks, 				 # sequence of axis marks to plot. Each element in marks[] should be a dictionary with the mandatory key "value" and optional keys "label", "linewidth", "linestyle", "linecolor". marks could for example represent major events in the stock market, to be highlighted in the plots. For example, marks = [{"value":2008, "label":"financial crisis"}, {"value":2000.3, "label":"dot-com bubble", "linecolor":"#AA0000"}]
					axis,				 # either "x" or "y"
					bbox_extra_artists): # (input/output) a list of boundary box artists. Will be updated in-place
	if(marks is None): return # nothing to do
	xlim = pyplot.gca().get_xlim()
	ylim = pyplot.gca().get_ylim()
	if(axis=="x"):
		# plot vertical mark lines
		mark_values = [mark["value"] for mark in marks]
		pyplot.scatter(	mark_values,
						[ylim[0]]*len(marks),
						c			= "#404040",
						edgecolor	= "#303030",
						s			= 5,
						marker		= 'o',
						clip_on		= False,
						alpha		= 0.8)
		for mark,mark_value in zip(marks,mark_values):
			# show vertical line at this mark
			if(not numpy.isfinite(mark_value)): continue
			pyplot.axvline(	x			= mark_value, 
							color		= mark.get("linecolor","#202020"),
							alpha		= 0.5,
							linewidth	= mark.get("linewidth",0.75), 
							linestyle	= mark.get("linestyle","dashed"))
			if((mark_label:=mark.get("label",None)) is not None):
				label = pyplot.text(x=mark_value, y=ylim[1], s=mark_label, horizontalalignment="center", verticalalignment="bottom")
				bbox_extra_artists.append(label)
	else:
		# plot horizontal mark lines
		mark_values = [mark["value"] for mark in marks]
		pyplot.scatter([xlim[0]]*len(marks), 
						mark_values, 
						c			= "#404040",
						edgecolor	= "#303030",
						s			= 5,
						marker		= 'o',
						clip_on		= False,
						alpha		= 0.8)
		for mark,mark_value in zip(marks,mark_values):
			# show horizontal line at this mark
			if(not numpy.isfinite(mark_value)): continue
			pyplot.axhline(	y			= mark_value, 
							color		= mark.get("linecolor","#202020"),
							alpha		= 0.5,
							linewidth	= mark.get("linewidth",0.75), 
							linestyle	= mark.get("linestyle","dashed"))
			if((mark_label:=mark.get("label",None)) is not None):
				label = pyplot.text(x=xlim[1], y=mark_value, s=" "+mark_label.replace("\n","\n "), horizontalalignment="left", verticalalignment="center")
				bbox_extra_artists.append(label)



# Create a barplot showing the values of multiple scalar variables.
# Optionally, one can group bars into NG groups of equal size NM, with extra space shown between groups. In this case, the values should be passed in the form of a NG x NM array, where each row corresponds to a separate group of bars and each m-th column corresponds to the m-th member in each group.
# Alternatively, each group of bars can contain a different number of members (NM differs between groups). In this "inhomogeneous" case, the values should be passed in the form of list of length NG whose elements are themselves lists of floats. Hence, each values[g] corresponds to a separate group of len(values[g]) bars. member_names[] & member_colors[] & errors[] will also need to have a similar structure.
# Thus, plotted bars show a flattened version of the data (in row-major format). In the case where NG=1, one obtains a simple barplot showing NM bars.
# This function displays similar information as stacked barplots, with the difference that bar segments are not stacked but instead interleaved (displayed next to each other).
def interleaved_barplot(plot_file,						# string, path to the output plot file. The file extension determines the file format (e.g., "pdf" or "jpg").
				 		values,							# either a 1D numpy array of size NM, or a 2D numeric numpy array of size NG x NM, or a list of lists of floats ("inhomogeneous" case), listing the values determining the bar heights. Rows correspond to groups of bars, columns correspond to individual members within each group.
						errors				= None,		# optional errors to be shown as whiskers. In the homogeneous case, either a 2D numpy array of size NG x NM specifying symmetric errors (+/- relative to the bar values) or a 3D numpy array of size NG x NM x 2 specifying errors downward and upward from the bar values (e.g., for bar (g,m) the whiskers will span from values[g,m]+errors[g,m,0] to values[g,m]+errors[g,m,1]). Attention: In the latter case, errors should NOT specify the actual values of the whisker ends, but instead their signed deviations from the bar values. In the inhomogeneous case, errors[] must be a list of lists of the same structure as values[], with elements either being floats (symmetric errors) or 2-tuples of floats (separate downward & upward signed errors).
						group_names			= None,		# optional list or 1D numpy array of strings of length NG, specifying names/labels for the various groups.
						member_names		= None,		# optional list or 1D numpy array of strings of length NM (homogeneous case) or a list of lists of strings (inhomogeneous case), specifying names/labels for the various group members. Alternatively, in the homogeneous case, this may be a 2D array of size NG x NM or an equivalent list of lists, specifying separate member labels for each group.
						member_colors		= None,		# either None, or a single color, or a list or 1D numpy array of length NM (in the homogeneous case) or a list of lists of colors (in the inhomogeneous case), specifying colors for the individual group members. The m-th member in each group will have color member_colors[m]. If a single color, then all bars will have the same color. If None, this is automatically chosen.
						vertical			= True,		# if True, bars are vertical, with labels underneath. If False, bars are horizontal, with labels to their left.
						frame_color			= "#000000",# HEX color for the plot's frame
						var_label			= "", 		# optional axis label for the various variables, e.g. 'stock'
						value_label			= "",		# optional label for the value axis, e.g. 'value (USD)'
						value_ref			= None,		# optional numeric, reference value at which to draw a horizontal line (if vertical==True) or a vertical line (if vertical==False)
						value_marks			= None,		# optional list of dictionaries, specifying value marks. If vertical==True, then value marks are shown at the right of the plot with horizontal extension lines.
						only_positives		= False,	# whether to only show positive values. If False, negative values are depicted as bars in the opposite direction. If True, no bars will appear for negative values.
						show_grid			= False,	# whether to show grid lines (horizontal lines if vertical==True, or vertical lines if vertical==False)
						plot_title			= "",		# string, e.g. 'Predicted stock values'. May be multi-line.
						plot_width			= None,		# numeric, plot width in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of bars if vertical==True.
						plot_height			= None,		# numeric, plot height in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of bars if vertical==False.
						dpi					= 300):		# resolution of the image, in dots per inch. Note that fonts scale with the inch-size of the figure, so a higher DPI means more smooth text for bitmap images. For PDFs this parameter is irrelevant. 
	# basic options checking
	inhomogeneous = (is_list_or_tuple(values) and (len(values)>0) and hasattr(values[0],"__len__") and (len(set(len(vv) for vv in values))>1)) # whether the number of members differs between groups
	if(not inhomogeneous): 
		values = numpy.asarray(values)
		if(values.ndim==1): values = values.reshape([1,len(values)]) # reformat as 2D numpy array with a single row
		(NG, NM) = values.shape
		Nbars = NG * NM
	else:
		NG    = len(values)
		NM 	  = numpy.asarray([len(vv) for vv in values], dtype=int)
		Nbars = sum(len(vv) for vv in values)
	bbox_extra_artists = []
			
	Nslots = Nbars+NG-1 # number of bar slots, i.e., number of bars + gaps between groups
	as_pdf = plot_file.lower().endswith(".pdf")
	
	# initialize figure
	if(plot_width is None):  plot_width = (0.4 * Nslots if vertical else 2)
	if(plot_height is None): plot_height = (2 if vertical else 0.25 * Nslots)
	bottom_margin	= 0.5 	# inches
	top_margin		= 0.5	# inches
	left_margin		= 1		# inches
	right_margin	= 1		# inches
	total_width		= plot_width+left_margin+right_margin
	total_height	= plot_height+top_margin+bottom_margin
	fig 			= pyplot.figure(figsize=(total_width, total_height))
	ax				= pyplot.gca()
	
	# figure out bar positions, colors and labels
	if(member_colors is None):
		if(NG==1):
			bar_fill_colors = "#AAD0E3"
			bar_edge_colors = "#505050"
		else:
			bar_fill_colors = [GENERIC_BAR_COLOR_PALETTE[m%len(GENERIC_BAR_COLOR_PALETTE)] for g in range(NG) for m in range(NM[g] if inhomogeneous else NM)]
			bar_edge_colors = [change_brightness(color,0.3) for color in bar_fill_colors]
	elif(is_list_or_tuple(member_colors) or isinstance(member_colors,numpy.ndarray)): 
		if(inhomogeneous):
			# assuming that member_colors is also an inhomogeneous list of lists of the same structure as values[]
			bar_fill_colors = [member_colors[g][m] for g in range(NG) for m in range(NM[g])]
		else:
			bar_fill_colors = numpy.tile(member_colors, NG)
		bar_edge_colors = [change_brightness(color,0.3) for color in bar_fill_colors]
	else:
		bar_fill_colors = member_colors
		bar_edge_colors = "#505050"
	if(isinstance(member_names,numpy.ndarray) and (member_names.ndim==2)):
		# the caller specified different member names for each group, as a 2D numpy array
		if(group_names is None): 
			bar_labels = member_names.flat
		else:
			bar_labels = [(("" if (group_names is None) else (group_names[g]+" - "))+member_names[g,m]) for g in range(NG) for m in range(NM)]
	elif(isinstance(member_names,list) and isinstance(member_names[0],list)):
		# the caller specified different member names for each group, as a list of lists
		bar_labels = [(("" if (group_names is None) else (group_names[g]+" - "))+member_names[g][m]) for g in range(NG) for m in range(NM[g] if inhomogeneous else NM)]
	else:
		# the caller specified the same member names for each group, presumably as a list of strings
		bar_labels = [join_nonempty((("" if (group_names is None) else group_names[g]), ("" if (member_names is None) else member_names[m])), " - ") for g in range(NG) for m in range(NM)]
		
	# determine bar positions
	if(inhomogeneous):
		next_position = 0
		bar_positions = []
		for g in range(NG):
			for m in range(NM[g]):
				bar_positions.append(next_position)
				next_position += 1
			next_position += 1
	else:
		bar_positions = [(g*(NM+1)+m) for g in range(NG) for m in range(NM)]

	# show grid lines
	if(show_grid):
		pyplot.grid(visible=True, which='major', axis=('y' if vertical else 'x'), alpha=1, color="#B0B0B0", zorder=0)

	# plot reference line
	if(value_ref is not None):
		if(vertical):
			# horizontal reference line
			pyplot.axhline(y=value_ref, color="#000000", alpha=1, linewidth=1)
		else:
			# vertical reference line
			pyplot.axvline(x=value_ref, color="#000000", alpha=1, linewidth=1)			
		
	# format errors as needed, either into a 1D array of length Nbars (symmetric errors) or into a 2D array of size 2 x Nbars (separate downward & upward errors).
	if(errors is not None):
		if(inhomogeneous):
			# assuming that errors is also an inhomogeneous list of lists of the same structure as values[], with elements being either scalars (symmetric deviations) or 2-tuples (downward & upward deviations)
			errors = numpy.asarray([errors[g][m] for g in range(NG) for m in range(NM[g])], dtype=float)
			if(errors.ndim==2):
				# errors are not symmetric, i.e. separate downward & upward errors are given for each bar				
				errors = numpy.moveaxis(errors, source=1, destination=0).reshape((2,Nbars))
				errors[0,:] *= -1 # convention for pyplot.bar and pyplot.barh is to define errors as absolute deviations from the bar's value, rather than signed deviations.
				errors = numpy.maximum(0,errors) # pyplot does not accept negative error values
		else:
			errors = numpy.asarray(errors, dtype=float)
			if(errors.ndim==2):
				errors = errors.ravel()
			elif(errors.ndim==3):
				# errors are not symmetric, i.e. separate downward & upward errors are given for each bar
				errors = numpy.moveaxis(errors, source=2, destination=0).reshape((2,Nbars))
				errors[0,:] *= -1 # convention for pyplot.bar and pyplot.barh is to define errors as absolute deviations from the bar's value, rather than signed deviations.
				errors = numpy.maximum(0,errors) # pyplot does not accept negative error values

	# plot bars
	values_flat = ([v for vv in values for v in vv] if inhomogeneous else values.flatten())
	if(vertical):
		pyplot.bar( bar_positions,
					height 	   	= values_flat,
					tick_label 	= bar_labels,
					yerr		= errors,
					color 	   	= bar_fill_colors,
					edgecolor  	= bar_edge_colors,
					alpha	   	= 0.85,
					width	   	= 0.75, 
					linewidth  	= 1,
					error_kw	= {'elinewidth':0.4},
					capsize		= (8 if as_pdf else 0.03*dpi)*plot_height/Nslots,
					zorder		= 3)
	else:
		pyplot.barh(bar_positions,
					width 	   	= values_flat,
					tick_label 	= bar_labels,
					xerr		= errors,
					color 	   	= bar_fill_colors,
					edgecolor  	= bar_edge_colors,
					alpha	   	= 0.85,
					height	   	= 0.75, 
					linewidth  	= 1,
					error_kw	= {'elinewidth':0.4},
					capsize		= (8 if as_pdf else 0.03*dpi)*plot_height/Nslots,
					zorder		= 3)
		ax.invert_yaxis()

	# plot reference line at zero
	if(vertical):
		# horizontal reference line
		pyplot.axhline(y=0, color="#000000", alpha=1, linewidth=1, zorder=3)
	else:
		# vertical reference line
		pyplot.axvline(x=0, color="#000000", alpha=1, linewidth=1, zorder=3)
		
	# limit to only positive values if needed
	if(only_positives):
		plot_max_value = 1.1*max(0,(0 if (value_ref is None) else value_ref),numpy.nanmax(values_flat))
		if(vertical):
			pyplot.ylim(0, plot_max_value)
		else:
			pyplot.xlim(0, plot_max_value)
	
	# axis labels, titles and axis ticks, frame color
	pyplot.title(plot_title+"\n",fontsize="medium", fontname="Helvetica")
	if(var_label!=""): 
		if(vertical):
			pyplot.xlabel(var_label, fontsize="medium", fontname="Helvetica")
		else:
			pyplot.ylabel(var_label, fontsize="medium", fontname="Helvetica", rotation=90, horizontalalignment="center")
	if(value_label!=""): 
		if(vertical):
			pyplot.ylabel(value_label, fontsize="medium", fontname="Helvetica", rotation=90, horizontalalignment="center")
		else:
			pyplot.xlabel(value_label, fontsize="medium", fontname="Helvetica")
	if(vertical): pyplot.xticks(rotation=45, ha="right")
	pyplot.setp(ax.spines.values(), color=frame_color)

	if(vertical):
		plot_axis_marks(marks=value_marks, axis="y", bbox_extra_artists=bbox_extra_artists)	
	else:
		plot_axis_marks(marks=value_marks, axis="x", bbox_extra_artists=bbox_extra_artists)
	
	if((parent:=os.path.dirname(plot_file))!=""): os.makedirs(parent, exist_ok=True)
	fig.savefig(plot_file, dpi=dpi, format=os.path.splitext(plot_file)[1][1:], bbox_inches='tight', bbox_extra_artists=bbox_extra_artists)
	pyplot.close(fig)





# convenience function for showing pairwise correlations as a heatmap, essentially a wrapper around plot_heatmap
def plot_correlations(	plot_file,						# string, path to the output plot file. The file extension determines the file format (e.g., "pdf" or "jpg").
						correlations,					# 2D numeric numpy array of size N1 x N2, listing the correlation values. The first axis enumerates rows, the second axis enumerates columns. Columns with greater index are plotted further to the right, rows with greater index are plotted further down.
						names,							# list or 1D numpy array of strings, listing the names of the first set of variables. These will be used as row labels (or column & row labels, if names2 is None).
						names2				= None,		# optional list or 1D numpy array of strings, listing the names of the second set of variables. If None, this is set to names2, and it is assumed that the correlations refer to the same sets of variables (which means N1=N2).
						axis_label			= "", 		# string specifying the label for the first (vertical) axis (or both axes if axis_label2==None).
						axis_label2			= None,		# optional string, e.g. 'treatment'. If None, this is set equal to axis_label.
						clabel				= "correlation", # label for the colorbar axis, e.g. 'activity'
						show_values_above	= 0.01,		# display correlations in each heatmap cell whenever their modulus is above this threshold. If <0, all correlations are shown. If set1==set2, then regardless of this threshold the diagonal values are not shown (since they should all be 1).
						show_grid			= "auto",	# either "auto", "vertical", "horizontal", "all" or "none"
						cluster				= "none",	# whether and how to reorder rows & columns via hierarchically clustering. Options are "none", "rows", "columns", "rows_and_columns_joint" and "rows_and_columns_separate".
						plot_title			= "",
						plot_width			= None,		# numeric, plot width in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of columns.
						plot_height			= None,		# numeric, plot height in inches. Typical values are 3-4. To convert to pixels (in the case of bitmap images), multiply this by DPI. If None, this is chosen automatically, scaling with the number of columns.
						dpi					= 300):
	# basic input checking
	if((names is not None) and (len(names)!=correlations.shape[0])): raise ValueError("Got %d names, but correlations table only has %d rows"%(len(names),correlations.shape[0]))
	if((names2 is not None) and (len(names2)!=correlations.shape[1])): raise ValueError("Got %d names2, but correlations table only has %d columns"%(len(names2),correlations.shape[1]))
	same_set = (names2 is None)
	if(same_set and (correlations.shape[0]!=correlations.shape[1])): raise ValueError("Assuming that correlations refer to the same sets of variables, but correlations table of size %s is not square"%(sprint_shape(correlations)))
	# determine which correlations to display in the heatmap cells
	show_values = (numpy.abs(correlations[:,:])>show_values_above)
	if(same_set): numpy.fill_diagonal(show_values, False)
	# plot
	plot_heatmap(table			= correlations,
				column_values	= (names if (names2 is None) else names2),
				row_values		= names,
				show_colorbar	= True,
				colormap		= DEFAULT_COLORMAP_FOR_CORRELATIONS,
				colormap_min	= -1,
				colormap_max	= +1,
				xlabel			= (axis_label if (axis_label2 is None) else axis_label2),
				ylabel			= axis_label,
				clabel			= clabel,
				show_values		= show_values,
				value_format	= "%.2g",	# format specifier for converting table values to strings, when displaying them inside cells. Only relevant if show_values!=False
				show_grid		= show_grid,
				cluster			= cluster,
				plot_title		= plot_title,
				plot_width		= plot_width,
				plot_height		= plot_height,
				plot_file		= plot_file,
				dpi				= dpi)



##############################################################
# improved dictionary, with more flexible syntax for assigning & querying members.
# Use e.g., as:
# mydict = IDict(x=10, y=20, city="Vancouver")
# z = mydict.x + mydict.y
# print(mydict.city)
# print(hasattr(mydict, "city")) # this will print True
# Since it is a subclass of the standard dict, one can also use all of classical dict syntax, for example:
# print(mydict.keys())
# print(mydict.values())
# print(mydict["city"])
# print(mydict.get("town", None))
# print("city" in mydict) # this will print True
# One can even pass a IDict as a dictionary of arguments to a function, for example:
# myfunction(**mydict)

class IDict(dict):
	__delattr__ = dict.__delitem__

	# add or modify member elements in-place, and return self
	# this replaces the standard dict.update() method, the only difference being that it returns self instead of None.
	def update(self,**kwargs):
		super().update(kwargs)
		return self

	# function for creating a shallow copy of the IDict instance and adding (or modifying) one or more member elements
	# use for example as:
	#	newdict = olddict.copy_and_update(nearest_town="Whistler")
	#	print(newdict.nearest_town)
	# Note that copying is not deep, i.e., the values stored in the parent & child may both still refer to the same objects.
	def copy_and_update(self,**kwargs):
		new = copy.copy(self)
		new.update(**kwargs)
		return new

	# enable retrieving a member value using the syntax <my_named_list>.<key>
	# standard dictionary attributes take precedence over stored keys, so you should avoid storing keys likely to be standard dict attributes
	def __getattr__(self, key):
# 		if((key=="__array_struct__") or (key=="__array_interface__") or (key=="__array__")):
# 			# explicitly tell numpy that these attributes do not exist, so that it knows the dict is not an array
# 			raise AttributeError
# 		try:
# 			return dict.__getattribute__(self, key)
# 		except:
# 			return self[key]
		if(key.startswith("__")): return dict.__getattribute__(self, key)
		try:
			return dict.__getattribute__(self, key)
		except:
			return self[key]


	# enable setting a member value using the syntax <my_named_list>.<key> = <value>
# 	def __setattr__(self, key, value):
# 		self[key] = value
	__setattr__ = dict.__setitem__




########################################################

# Simple context manager for timing a piece of code
# If task_name is provided during initialization, the total time elapsed is printed upon exit of the context manager.
#
# Use e.g. as:
#	with Timer("  Fitting"):
#		model.fit(...)
#
# or as:
#	with Timer() as tim:
#		model.fit(...)
#	print("Fitting took %g real sec, %g compute sec"%(tim.real_time,tim.process_time))
class Timer:
	def __init__(self, task_name=None):
		self.task_name = task_name		

	def __enter__(self):
		self.real_time  	= time.perf_counter()
		self.process_time 	= time.process_time()
		if(self.task_name is not None): print(self.task_name+"..")
		return self

	def __exit__(self, type, value, traceback):
		self.real_time  	= time.perf_counter() - self.real_time		# real time (seconds) elapsed since the timer was entered
		self.process_time 	= time.process_time() - self.process_time 	# the CPU time (seconds) taken by the current process
		if(self.task_name is not None): print("  %s took %g seconds"%(self.task_name,self.real_time))



##########################################################
# MARK File I/O


# get the right-most extension of a file
# if skip_gz==True, and the file name ends with ".gz", then the ".gz" part is ignored (for example "data.tsv.gz" yields "tsv").
def file_extension(filepath, skip_gz):
	if(skip_gz and filepath.lower().endswith(".gz")): filepath = filepath[:-3]
	extension = filepath.rsplit(".",1)[-1]
	return extension


# guess the column delimiter of a classical table file, based on the file extension
def infer_file_delimiter(filepath, default=" "):
	extension = file_extension(filepath, skip_gz=True).lower()
	return FILE_EXTENSION_TO_COLUMN_DELIMITER.get(extension, default)


# open a local file for reading/writing, supporting optionally gzip compression (file extension ".gz")
# mode can be 'wt', 'rt', 'wb', 'rb' or other standard modes
def open_file(filepath, mode):
	if(mode.startswith("w") or mode.startswith("a")):
		parent = os.path.dirname(filepath)
		if(parent!=""): os.makedirs(parent, exist_ok=True)
	if(filepath.lower().endswith(".gz")):
		return gzip.open(filepath,mode)
	else:
		return open(filepath,mode)





# save an object as a local pickled binary file
# This is essentially a wrapper around pickle.dump, but using the parallelized pigz tool for gzip-compression and ensuring existence of the parent directory
def save_object(obj, filepath, compression_level=6):
	parent = os.path.dirname(filepath)
	if(parent!=""): os.makedirs(parent, exist_ok=True)
	if(filepath.lower().endswith(".gz")):
		temp_filepath = filepath[:-3]
		with open(temp_filepath, 'wb') as fout:
			pickle.dump(obj, fout)
		os.system("pigz -%d -c '%s' > '%s'; rm -f '%s'"%(compression_level,temp_filepath,filepath,temp_filepath))
	else:
		with open(filepath, 'wb') as fout:
			pickle.dump(obj, fout)


# load an object from a local pickled binary file
# If graceful==False, and filepath is None or the input file is missing or the checksum is inconsistent, an error is raised.
# If graceful==True, and filepath is None or the input file is missing, then default_value is returned.
def load_object(filepath, 				# local or remote filepath. If a remote path, it must be in the form <username>@<hostname or IP>:<remote path>
				graceful		= True, 
				default_value	= None, 
				checksum		= None):
	if(filepath is None):
		if(graceful): return default_value	
		else: raise RuntimeError("Missing file path to load")
	elif(not os.path.exists(filepath)):
		if(graceful): return default_value
		else: raise RuntimeError("Missing pickle file '%s'"%(filepath))
	try:
		with open_file(filepath, 'rb') as fin:
			obj = pickle.load(file=fin)
	except Exception as e:
		if(graceful): return default_value
		else: raise RuntimeError("Failed to load pickle file '%s': %s"%(filepath,e))
	if((checksum is not None) and ((not hasattr(obj,"checksum")) or (obj.checksum!=checksum))):
		if(graceful): return default_value
		else: raise RuntimeError("Loaded object checksum differs from expectation")
	return obj


def check_output_file(file,force,verbose,verbose_prefix):
	if(os.path.exists(file)):
		if(force): 
			if(os.path.isdir(file)): shutil.rmtree(file)
			else: os.remove(file)
			if(verbose): print("%sNote: Replacing output file '%s'" % (verbose_prefix,file))
		else:
			print("%sERROR: Output file '%s' already exists\n%s       Cowardly refusing to continue.\n%s       Use --force to ignore this message" % (verbose_prefix,file,verbose_prefix,verbose_prefix))
			sys.exit(1)
	else:
		parent = os.path.dirname(file)
		if((parent!='') and (not os.path.exists(parent))): os.makedirs(parent, exist_ok=True)


def save_table(	filepath,			# path to output file. If it exists, it will be silently replaced.
				table,				# 2D numpy object or float or string array of size NR x NC
				row_names,			# list of size NR. Can be None (don't save row names)
				column_names, 		# can be None
				row_names_column, 	# column name under which to store row names. Can be None (don't save row names).
				save_column_names_in	= "first_data_line",	# either 'none' or 'last_comment_line' or 'first_data_line'
				delimiter				= None,					# column delimiter, e.g. "\t" or " " or ",". If None, this is determined automatically based on the file extension.
				comment_prefix			= "#",
				comments				= ""):	# optional string of comments to include at the top of the table
	check_output_file(file=filepath, force=True, verbose=False, verbose_prefix="")
	if(delimiter is None): delimiter = infer_file_delimiter(filepath)
	include_row_names = ((row_names is not None) and (row_names_column is not None))
	include_column_names = ((column_names is not None) and (save_column_names_in!="none"))
	if(include_row_names and (column_names is not None)): column_names = [row_names_column] + list(column_names) # add row-names column
	with open_file(filepath,'wt') as fout:
		# write any header comments
		if(comments!=""):
			fout.write(comment_prefix + " " + comments.replace("\n","\n"+comment_prefix+" ") + "\n")
		# write column names
		if(include_column_names):
			if(save_column_names_in=="last_comment_line"): fout.write(comment_prefix+" ")
			fout.write(delimiter.join(column_names)+"\n")
		# write table contents
		if((table.dtype.type is numpy.float_) or (table.dtype.type is numpy.cfloat)):
			for r in range(table.shape[0]):
				row_data = delimiter.join("{:.10g}".format(value) for value in table[r,:])
				fout.write((row_names[r]+delimiter if include_row_names else "") + row_data + "\n")
		else:
			for r in range(table.shape[0]):
				row_data = delimiter.join(table[r,:])
				fout.write((row_names[r]+delimiter if include_row_names else "") + row_data + "\n")



# Load a single scalar profile (along a 1D grid) from a CSV/TSV/XLSX file, returning a 1D numeric array
# The function assumes that grid points are in the first column and values are in a column named colname
# If grid_out is None, then the loaded grid points are returned as well. If grid_out is not None, the loaded values are first interpolated onto grid_out, and the latter is not returned.
def load_profile_from_file(filepath, colname, grid_out=None, check_nan=True):
	if(filepath.lower().endswith(".xlsx")):
		DF = pandas.read_excel(filepath, index_col=None, na_filter=False, comment="#")
	else:
		DF = pandas.read_csv(filepath, index_col=None, sep=infer_file_delimiter(filepath, default=""), na_filter=False, comment="#")
	if(colname not in DF): raise RuntimeError("Column '%s' not found in table file '%s'"%(colname,filepath))
	grid   = numpy_float_or_nan(DF.iloc[:,0])
	values = numpy_float_or_nan(DF[colname])
	if(grid_out is None):
		return grid, values
	else:
		return interpolate_time_series(x=grid_out, xp=grid, fp=values, axis=0, left=None, right=None, check_nan=check_nan)
		

