# Parsing and interpreting chemical reactions & microbial dynamics
#
# Stilianos Louca
# Copyright 2024
#
# LICENSE AGREEMENT
# - - - - - - - - -
# All rights reserved.
# Use and redistributions of this code is permitted for commercial and non-commercial purposes,
# under the following conditions:
#
#	* Redistributions must retain the above copyright notice, this list of 
#	  conditions and the following disclaimer in the code itself, as well 
#	  as in documentation and/or other materials provided with the code.
#	* Neither the name of the original author (Stilianos Louca), nor the names 
#	  of its contributors may be used to endorse or promote products derived 
#	  from this code without specific prior written permission.
#	* Proper attribution must be given to the original author, including a 
#     reference to the peer-reviewed publication through which the code was published.
#
# THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# - - - - - - - - -

import string
import sys, os
import numpy
import math
import pandas
import copy

# import other modules
from ReactionTransportModeling.general import *
from ReactionTransportModeling.solvers import *

# define some physical constants
Rgas 	= 8.3144621e-3 # gas constant (kJ/mol/K)
Faraday = 96.486 # Faraday constant, in kJ/V


# Class for defining a set of chemical reactions and their reactants/products.
# Optionally, some reactions may be classified as "biotic", which means that they are catalyzed by microbes whose cell densities follow simple growth dynamics in response to the reactions catalyzed.
# A ReactionNetwork is itself agnostic of any spatial and temporal dimension, i.e., all reactions and cell population dynamics take place in a well-mixed box following time-invariant kinetics.
# Instances of this class may be used to build ODE/PDE models of microbial systems, where some or all reactions are driven by microbes, for example a water column or a chemostat.
# Such microbial systems may include a spatial grid and take care of transport dynamics, buffering of some metabolites, external fluxes of metabolites etc.
#
# Abiotic reaction kinetics can be combinations of mass-action laws (i.e., first order in any given substrate) and Michaelis-Menten kinetics, including inhibitions:
#	rate = Vmax * ([C1] * .. * [Cn]) * ([X1] * .. * [Xm])/(([X1]+K1) * .. * ([Xm]+Km)) * (L1 * .. * Lk)/(([I1]+L1) * .. * ([Ik]+Lk))
#	where C1,..,Cn are substrates with mass-action contribution, X1,..,Xm are substrates with Michaelis-Menten-type contributions and I1,..,Ik are inhibitors, and K1,..,Km are half-saturation constants and L1,..,Lk are half-inhibition constants.
# Biotic reaction kinetics can be of a similar form, the only difference being that the rates are cell-specific, i.e., the kinetic law describes the reaction rate per cell rather than per volume.
# Some metabolites may have pre-specified (fixed) concentrations; such metabolites are henceforth called "fixed", while all other metabolites (i.e., whose concentrations are dynamic variables) are called "dynamic".
#
# Measurement units:
# 	The units used for metabolite concentrations can in principle be arbitrary, and are essentially dictated by the units used for the data provided in the input tables upon initialization.
# 	The units for cell densities (in the case of biotic reactions) should be equal to the units of cell yield factors ([yield]) multiplied by the units of metabolite concentrations.
# 	Below, the documentation of this class assumes metabolite concentrations are in mol/L, cell densities are in cells/L and yield factors are in cells/mol.
#   Attention: If yield factors are missing (nan) for some biotic reactions, the DeltaG of a reaction is used to predict yield factors, in which case the canonical concentration units (mol/L) & density units (cells/L) specified in the documentation must be followed.

class ReactionNetwork():
	def __init__(self, 
				Mtable,								# either a dataframe specifying metabolites (rows indexed by metabolite name, with columns "DeltaGf0", "description" & "epsilon") or a file path to an input table of similar format.
				Rtable,								# either a dataframe specifying reactions (rows indexed by reaction name, with columns "DeltaG0", "equation", "description", "biotic", "yield", "death", "cell_mass", "limiters", "biotic_affinity", "abiotic_affinity", "Nelectrons" and "membrane_potential") or a file path to an input table of similar format.
				only_metabolites		 = None,	# optional, either a string specifying a comma-separated list of metabolite names (or wildcards) to keep, or a list/set of such names/wildcards. If None or has len==0, this filter is ignored.
				omit_metabolites		 = None,	# optional, either a string specifying a comma-separated list of metabolite names (or wildcards) to omit, or a list/set of such names/wildcards. If None or has len==0, this filter is ignored.
				only_reactions			 = None,	# optional, either a string specifying a comma-separated list of reaction names (or wildcards or regular expressions) to keep, or a list/set of such names/wildcards. If None or has len==0, this filter is ignored.
				omit_reactions			 = None,	# optional, either a string specifying a comma-separated list of reaction names (or wildcards or regular expressions) to omit, or a list/set of such names/wildcards. If None or has len==0, this filter is ignored.
				omit_trivial_metabolites = True,	# if True, then any trivial metabolites (i.e., not involved in any reaction, after filtering) will be omitted as well
				omit_trivial_reactions	 = True):	# if True, then any trivial reactions (i.e., with no metabolites, after filtering) will be omitted as well					
		# load table files as dataframes if needed
		if(isinstance(Mtable,str)): Mtable = pandas.read_csv(Mtable, index_col="name", sep=infer_file_delimiter(Mtable, default=""), na_filter=False, comment="#")
		if(isinstance(Rtable,str)): Rtable = pandas.read_csv(Rtable, index_col="name", sep=infer_file_delimiter(Rtable, default=""), na_filter=False, comment="#")
		
		# sort metabolites & reactions alphabetically
		Mtable.sort_index(axis=0, inplace=True, key=lambda col: col.str.lower())
		Rtable.sort_index(axis=0, inplace=True, key=lambda col: col.str.lower())
	
		# construct stoichiometric matrix
		stoich = numpy.zeros((Mtable.shape[0], Rtable.shape[0]))
		for r,reaction,equation in zip(range(Rtable.shape[0]),Rtable.index,Rtable["equation"]):
			metabolites, coefficients = self.disect_reaction_equation(equation)
			missings = [metabolite for metabolite in metabolites if (metabolite not in Mtable.index)]
			if(len(missings)>0): raise ValueError("Invalid metabolites (e.g., '%s') specified in the equation of reaction '%s'"%(missings[0],reaction))
			stoich[[Mtable.index.get_loc(metabolite) for metabolite in metabolites], r] = coefficients	
		
		# filter metabolites & reactions if needed
		keep_metabolites = numpy.asarray(filter_names(names=Mtable.index, only_names=only_metabolites, omit_names=omit_metabolites, pattern_type="wildcard"), dtype=int)
		keep_reactions   = numpy.asarray(filter_names(names=Rtable.index, only_names=only_reactions, omit_names=omit_reactions, pattern_type="wildcard"), dtype=int)
		stoich	 		 = stoich[numpy.ix_(keep_metabolites,keep_reactions)]
		if(omit_trivial_metabolites):
			nontrivials 	 = numpy.where(numpy.any(stoich!=0, axis=1))[0]
			stoich	 		 = stoich[nontrivials,:]
			keep_metabolites = keep_metabolites[nontrivials]
		if(omit_trivial_reactions):
			nontrivials 	= numpy.where(numpy.any(stoich!=0, axis=0))[0]
			stoich			= stoich[:,nontrivials]
			keep_reactions	= keep_reactions[nontrivials]
		Mtable = Mtable.iloc[keep_metabolites,:]
		Rtable = Rtable.iloc[keep_reactions,:]
		
		self.NM, self.NR 	= Mtable.shape[0], Rtable.shape[0]
		self.stoich 		= stoich
		self.metabolites 	= Mtable.index.to_numpy()
		self.MDeltaGf0 		= Mtable["DeltaGf0"].to_numpy()
		self.Mdescriptions	= Mtable["description"].to_numpy()
		self.Mepsilon		= Mtable["epsilon"].to_numpy()
		self.reactions   	= Rtable.index.to_numpy()
		self.RDeltaG0		= Rtable["DeltaG0"].to_numpy()
		self.Requations		= Rtable["equation"].to_numpy()
		
		# determine any missing standard Gibbs free energies of reactions based on the formation energies of their constituents
		missing_DeltaG0s = numpy.where(~numpy.isfinite(self.RDeltaG0))[0]
		if(len(missing_DeltaG0s)>0):
			relevant_metabolites = numpy.any(stoich[:,missing_DeltaG0s]!=0,axis=1)
			self.RDeltaG0[missing_DeltaG0s] = self.MDeltaGf0[relevant_metabolites] @ stoich[relevant_metabolites,missing_DeltaG0s]
		
		# determine metabolites used & produced by each reaction
		# Each Rreactants[r] will be a list of integers from 0 to NM-1, specifying which of the metabolites are reactants for the r'th reaction
		# Each Rproducts[r] will be a list of integers from 0 to NM-1, specifying which of the metabolites are products of the r'th reaction
		self.Rreactants = [numpy.where(stoich[:,r]<0)[0] for r in range(self.NR)]
		self.Rproducts  = [numpy.where(stoich[:,r]>0)[0] for r in range(self.NR)]
			
		# assign reaction kinetics, assuming mass action laws
		self.set_kinetics(Rtable)
		


	@property
	def shape(self):
		return (self.stoich.shape)


	def __repr__(self):
		max_metabolite_len = max(len(metabolite) for metabolite in self.metabolites)
		max_reaction_len   = max(len(reaction) for reaction in self.reactions)
		representation = "Reaction pool with %d metabolites x %d reactions"%(self.NM,self.NR)
		if(self.NM>0):
			representation += "\nMetabolites:" + "".join("\n  %s: %s"%(pad_string_right(metabolite,min_len=max_metabolite_len+1),description) for metabolite,description in zip(self.metabolites,self.Mdescriptions))
		if(len(self.Rabiotic)>0):
			representation += "\nAbiotic reactions (%d):"%(len(self.Rabiotic))\
							  + "".join("\n  %s: %s"%(pad_string_right(self.reactions[r],min_len=max_reaction_len+1),self.Requations[r]) for r in self.Rabiotic)		
		if(self.NRB>0):
			representation += "\nBiotic reactions (%d):"%(self.NRB)\
							  + "".join("\n  %s: %s"%(pad_string_right(self.reactions[r],min_len=max_reaction_len+1),self.Requations[r]) for r in self.Rbiotic)		
		if(numpy.any(numpy.isfinite(self.RBmembrane_potentials))):
			representation += "\nMembrane potentials & electrons transferred for %d biotic reactions:"%(self.NRB)\
							  + "".join("\n  %s: %g V, %g electrons"%(pad_string_right(self.reactions[r],min_len=max_reaction_len+1),self.RBmembrane_potentials[rb],self.RNelectrons[r]) for rb,r in enumerate(self.Rbiotic))		
		return representation			


	@property
	def biotic_reactions(self):
		return self.reactions[self.Rbiotic]


	# Set or modify the kinetics of each reaction in the form of either mass-action laws or Monod kinetics
	# Each reaction is declared as either biotic or abiotic. Some parameters (such as cell yield factors) and dynamic variables (such as cell densities) are only relevant for biotic reactions.
	# For abiotic reactions, kinetics specify volume-specific rates (e.g., mol per volume per time), while for biotic reactions kinetics specify cell-specific rates (e.g., mol per cell per time).
	# The affinity of an abiotic reaction is measured in day^{-1} * molar^{1-k}, where k is the number of limiting metabolites for the reaction.
	#	For example, for a reaction limited by two metabolites, the affinity is measured in 1/(day*molar).
	#   The affinity is defined as the rate density of a reaction (in molar per day) divided by the concentrations of all limiting metabolites, when the latter are rare.
	#	For example, if a reaction proceeds at 10 μM/day and is limited by two rare metabolites, each with concentration of 1 μM, then its affinity is 10/(μM*day) = 1e7 day^{-1} molar^{-1}
	#	The abiotic affinity is only elevant for reactions declared as abiotic.
	# The affinity of a biotic reaction is measured in L * day^{-1} * molar^{1-k} * cell^{-1}, where k is the number of limiting metabolites for the reaction.
	#	For example, for a reaction limited by two metabolites, the affinity is measured in L/(day*molar*cell).
	#   The biotic affinity is defined as the rate density of a reaction (in molar per day) divided by the concentrations of all limiting metabolites and the concentrations of cells performing the reaction, when the substrates are all rare.
	#	For example, if 10^6 cells/L perform a reaction at 10 μM/day that is limited by two rare metabolites, each with concentration of 1 μM, then the affinity is 10 L/(day*molar*cell).
	#	The biotic affinity is only elevant for reactions declared as biotic.
	# If half-saturation constants are provided and finite, then kinetics are of Monod type (in the respective metabolites), whereas if half saturation constants are missing or non-finite mass-action laws are assumed.
	#   Specifically, Monod kinetics are assumed to be of the form:
	#		rate = Vmax * ((C1 * .. * Cn) / (([C1]+K1) * .. * ([Cn]+Kn))) * ((L1 * .. * Lm) / (([I1]+L1) * .. * ([Im]+Lm)))
	#	where C1 ... Cn are limiting substrates with half-saturation constants K1 .. Kn, and I1 .. Im are inhibiting substrates with half-inhibition constants L1 .. Lm.
	#	This is equivalent to:
	#		rate = affinity * (C1 * .. * Cn) * (K1 * .. * Kn) * (L1 * .. * Lm) /(([C1]+K1) * .. * ([Cn]+Kn) * ([I1]+L1) * .. * ([Im]+Lm))
	#	Thus, substrates with finite half-saturation constants have a similar mathematical contribution as inhibors, the only difference being that for the latter we don't include their concentrations in the numerator.
	# Ktable may be one of the following:
	#	A dataframe specifying the mass-action kinetics of reactions, with rows indexed by reaction names, 
	#		and columns including "biotic" (bool), "yield" (float, cells per kJ, optional), "death" (float, 1/day, optional), "cell_mass" (float, gram, optional), "limiters" (string, comma-separated list of metabolites, optionally specifying half-saturation concentrations for some of them), "abiotic_affinity" & "biotic_affinity" (float, see above for units), and "Nelectrons" (float, optional) and "membrane_potential" (float, Volt, optional).
	#		This dataframe may contain superfluous reactions, i.e. not in the model, which will be silently ignored.
	#	A file path to an input table listing kinetics in the same format, one row per reaction.
	#	A dictionary with keys "biotic", "yield" (optional), "death" (optional), "cell_mass" (optional), "abiotic_affinity", "biotic_affinity", "limiters", "Nelectrons" (optional) and "membrane_potential" (optional), and with each value either being a scalar or a 1D array of size self.NR synchronized with self.reactions[]
	def set_kinetics(self, Ktable):
		if(isinstance(Ktable,dict)):
			self.Rbiotic  				= numpy.where(array_if_scalar(value=Ktable["biotic"], shape=(self.NR,), dtype=bool))[0]
			self.RByields 				= array_if_scalar(value=Ktable.get("yield",nan), shape=(self.NR,), dtype=float)[self.Rbiotic]
			self.RBdeaths 				= array_if_scalar(value=Ktable.get("death",0), shape=(self.NR,), dtype=float)[self.Rbiotic]
			self.RBcell_masses			= array_if_scalar(value=Ktable.get("cell_mass",nan), shape=(self.NR,), dtype=float)[self.Rbiotic]
			self.RBmembrane_potentials 	= array_if_scalar(value=Ktable.get("membrane_potential",-inf), shape=(self.NR,), dtype=float)[self.Rbiotic]
			self.RNelectrons			= array_if_scalar(value=Ktable.get("Nelectrons",0), shape=(self.NR,), dtype=float)
			abiotic_affinities			= array_if_scalar(value=Ktable["abiotic_affinity"], shape=(self.NR,), dtype=float)
			biotic_affinities			= array_if_scalar(value=Ktable["biotic_affinity"], shape=(self.NR,), dtype=float)
			limiter_specs				= array_if_scalar(Ktable["limiters"], shape=(self.NR,), dtype=str)
			inhibiter_specs				= (array_if_scalar(Ktable["inhibiters"], shape=(self.NR,), dtype=str) if ("inhibiters" in Ktable) else [""]*self.NR)
		else:
			if(isinstance(Ktable,str)): Ktable = pandas.read_csv(Ktable, index_col="name", sep=infer_file_delimiter(Ktable, default=""), na_filter=False, comment="#")
			Ktable						= Ktable.loc[self.reactions,:] # filter & synchronize Ktable rows with self.reactions
			self.Rbiotic				= numpy.where(Ktable.loc[:,"biotic"].to_numpy(dtype=bool))[0] # 1D array listing the indices of biotic reactions (indices w.r.t. self.reactions)
			self.RByields				= (numpy.full(len(self.Rbiotic),nan) if ("yield" not in Ktable) else numpy_float_or_nan(Ktable.loc[self.reactions[self.Rbiotic],"yield"])) # 1D array synchronized with self.Rbiotic[]. Cell yield factors (cells produced per mol reaction).
			self.RBdeaths				= (numpy.full(len(self.Rbiotic),0) if ("death" not in Ktable) else numpy_float_or_nan(Ktable.loc[self.reactions[self.Rbiotic],"death"])) # 1D array synchronized with self.Rbiotic[]. Per-capita death rates of cells, in 1/day.
			self.RBcell_masses			= (numpy.full(len(self.Rbiotic),nan) if ("cell_mass" not in Ktable) else numpy_float_or_nan(Ktable.loc[self.reactions[self.Rbiotic],"cell_mass"])) # 1D array synchronized with self.Rbiotic[]. Dry cell masses, in gram.
			self.RBmembrane_potentials 	= (numpy.full(len(self.Rbiotic),-inf) if ("membrane_potential" not in Ktable) else numpy_float_or_nan(Ktable.loc[self.reactions[self.Rbiotic],"membrane_potential"])) # 1D array synchronized with self.Rbiotic[]. Cell membrane potential, in Volt.
			self.RNelectrons			= (numpy.full(self.NR,0) if ("Nelectrons" not in Ktable) else numpy_float_or_nan(Ktable.loc[:,"Nelectrons"])) # 1D array synchronized with self.reactions. Mole electrons transferred per mol reaction.
			abiotic_affinities 			= numpy_float_or_nan(Ktable.loc[:,"abiotic_affinity"])
			biotic_affinities 			= numpy_float_or_nan(Ktable.loc[:,"biotic_affinity"])
			limiter_specs				= Ktable.loc[:,"limiters"]
			inhibiter_specs				= (Ktable.loc[:,"inhibiters"] if ("inhibiters" in Ktable) else [""]*self.NR)
		self.Rabiotic 	 = integer_complement(N=self.NR, subset=self.Rbiotic)
		self.Raffinities = array_insert(abiotic_affinities, indices=self.Rbiotic, values=biotic_affinities[self.Rbiotic], inplace=False)
		# determine limiting substrates & half-saturation constants (if available) for each reaction
		# mathematically, we can treat the saturation effect of substrates with finite Khalf similarly to the inhibition effects of inhibiters
		metabolite2index = {metabolite:m for m,metabolite in enumerate(self.metabolites)}
		self.Rlimiters   = [[] for r in range(self.NR)] # list of length NR, synchronized with self.reactions[]. Each element will be a list of integers (indices in self.metabolites), specifying which metabolites are limiting a given reaction (regardless of whether their Khalf is finite or not).
		self.Rsaturaters = [[] for r in range(self.NR)] # Metabolites with a saturation or inhibition effect for each reaction. List of length NR, synchronized with self.reactions[]. Each element will be a list of integers (indices in self.metabolites), specifying the inhibiters as well as saturated metabolites (i.e., with finite Khalf) that limit a given reaction
		self.RKhalfs 	 = [[] for r in range(self.NR)] # half-saturation or half-inhibition constants for each reaction. List of length NR, synchronized with self.reactions[]. Each r-th element will be a list of floats, synchronized with Rinhibiters[r], specifying the half-saturation or half-inhibition constants for the r-th reaction for the metabolites listed in self.Rinhibiters[r].
		for r,equation,limiter_spec,inhibiter_spec in zip(range(self.NR),self.Requations,limiter_specs,inhibiter_specs):
			reaction_metabolites, dummy = self.disect_reaction_equation(equation)
			# parse limiter spec
			if(limiter_spec.strip()!=""):
				Khalfs	 = [(float_or_nan(limiter.split(":",1)[1].strip()) if (":" in limiter) else nan) for limiter in limiter_spec.split(",")] # extract half-saturation constants from limiter list, if available
				limiters = [limiter.split(":",1)[0].strip() for limiter in limiter_spec.split(",")] # extract metabolite names from limiter list
				missings = [metabolite for metabolite in limiters if (metabolite not in reaction_metabolites)]
				if(len(missings)>0): raise ValueError("Invalid limiting metabolites (e.g., '%s') specified for reaction '%s': Not found in the reaction's equation"%(missings[0],self.reactions[r]))
				self.Rlimiters[r]   = [metabolite2index[limiter] for limiter in limiters if (limiter in metabolite2index)]
				self.Rsaturaters[r] = [metabolite2index[limiter] for limiter,Khalf in zip(limiters,Khalfs) if ((limiter in metabolite2index) and numpy.isfinite(Khalf))]
				self.RKhalfs[r] 	= [Khalf for limiter,Khalf in zip(limiters,Khalfs) if ((limiter in metabolite2index) and numpy.isfinite(Khalf))]
			# parse inhibiter spec
			if(inhibiter_spec.strip()!=""):
				Khalfs	 			= [float_or_nan(inhibiter.split(":",1)[1].strip()) for inhibiter in inhibiter_spec.split(",")] # extract half-inhibition constants from inhibiter list. We always expect a half inhibition constant for inhibiters.
				inhibiters 			= [inhibiter.split(":",1)[0].strip() for inhibiter in inhibiter_spec.split(",")] # extract metabolite names from inhibiter list
				self.Rsaturaters[r] += [metabolite2index[inhibiter] for inhibiter in inhibiters if (inhibiter in metabolite2index)]
				self.RKhalfs[r] 	+= Khalfs
		self.NRB = len(self.Rbiotic)


	# disect a half-reaction equation into its parts: metabolites & coefficients
	# The input string should be of the form:
	#	coefficient1 * substance1 + coefficient2 * substance2 + ...
	# Optionally coefficients may be omitted if they are equal to 1.
	@classmethod
	def disect_reaction_side(cls,side):
		parts 		 = side.split("+")
		metabolites	 = [part.split("*")[-1].strip() for part in parts]
		coefficients = [(float(part.split("*")[0]) if ("*" in part) else 1) for part in parts]
		return metabolites, coefficients
	
	
	# disect a reaction equation into its parts: reactants & products, coefficients
	# The input string should be of the form:
	#	coefficientA1 * reactant1 + coefficientA2 * reactant2 + ... = coefficientB1 * product1 + coefficientB2 * product2 + ...
	# Optionally coefficients may be omitted if they are equal to 1.
	# By convention the coefficients of reactants (left hand side of the equation) are negated.
	@classmethod
	def disect_reaction_equation(cls,equation):
		sides = equation.split("=")
		if(len(sides)!=2): raise ValueError("Could not parse reaction '%s': Expected exactly one '=' sign, instead got %d"%(reaction,len(sides)-1))
		reactants, reactant_coefficients = cls.disect_reaction_side(sides[0])
		products, product_coefficients   = cls.disect_reaction_side(sides[1])
		metabolites	 = reactants + products
		coefficients = [-k for k in reactant_coefficients] + product_coefficients
		return metabolites, coefficients
	


	# convert Gibbs free energy of a metabolic reaction (kJ per mole reaction) to the expected biomass yield (gram dry weight)
	# Reference: Roden and Jin (2011) Thermodynamics of Microbial Growth Coupled to Metabolism of Glucose, Ethanol, Short-Chain Organic Acids, and Hydrogen. Applied and Environmental Microbiology. 77:1907.
	@classmethod
	def DeltaG2biomass(cls,DeltaG):
		return numpy.maximum(0, 2.08 - 0.0211*DeltaG)


	# Function for computing thes rate of change (ROC) of metabolite concentrations and cell densities (corresponding to biotic reactions) based on their current values
	# This function can be used to build a "source" term in a PDE model for the metabolite & cell concentrations of aquatic chemical reaction or microbial systems.
	# The input metabolite concentrations & cell densities can have an arbitrary number of dimensions, as long as the last axis' size matches the number of metabolites and biotic reactions, respectively.
	# 	The first n axes will typically represent a spatial or spatiotemporal "grid", at each point of which the rates of change are independently requested.
	#	Non-spatial systems would correspond to n=0.
	# Cell population dynamics and reaction kinetics are governed by previously specified "kinetics" (see e.g. self.set_kinetics).
	def rates_of_change(self,
						concentrations,				# (n+1)-dimensional array of shape N1 x .. x Nn x NM, listing concentrations of metabolites
						cell_densities	= None,		# (n+1)-dimensional array of shape N1 x .. x Nn x NRB, listing cell densities for biotic reactions, in the order specified by self.Rbiotic. Only needed if NRB>0.
						temperature		= 25,		# temperature in degrees Celsius. Either a scalar, or an n-dimensional array broadcastable with the first n axes of concentrations[], i.e. of size [1 | N0] x ... x [1 | N_{n-1}]. Only used to compute Gibbs free energies, for biotic reactions lacking an explicitly specified yield factor.
						include_details = False,	# whether to include details (secondary results), such as gross production & consumption rates for metabolites or per-capita rates for cells in the returned results. If these are not needed, keep this at False for computational efficiency.
						Mfocal			= None):	# optional integer sequence (or slice) with values in 0,..,NM-1, specifying metabolites to which to restrict the returned results. For example, these may be a subset of "dynamic" metabolites in the context of a BioreactorColumn.
		# basic input checking
		if(self.NM!=concentrations.shape[-1]): raise ValueError("Invalid shape of input metabolite concentrations (%s): Last axis should have size %d (number of metabolites)"%(sprint_shape(concentrations),self.NM))
		if((cell_densities is None) and (self.NRB>0)): raise ValueError("Missing cell_densities, needed since the number of biotic reactions is non-zero")
		elif((cell_densities is not None) and (cell_densities.size>0) and (self.NRB==0)): raise ValueError("Unexpectedly got cell_densities, but the number of biotic reactions is zero")
		if(cell_densities is None): cell_densities = numpy.empty((1,)*(concentrations.ndim-1)+(0,)) # create an empty densities array for easier handling below
		if(self.NRB!=cell_densities.shape[-1]): raise ValueError("Invalid shape of input cell_densities (%s): Last axis should have size %d (number of biotic reactions)"%(sprint_shape(cell_densities),self.NRB))
		if(concentrations.shape[:-1]!=cell_densities.shape[:-1]): raise ValueError("Metabolite concentrations and cell_densities must have the same shape except for the last axis. Instead, got shapes %s and %s"%(sprint_shape(concentrations),sprint_shape(cell_densities)))
		cell_densities	= numpy.maximum(cell_densities,0)
		concentrations 	= numpy.maximum(concentrations,0)
		grid_shape 		= concentrations.shape[:-1]
		grid_ndim		= concentrations.ndim-1
		if(not numpy.isscalar(temperature)): 
			if(temperature.shape!=grid_shape): raise ValueError("Temperature has invalid shape %s: Expected either a scalar or an array of shape %s"%(sprint_shape(temperature),sprint_vector(grid_shape, separator=" x ")))
			temperature = numpy.expand_dims(temperature,axis=-1) # make temperature broadcastable with concentrations[]
		if(Mfocal is None): Mfocal = slice(None)
		temperature = temperature + 273.15 # convert to Kelvin

		# compute abiotic reaction rates
		reaction_rates	= numpy.empty(grid_shape+(self.NR,))
		for r in self.Rabiotic:
			reaction_rates[...,r] = self.Raffinities[r]*numpy.prod(concentrations[...,self.Rlimiters[r]],axis=-1)
			if(len(self.Rsaturaters[r])>0): reaction_rates[...,r] *= numpy.prod(self.RKhalfs[r]) / numpy.prod(concentrations[...,self.Rsaturaters[r]]+numpy.expand_dims(self.RKhalfs[r],axis=tuple(range(grid_ndim))),axis=-1)

		# prepare for computing DeltaGs of biotic reactions, for example if some of them lack cell-yield factors
		RB_need_DelgaG = (~numpy.isfinite(self.RByields)) | (numpy.isfinite(self.RBmembrane_potentials) & (self.RNelectrons[self.Rbiotic]!=0))
		if(numpy.any(RB_need_DelgaG)):
			log_concentrations   = log_nansafe(numpy.maximum(numpy.expand_dims(self.Mepsilon,axis=tuple(range(grid_ndim))),concentrations))
			cell_masses		 	 = numpy.expand_dims(self.RBcell_masses, axis=tuple(range(grid_ndim))) # add dummy axes to make cell_masses broadcastable along the grid
		
		# compute reaction rates and cell density ROC's for biotic reactions
		cell_net_productions = numpy.empty_like(cell_densities)
		if(include_details): 
			cell_gross_productions = numpy.empty_like(cell_densities)
			cell_gross_deaths		= numpy.empty_like(cell_densities)
		for rb,r in enumerate(self.Rbiotic):
			# compute Gibbs free energies for this reaction if needed
			if(RB_need_DelgaG[rb]):
				DeltaG = self.RDeltaG0[r] + Rgas * temperature * numpy.tensordot(log_concentrations, self.stoich[:,r], axes=(-1,0))
				DeltaG[numpy.any(concentrations[...,self.Rreactants[r]]<=0,axis=-1)] = inf # if any of the reactants of this reaction have zero concentration, set DeltaG to inf			
			# compute rates for this biotic reaction
			rates_per_cell = self.Raffinities[r]*numpy.prod(concentrations[...,self.Rlimiters[r]],axis=-1)
			if(len(self.Rsaturaters[r])>0): rates_per_cell *= numpy.prod(self.RKhalfs[r]) / numpy.prod(concentrations[...,self.Rsaturaters[r]]+numpy.expand_dims(self.RKhalfs[r],axis=tuple(range(grid_ndim))),axis=-1)
			if(numpy.isfinite(self.RBmembrane_potentials[rb]) and (self.RNelectrons[r]!=0)):
				# incorporate potential factor
				rates_per_cell /= numpy.exp(((DeltaG/self.RNelectrons[r]) + Faraday*self.RBmembrane_potentials[rb])/(Rgas*temperature))+1
			reaction_rates[...,r] = cell_densities[...,rb]*rates_per_cell
			# compute rate of change of cell densities for this biotic reaction, based on the computed reaction rates
			if(numpy.isfinite(self.RByields[rb])):
				Y = self.RByields[rb]
			else:
				Y = self.DeltaG2biomass(DeltaG)/cell_masses[...,rb]
			cell_net_productions[...,rb] = (Y * rates_per_cell - self.RBdeaths[rb])*cell_densities[...,rb]
			if(include_details): 
				cell_gross_productions[...,rb] = Y * rates_per_cell
				cell_gross_deaths[...,rb] = self.RBdeaths[rb]*cell_densities[...,rb]
		
		# compute rates of change of metabolite concentrations based on abiotic & biotic reaction rates
		metabolite_net_productions = numpy.tensordot(reaction_rates, self.stoich[Mfocal,:], axes=(-1,1))

		results = IDict(metabolite_net_productions	= metabolite_net_productions,
						cell_net_productions	 	= cell_net_productions, 
						reaction_rates	 			= reaction_rates)

		# also include gross metabolite production and consumption rates, as well as gross cell production rates and death rates, if needed
		if(include_details):
			Mfluxes_split = numpy.expand_dims(self.stoich[Mfocal,:],axis=tuple(range(grid_ndim))) * numpy.expand_dims(reaction_rates, axis=grid_ndim) # ROCs of metabolite concentrations split by reaction (in axis -1), separately for each metabolite (axis -2)
			results.metabolite_gross_productions  		= numpy.sum(numpy.maximum(0,Mfluxes_split),axis=-1)
			results.metabolite_gross_consumptions 		= numpy.sum(numpy.maximum(0,-Mfluxes_split),axis=-1)
			results.cell_gross_productions		  		= cell_gross_productions
			results.cell_gross_deaths			  		= cell_gross_deaths
			results.cell_per_capita_net_productions		= divide_nansafe(cell_net_productions,cell_densities)
			results.cell_per_capita_gross_productions	= divide_nansafe(cell_gross_productions,cell_densities)
			results.cell_per_capita_gross_deaths		= divide_nansafe(cell_gross_deaths,cell_densities)

		return results




###########################################################################

# Class for defining a 1D chemical reaction-advection-diffusion system, for example along a water column.
# Reactions may be abiotic or be catalyzed by microbes (one species per reaction), whose growth depends on their catalized reaction rates.
# Note that this class only supports time-invariant dynamics, i.e. there is no explicit time dependence allowed in any of the model parameters or boundary conditions.
class BioreactorColumn:
	def __init__(self,
				chemistry, 		# object of class ReactionNetwork, defining all metabolites and all abiotic & biotic reactions and their kinetics.
				grid,			# 1D numeric array of size NG, listing spatial grid coordinates in ascending order. Spatial grid used for all simulations. A finer grid improves accuracy, at the cost of more computation. For example, these may be depths along a sediment column.
				BCleft,			# specification of boundary conditions for dynamic metabolites and cell catalyzing biotic reactions, at the left boundary. Either a single 2-tuple of format (<boundary condition type>,<boundary condition value>), or a dictionary whose keys are metabolite/reaction names and whose values are 2-tuples. Superfluous keys are silently ignored, but missing necessary keys will throw an error.
				BCright,		# similar to BCleft, but for the right boundary.
				diffusivity,	# either a scalar float (same diffusivity for all metabolites & cells at all grid points), or a 1D numeric array of size NG specifying diffusivities varying along the grid (but equal for all metabolites & cells), or a dictionary (str-->float) that maps metabolite names and reaction names to their respective scalar diffusivities or to 1D arrays of size NG, or a string (file path) pointing to an input CSV/TSV/XLSX file listing locations/depths in the 1st column and diffusivities in a column named "diffusivity". In the case of a dictionary, superfluous keys are silently ignored, but missing necessary keys will throw an error.
				advection			= 0,	# advection speed. Allowed formats are similar to diffusivity.
				concentrations		= None,	# optional fixed concentrations, i.e., specified externally and that do not change due to internal reactions. Must be a dictionary, whose keys are metabolites and whose values are either scalar floats (same concentration on the entire grid) or 1D numeric arrays of size NG (different concentration at each grid point) or string specifying the path to an input CSV/TSV/XLSX file with locations/depths in the first column and concentrations in a column named as the metabolite. For example, this may be used to specify a fixed concentration of H+ ions (determined by pH) or a fixed concentration profile for particulate organic matter in a water column. Any keys (metabolite names) not in chemistry will be silently ignored.
				temperature			= 25,	# temperature in degrees Celsius. Either a scalar float (same temperature at all grid points), or a 1D numeric array of size NG specifying a varying temperature across grid points, or a string specifying the file path to a TSV/CSV/XLSX table with locations/depths in the first column and temperatures in a column named "temperature".
				vertical			= True,	# if True, the bioreactor column is assumed to be vertically oriented, i.e., the spatial axis points downwards. If False, the spatial axis points rightwards. This only affects the visualizations, not the computations.
				name				= "Bioreactor column", # optional name/ID for this object, for example "Saanich Inlet sediments"
				distance_unit		= "m",
				time_unit			= "d",
				concentration_unit	= "M",
				density_unit		= "cells/L",
				rtol				= 1e-4,	# relative error tolerance during simulation, for the differential equation solvers. Note that the absolute error is controlled via density_epsilon and chemistry.Mepsilons.
				density_epsilon		= 10,
				Mfocals				= None, # optional specification of focal metabolites, e.g., for plotting. Either a sequence of strings (names or wildcards) or a single string encoding a comma-separated list of names/wildcards. This option only affects visualizations, not the actual computations.
				Rfocals				= None,	# optional specification of focal reactions, e.g., for plotting. Either a sequence of strings (names or wildcards) or a single string encoding a comma-separated list of names/wildcards. This option only affects visualizations, not the actual computations.
				plot_sizeZ			= 2,	# plot size (inches) for spatial axis in all profile & contour plots.
				plot_sizeV			= 1.2,	# plot size (inches) for value axis in all profile plots.
				plot_sizeT			= 3):	# plot size (inches) for time axis in all contour plots.
		self.chemistry 	= chemistry
		self.grid 		= numpy.asarray(grid, dtype=float)
		self.NG 		= len(self.grid)
				
		# parse fixed metabolite concentrations
		# Mfixed[] will be a 1D integer array listing which metabolites have fixed concentrations/concentration profiles
		# MFconcentrations[] will be a 2D array of size [1 | NG] x NMF, with columns synchronized with self.Mfixed, specifying the concentrations of fixed metabolites
		if(concentrations is None):
			self.Mfixed   = numpy.empty(0, dtype=int)
			self.Mdynamic = numpy.arange(chemistry.NM, dtype=int)
			self.MFconcentrations = numpy.empty((1,0), dtype=int)
		else:
			for metabolite,concentration in concentrations.items():
				if((not numpy.isscalar(concentration)) and (concentration.shape!=(self.NG,))): raise ValueError("Fixed concentration specified for '%s' has invalid shape %s: Expected a scalar or a 1D array of size %d"%(metabolite,sprint_shape(concentration),self.NG))
			self.Mfixed    = numpy.asarray([m for m,metabolite in enumerate(chemistry.metabolites) if (metabolite in concentrations)], dtype=int)
			self.Mdynamic  = integer_complement(chemistry.NM, self.Mfixed)
			concentrations = [concentrations[chemistry.metabolites[m]] for m in self.Mfixed] # convert to a list, synchronized with self.Mfixed
			if(all(((not isinstance(concentration,str)) and numpy.isscalar(concentration)) for concentration in concentrations)):
				# all fixed concentrations are scalars, i.e. they don't vary across the grid, so we can keep the first (spatial) axis of MFconcentrations[] at size 1
				self.MFconcentrations = numpy.empty((1,len(self.Mfixed)), dtype=float)
			else:
				# at least some fixed concentrations vary across the grid
				self.MFconcentrations = numpy.empty((self.NG,len(self.Mfixed)), dtype=float)
			for mf,(m,concentration) in enumerate(zip(self.Mfixed,concentrations)):
				if(isinstance(concentration,str)):
					# load concentration profile from a file
					concentration = load_profile_from_file(concentration, colname=chemistry.metabolites[m], grid_out=grid, check_nan=True)
				self.MFconcentrations[:,mf] = concentration
		self.NMD = len(self.Mdynamic)
		self.NMF = len(self.Mfixed)
		self.dynamic_metabolites = chemistry.metabolites[self.Mdynamic]

		# parse provided temperature
		if(isinstance(temperature,str)):
			# load temperature profile from a file
			temperature = load_profile_from_file(temperature, colname="temperature", grid_out=grid, check_nan=True)
		elif((not numpy.isscalar(temperature)) and (temperature.shape!=(self.NG,1))):
			raise ValueError("temperature has invalid shape %s: Expected a scalar or a 1D array of size %d"%(sprint_shape(temperature),self.NG))
		self.temperature = temperature
		
		def aux_check_names_in_specs(specs, specs_name):
			if(isinstance(specs,dict)):
				missings = [metabolite for metabolite in self.dynamic_metabolites if (metabolite not in specs)] + [reaction for reaction in chemistry.biotic_reactions if (reaction not in specs)]
				if(len(missings)>0): raise ValueError("Missing %d dynamic metabolites and/or biotic reactions in %s: %s"%(len(missings),specs_name,", ".join(missings)))
		aux_check_names_in_specs(BCleft, "BCleft")
		aux_check_names_in_specs(BCright, "BCright")
		aux_check_names_in_specs(diffusivity, "diffusivity")
		aux_check_names_in_specs(advection, "advection")
		
		def reformat_transport_parameter(param, param_name):
			if(isinstance(param,dict)):
				individual_values = [param[metabolite] for metabolite in self.dynamic_metabolites] + [param[reaction] for reaction in chemistry.biotic_reactions]
				if(all(numpy.isscalar(v) for v in individual_values)):
					# all param values are scalar (i.e., constant across the grid), so return a 2D array of size 1 x (NMD+NRB).
					return numpy.expand_dims(individual_values, axis=0)
				else:
					# at least some individual values are arrays (i.e., values differ between grid points), so return a 2D array of size NG x (NMDxNRB)
					P = numpy.empty((self.NG,self.NMD+self.NRB), dtype=float)
					for i,v in enumerate(individual_values):
						if((not numpy.isscalar(v)) and (v.shape!=(self.NG,))): raise ValueError("%s provided for at least metabolites or cells has invalid shape %s: Expected a scalar or a 1D array of size %d"%(param_name,sprint_shape(v),self.NG))
						P[:,i] = v
					return P
			elif(isinstance(param,str)):
				# load param profile from a file
				return numpy.expand_dims(load_profile_from_file(param, colname=param_name, grid_out=grid, check_nan=True), axis=1)
			elif(numpy.isscalar(param)):
				return param
			elif(isinstance(param,numpy.ndarray)):
				if((param.ndim!=1) or (param.shape[0]!=self.NG)): raise ValueError("%s has invalid shape %s: Expected a scalar or a 1D array of size %d"%(param_name,sprint_shape(param),self.NG))
				return numpy.expand_dims(param,axis=1)
		self.diffusivity = reformat_transport_parameter(diffusivity,"diffusivity")
		self.advection 	 = reformat_transport_parameter(advection,"advection")

		self.BCleft  			= (BCleft if isinstance(BCleft,dict) else defaultdict(lambda: BCleft))
		self.BCright 			= (BCright if isinstance(BCright,dict) else defaultdict(lambda: BCright))
		self.distance_unit		= distance_unit
		self.time_unit			= time_unit
		self.concentration_unit	= concentration_unit
		self.density_unit		= density_unit
		self.reaction_rate_unit	= "%s/%s"%(concentration_unit,time_unit)
		self.cell_growth_unit	= "%s/%s"%(density_unit,time_unit)
		self.density_epsilon	= density_epsilon
		self.vertical			= vertical
		self.rtol				= rtol
		self.name				= name
		self.side_names			= (["top","bottom"] if vertical else ["left","right"])
		self.plot_sizeZ			= plot_sizeZ
		self.plot_sizeV			= plot_sizeV
		self.plot_sizeT			= plot_sizeT
		
		# determine focal metabolites & reactions
		# Hence, self.Mfocals will be an integer list with values in 0,..,NM-1, specifying focal metabolites
		# Similarly, self.MDfocals will be an integer list with values in 0,..,NMD-1, specifying focal dynamic metabolites
		if(Mfocals is None): 
			self.Mfocals, self.MDfocals = None, None
		else:
			self.Mfocals  = filter_names(names=self.metabolites, only_names=Mfocals, pattern_type="wildcard")
			self.MDfocals = filter_names(names=self.dynamic_metabolites, only_names=Mfocals, pattern_type="wildcard")
		if(Rfocals is None): 
			self.Rfocals, self.RBfocals = None, None
		else:
			self.Rfocals = filter_names(names=self.reactions, only_names=Rfocals, pattern_type="wildcard")
			self.RBfocals = filter_names(names=self.biotic_reactions, only_names=Rfocals, pattern_type="wildcard")

		# define standard observables always expected in simulation results, e.g. for plotting and saving to files
		self.observable_names 	= ["concentrations", 			"cell_densities",		"reaction_rates", 			"metabolite_net_productions", 	"metabolite_gross_productions", "metabolite_gross_consumptions", 	"cell_net_productions", 			"cell_gross_productions",			"cell_gross_deaths",				"cell_per_capita_net_productions",	"cell_per_capita_gross_productions",	"cell_per_capita_gross_deaths",	"metabolite_fluxes",													"cell_fluxes"]
		self.observable_labels 	= ["concentration", 			"density",				"rate", 					"rate", 						"rate", 						"rate", 							"rate", 							"rate",								"rate",								"per-capita rate",					"per-capita rate",						"per-capita rate",				"flux",																	"flux"]
		self.observable_parts 	= [self.dynamic_metabolites,	self.biotic_reactions,	chemistry.reactions,		self.dynamic_metabolites,		self.dynamic_metabolites,		self.dynamic_metabolites,			self.chemistry.biotic_reactions,	self.chemistry.biotic_reactions,	self.chemistry.biotic_reactions,	self.chemistry.biotic_reactions,	self.chemistry.biotic_reactions,		self.chemistry.biotic_reactions,self.dynamic_metabolites,												self.biotic_reactions]
		self.observable_units 	= [concentration_unit,			density_unit,			self.reaction_rate_unit,	self.reaction_rate_unit,		self.reaction_rate_unit,		self.reaction_rate_unit,			self.cell_growth_unit,				self.cell_growth_unit,				self.cell_growth_unit,				"1/"+time_unit,						"1/"+time_unit,							"1/"+time_unit,					"%s*%s/%s"%(self.concentration_unit,self.distance_unit,self.time_unit),	"%s*%s/%s"%(self.density_unit,self.distance_unit,self.time_unit)]
		self.observable_flux	= [False,						False,					True,						True,							True,							True,								True,								True,								True,								True,								True,									True,							True,																	True]
		self.observable_focals 	= [self.MDfocals,				self.RBfocals,			self.Rfocals,				self.MDfocals,					self.MDfocals,					self.MDfocals,						self.RBfocals,						self.RBfocals,						self.RBfocals,						self.RBfocals,						self.RBfocals,							self.RBfocals,					self.MDfocals,															self.RBfocals]



	def __repr__(self):
		description = "%s, grid spanning %d points across %g - %g %s"%(self.name,len(self.grid),self.grid[0],self.grid[-1],self.distance_unit)\
					+ "\nTransport time scales: %g %s (diffusion) & %g %s (advection)"%(self.diffusion_time_scale(),self.time_unit,self.advection_time_scale(),self.time_unit)\
					+ "\n"+str(self.chemistry)
		if(len(self.Mfixed)>0):
			if(self.NMD==0): 
				description += "\nAll metabolites are fixed:"
			else:
				description += "\nFixed metabolite concentrations:"
			max_metabolite_len = max(len(metabolite) for metabolite in self.chemistry.metabolites)
			if(self.MFconcentrations.shape[0]==1):
				description += "".join("\n  %s: %g %s"%(pad_string_right(self.chemistry.metabolites[m],min_len=max_metabolite_len+1),self.MFconcentrations[0,mf],self.concentration_unit) for mf,m in enumerate(self.Mfixed))
			else:
				description += "".join("\n  %s: %s, fixed concentration profile ranges from %g to %g %s"%(pad_string_right(self.chemistry.metabolites[m],min_len=max_metabolite_len+1),self.chemistry.Mdescriptions[m],numpy.min(self.MFconcentrations[:,mf]),numpy.max(self.MFconcentrations[:,mf]),self.concentration_unit) for mf,m in enumerate(self.Mfixed))
				
		if(self.NMD>0):
			if(self.NMF==0): description += "\nAll metabolites are dynamic"
			description += "\nDynamic metabolite boundary conditions:" + "".join("\n  %s: %s %s %g, %s %s %g"%(pad_string_right(metabolite,min_len=max_metabolite_len+1),self.side_names[0],self.BCleft[metabolite][0],self.BCleft[metabolite][1],self.side_names[1],self.BCright[metabolite][0],self.BCright[metabolite][1]) for metabolite in self.dynamic_metabolites)
		return description
						
			

	# inherit some properties from the reaction pool for convenience
	@property
	def NR(self): return self.chemistry.NR

	@property
	def NM(self): return self.chemistry.NM

	@property
	def NRB(self): return self.chemistry.NRB

	@property
	def biotic_reactions(self): return self.chemistry.biotic_reactions

	@property
	def reactions(self): return self.chemistry.reactions

	@property
	def metabolites(self): return self.chemistry.metabolites


	def diffusion_time_scale(self):
		return divide_nansafe((self.grid[-1]-self.grid[0])**2,(self.diffusivity if numpy.isscalar(self.diffusivity) else numpy.mean(numpy.min(self.diffusivity,axis=1))))
		
	def advection_time_scale(self):
		return divide_nansafe((self.grid[-1]-self.grid[0]),(self.advection if numpy.isscalar(self.advection) else numpy.mean(numpy.min(self.advection,axis=1))))
		
	def copy(self):
		 return copy.deepcopy(self)
		
	# return a hash string representing this BioreactorColumn instance, focusing only on attributes that affect its dynamics but not visualizations 
	def hash_dynamics(self):
		return hash_object([self.Mfixed, self.Mdynamic, self.MFconcentrations, self.chemistry, self.temperature, self.grid, self.BCleft, self.BCright, self.diffusivity, self.advection, self.distance_unit, self.concentration_unit, self.time_unit, self.density_unit, self.rtol, self.density_epsilon])


	# add the final profiles & their averages to simulation results in-place, if missing
	def ensure_final_state(self, sim):
		for observable in self.observable_names:
			if(("final_"+observable) in sim):
				# final profile is already present
				pass
			elif(observable not in sim):
				continue
			elif(sim[observable].ndim==3):
				# observable is in sim but includes time axis, so just extract the last time point
				sim["final_"+observable] = sim[observable][-1,...]
			elif(sim[observable].shape[0]==len(sim.grid)):
				sim["final_"+observable] = sim[observable]
			# also compute average across the grid, i.e. collapsing the spatial axis
			if(("final_mean_"+observable) not in sim):
				sim["final_mean_"+observable] = integrate_on_grid(grid=sim.grid, values=sim["final_"+observable], axis=0, average=True, degree=1, check_nan=True)


	# save simulation results at a single final time point, as a tabular file (one row per grid point)
	def save_final_state(self,
						 sim, 				# simulation results, as returned e.g. by self.simulate()
						 profiles_filepath,	# string, path to an output CSV/TSV/XLSX file to save the final profiles (one row per grid point, one column per observable)
						 means_filepath):	# string, path to an output CSV/TSV/XLSX file to save the final grid-averages (one row per observable)
		self.ensure_final_state(sim)
		# save grid profiles
		DF = pandas.DataFrame({("depth" if self.vertical else "location"):self.grid} | {(observable+"."+part):sim["final_"+observable][:,p] for observable,parts in zip(self.observable_names,self.observable_parts) for p,part in enumerate(parts)}, copy=False)
		if((parent:=os.path.dirname(profiles_filepath))!=""): os.makedirs(parent, exist_ok=True)
		if(profiles_filepath.lower().endswith(".xlsx")):
			DF.to_excel(profiles_filepath, sheet_name="final", na_rep="", index=False, freeze_panes=(1,1))
		else:
			DF.to_csv(profiles_filepath, sep=infer_file_delimiter(profiles_filepath, default="\t"), index=False, na_rep="")
		# save grid-means
		DF = pandas.DataFrame({	"observable":[(observable+"."+part) for observable,parts in zip(self.observable_names,self.observable_parts) for part in parts],\
								"final_mean":[sim["final_mean_"+observable][p] for observable,parts in zip(self.observable_names,self.observable_parts) for p,part in enumerate(parts)],\
								"unit":[unit for unit,parts in zip(self.observable_units,self.observable_parts) for part in parts]},
							  copy = False)
		if((parent:=os.path.dirname(means_filepath))!=""): os.makedirs(parent, exist_ok=True)
		if(means_filepath.lower().endswith(".xlsx")):
			DF.to_excel(means_filepath, sheet_name="final_mean", na_rep="", index=False, freeze_panes=(1,0))
		else:
			DF.to_csv(means_filepath, sep=infer_file_delimiter(means_filepath, default="\t"), index=False, na_rep="")


	# plot simulation results at a single final time point, as curves (one file per variable type, one subplot per variable)
	# For example, this may be used to save & plot the equilibrium state of the system after convergence
	def plot_final_profiles(self,
							sim, # simulation results, as returned e.g. by self.simulate()
							plot_basepath):
		self.ensure_final_state(sim)
		Zlabel = ("depth" if self.vertical else "location")+" ("+self.distance_unit+")"
		Tlabel = "time (%s)"%(self.time_unit)
		for obs,observable in enumerate(self.observable_names):
			if(len(self.observable_parts[obs])==0): continue
			value = sim["final_"+observable]
			axis_factor_and_units = [best_axis_factor(max_abs_value=numpy.nanmax(numpy.abs(value[:,k])), unit=self.observable_units[obs]) for k in range(value.shape[1])] # determine nice axis rescalings, i.e., to get orders of magnitude around 1-100
			plot_profile(Zgrid			= sim.grid,
						 value			= value * numpy.expand_dims([fu[0] for fu in axis_factor_and_units], axis=0),
						 Zlabel			= Zlabel,
						 Vlabel			= ["%s (%s)"%(self.observable_labels[obs],fu[1]) for fu in axis_factor_and_units],
						 plot_file		= plot_basepath+observable+".pdf",
						 plot_title		= "Final "+observable.replace("_"," "),
						 subplot_titles	= self.observable_parts[obs],
						 reverseZ		= self.vertical,
						 verticalZ		= self.vertical,
						 plot_sizeZ		= self.plot_sizeZ,
						 plot_sizeV		= self.plot_sizeV)
			if((focals:=self.observable_focals[obs]) is not None):
				plot_profile(Zgrid			= sim.grid,
							 value			= value[:,focals] * numpy.expand_dims([axis_factor_and_units[k][0] for k in focals], axis=0),
							 Zlabel			= Zlabel,
							 Vlabel			= ["%s (%s)"%(self.observable_labels[obs],axis_factor_and_units[k][1]) for k in focals],
							 plot_file		= plot_basepath+observable+"_focals.pdf",
							 plot_title		= "Final "+observable.replace("_"," ")+", focals",
							 subplot_titles	= self.observable_parts[obs][focals],
							 reverseZ		= self.vertical,
							 verticalZ		= self.vertical,
							 plot_sizeZ		= self.plot_sizeZ,
							 plot_sizeV		= self.plot_sizeV)


	# plot spatiotemporal simulation results, as contour plots
	def plot_spatiotemporal(self,
							sim, # simulation results, as returned e.g. by self.simulate()
							plot_basepath):
		Zlabel = ("depth" if self.vertical else "location")+" ("+self.distance_unit+")"
		Tlabel = "time (%s)"%(self.time_unit)
		for obs, observable in enumerate(self.observable_names):
			if(len(self.observable_parts[obs])==0): continue
			plot_contour(Tgrid			= sim.times, 
						 Zgrid			= sim.grid,
						 value			= sim[observable],
						 Tlabel			= Tlabel,
						 Zlabel			= Zlabel,
						 plot_file		= plot_basepath+observable+".pdf",
						 plot_title		= "%s (%s) over space & time"%(observable.replace("_"," "),self.observable_units[obs]),
						 subplot_titles	= self.observable_parts[obs],
						 reverseZ		= self.vertical,
						 verticalZ		= self.vertical,
						 plot_sizeZ		= self.plot_sizeZ,
						 plot_sizeT		= self.plot_sizeT)
			if((focals:=self.observable_focals[obs]) is not None):
				plot_contour(Tgrid			= sim.times, 
							 Zgrid			= sim.grid,
							 value			= sim[observable][...,focals],
							 Tlabel			= Tlabel,
							 Zlabel			= Zlabel,
							 plot_file		= plot_basepath+observable+"_focals.pdf",
							 plot_title		= "%s (%s) over space & time"%(observable.replace("_"," "),self.observable_units[obs]),
							 subplot_titles	= self.observable_parts[obs][focals],
							 reverseZ		= self.vertical,
							 verticalZ		= self.vertical,
							 plot_sizeZ		= self.plot_sizeZ,
							 plot_sizeT		= self.plot_sizeT)



	# given concentrations of dynamic metabolites as a multidimensional array (last axis of size NMD), amend those with the fixed concentration profiles, thus returning a new (larger) array.
	# dynamic_concentrations[] may have arbitrary shape, except for the last 2 axes which must have size NG x NMD.
	def merge_dynamic_and_fixed_concentrations(self, dynamic_concentrations):
		concentrations = numpy.empty(dynamic_concentrations.shape[:-1]+(self.NM,), dtype=float)
		concentrations[...,self.Mdynamic] = dynamic_concentrations
		concentrations[...,self.Mfixed]   = numpy.expand_dims(self.MFconcentrations,axis=tuple(range(dynamic_concentrations.ndim-2)))
		return concentrations


	# parse the start metabolite concentrations & start cell densities and merge them into a single start_value numpy array for the PDE solver
	def parse_start_values_for_solver(self,
									  concentrations0	= 0,  # start concentrations of dynamic metabolites at time 0. Either a scalar float (same start concentration for all dynamic metabolites at all grid points), or a 1D numeric array of size NG specifying start concentrations varying along the grid (but equal for all metabolites), or a dictionary mapping dynamic metabolite names to scalar start concentrations, or a 2D numeric array of size [1 | NG] x [1 | NDM] specifying a separate initial concentration for each dynamic metabolite at each grid point; axes of size 1 are broadcasted.
									  cell_densities0	= 0): # start cell densities at time 0. Allowed formats are similar to concentrations0. If a dictionary, keys must be biotic reaction names. Only relevant if the reaction pool includes biotic reactions.
		if(numpy.isscalar(concentrations0) and numpy.isscalar(cell_densities0) and (concentrations0==cell_densities0)):
			# concentrations0 & cell_densities0 are equal scalars, so simply keep one scalar as start value
			start_value = concentrations0
		else:
			# format the start concentrations of dynamic metabolites into a 2D array of size [1 | NG ] x [1 | NMD]
			if(isinstance(concentrations0,dict)):
				concentrations0 = numpy.expand_dims([param[metabolite] for metabolite in self.dynamic_metabolites], axis=0)
			elif(numpy.isscalar(concentrations0)):
				concentrations0 = numpy.full((1,self.NMD), concentrations0, dtype=float)
			elif(isinstance(concentrations0,numpy.ndarray)):
				if((concentrations0.ndim==2) and (concentrations0.shape[0] in [1,self.NG]) and (concentrations0.shape[1] in [1,self.NMD])):
					pass
				elif((concentrations0.ndim==1) and (concentrations0.shape[0]==self.NG)):
					concentrations0 = numpy.expand_dims(concentrations0,axis=1)
				else:
					raise ValueError("concentrations0 has invalid shape %s: Expected a scalar or a 1D array of size %d"%(sprint_shape(concentrations0),self.NG))
			# format the start cell densities of biotic reactions into a 2D array of size [1 | NG ] x [1 | NRB]
			if(isinstance(cell_densities0,dict)):
				cell_densities0 = numpy.expand_dims([param[reaction] for reaction in self.chemistry.biotic_reactions], axis=0)
			elif(numpy.isscalar(cell_densities0)):
				cell_densities0 = numpy.full((1,self.NRB), cell_densities0, dtype=float)
			elif(isinstance(cell_densities0,numpy.ndarray)):
				if((cell_densities0.ndim==2) and (cell_densities0.shape[0] in [1,self.NG]) and (cell_densities0.shape[1] in [1,self.NRB])):
					pass
				elif((cell_densities0.ndim==1) and (cell_densities0.shape[0]==self.NG)):
					cell_densities0 = numpy.expand_dims(cell_densities0,axis=1)
				else:
					raise ValueError("cell_densities0 has invalid shape %s: Expected a scalar or a 1D array of size %d"%(sprint_shape(cell_densities0),self.NG))
			# merge into a 1D or 2D array
			if((self.NRB==0) and (self.NMD>0)):
				return concentrations0
			elif((self.NRB>0) and (self.NMD==0)):
				return cell_densities0
			elif((concentrations0.shape[0]==1) and (cell_densities0.shape[0]==1)):
				# both concentrations0 & cell_densities0 are row vectors, so just concatenate into a single row
				start_value = numpy.hstack((concentrations0,cell_densities0))
			elif((concentrations0.shape[1]==1) and (cell_densities0.shape[1]==1) and numpy.all(concentrations0==cell_densities0)):
				# both concentrations0 & cell_densities0 are column vectors and equal at all grid points, so keep a single column vector
				start_value = concentrations0
			else:
				# at least one of concentrations0 & cell_densities0 is a column vector (i.e., varies along the grid) and they are not equal, so merge into a full 2D array of size NG x (NMD+NRB)
				start_value = numpy.empty((self.NG,self.NMD+self.NRB))
				start_value[:,:self.NMD] = concentrations0
				start_value[:,self.NMD:] = cell_densities0
		return start_value


	def get_boundary_conditions_for_solver(self):
		BCleftType		= [self.BCleft[metabolite][0] for metabolite in self.dynamic_metabolites] + [self.BCleft[reaction][0] for reaction in self.biotic_reactions]
		BCleftTarget	= [self.BCleft[metabolite][1] for metabolite in self.dynamic_metabolites] + [self.BCleft[reaction][1] for reaction in self.biotic_reactions]
		BCrightType		= [self.BCright[metabolite][0] for metabolite in self.dynamic_metabolites] + [self.BCright[reaction][0] for reaction in self.biotic_reactions]
		BCrightTarget	= [self.BCright[metabolite][1] for metabolite in self.dynamic_metabolites] + [self.BCright[reaction][1] for reaction in self.biotic_reactions]
		return BCleftType, BCleftTarget, BCrightType, BCrightTarget

	
	def get_source_function_for_solver(self):
		def source(time,U):
			# get ROCs of all metabolites (dynamic & fixed) & cells due to the chemical reactions
			# assuming U is of shape N1 x ... x Nn x NG x (NMD+NRB)
			ROCs = self.chemistry.rates_of_change(	concentrations	= self.merge_dynamic_and_fixed_concentrations(U[...,:self.NMD]),
													cell_densities	= U[...,self.NMD:],
													include_details	= False,
													temperature		= self.temperature,
													Mfocal			= self.Mdynamic)
			# return ROCs of dynamic metabolites and cells
			return numpy.concatenate((ROCs.metabolite_net_productions, ROCs.cell_net_productions), axis=-1)
		return source


	def simulate(self,
				 times,
				 concentrations0	= 0,	# start concentrations of dynamic metabolites at time 0. Either a scalar float (same start concentration for all dynamic metabolites at all grid points), or a 1D numeric array of size NG specifying start concentrations varying along the grid (but equal for all metabolites), or a dictionary mapping metabolite names to scalar start concentrations, or a 2D numeric array of size [1 | NG] x [1 | NDM] specifying a separate initial concentration for each dynamic metabolite at each grid point; axes of size 1 are broadcasted.
				 cell_densities0	= 0,	# start cell densities at time 0. Allowed formats are similar to concentrations0. If a dictionary, keys must be biotic reaction names. Only relevant if the reaction pool includes biotic reactions.
				 grid_out			= None,
				 plot_basepath 		= None,	# optional string, basepath for output plot files. If provided, the simulation results will be visualized through various contour & profile plots, as multiple PDF files.
				 Nthreads			= 0,
				 verbosity 			= 0,	
				 verbose_prefix 	= ""):			
		BCleftType, BCleftTarget, BCrightType, BCrightTarget = self.get_boundary_conditions_for_solver()
		if(verbosity>=1): print("%sSolving PDE from time %g to %g %s.."%(verbose_prefix,times[0],times[-1],self.time_unit))
		sol = solve_reaction_diffusion_advection_1D(M				= self.NMD+self.NRB,
													grid			= self.grid,
													advection		= self.advection,
													diffusivity		= self.diffusivity,
													source			= self.get_source_function_for_solver(),
													BCleftType		= BCleftType,
													BCleftTarget	= BCleftTarget,
													BCrightType		= BCrightType,
													BCrightTarget	= BCrightTarget,
													start_value		= self.parse_start_values_for_solver(concentrations0,cell_densities0),
													rtol			= self.rtol,
													atol			= numpy.expand_dims(numpy.concatenate((self.chemistry.Mepsilon[self.Mdynamic],[self.density_epsilon]*self.chemistry.NRB)),axis=0),
													times_out		= times,
													grid_out		= grid_out,
													include_roc		= False,
													include_flux	= True,
													Nthreads		= Nthreads,
													verbosity		= max(0,verbosity-1),
													verbose_prefix	= verbose_prefix+"  ")
		if(not sol.success): return IDict(success=False, error="Solving PDE failed: "+sol.error)
		results = IDict(success=True,
						times				= sol.times,
						grid				= sol.grid,
						concentrations 		= sol.value[...,:self.NMD],
						cell_densities		= sol.value[...,self.NMD:],
						metabolite_fluxes	= sol.flux[...,:self.NMD],
						cell_fluxes			= sol.flux[...,self.NMD:])
		
		# compute reaction rates & net metabolite productions & net cell productions over space & time, since these are not returned by the PDE solver
		ROCs = self.chemistry.rates_of_change(concentrations	= self.merge_dynamic_and_fixed_concentrations(results.concentrations),
											  cell_densities	= results.cell_densities,
											  include_details		= True,
											  temperature		= self.temperature,
											  Mfocal			= self.Mdynamic)
		for key,value in ROCs.items():
			results[key] = value
			results["mean_"+key] = integrate_on_grid(grid=results.grid, values=value, axis=1, average=True, degree=1, check_nan=True) # observable averaged over the entire grid, i.e., with the spatial axis collapsed

		if(plot_basepath is not None):
			if(verbosity>=1): print("%sPlotting simulation results.."%(verbose_prefix))
			self.plot_final_profiles(sim=results, plot_basepath=plot_basepath+"final_")
			self.plot_spatiotemporal(sim=results, plot_basepath=plot_basepath)
		return results



	# simulate the system long enough until convergence to a steady state
	def equilibrate(self,
					max_time,					# maximum time up until which the PDE should be integrated, before giving up. The greater this value, the more likely it is that convergence is achieved, at the potential expense of proportionally longer computation. There will thus be at most ceil((max_time-start_time)/period) iterations.
					period			= None,		# time period to consider with each iteration, i.e., how far into the future should the PDE be solved at each iteration. If None, this is chosen automatically heuristically.
					concentrations0	= 0,		# start concentrations of dynamic metabolites at time 0. Allowed formats are similar to concentrations0 in self.simulate().
					cell_densities0	= 0,		# start cell densities at time 0. Allowed formats are similar to cell_densities0 in self.simulate().
					max_rel_change	= 0.001,	# Maximum allowed relative change in the value between successive iterations (i.e. one period apart) for convergence. Either a scalar float or a dictionary mapping dynamic metabolite & biotic reaction names to scalar floats. See the documentation of solvers.equilibrate_reaction_diffusion_advection_1D() for more details.
					max_norm_change	= 0.0001,	# Maximum allowed normalized change in the value between successive iterations (i.e. one period apart) for convergence; normalization is done by dividing each variable by its mean absolute value across the entire grid. Allowed structured are as for max_rel_change. See the documentation of solvers.equilibrate_reaction_diffusion_advection_1D() for more details.
					max_abs_change	= 0,		# Maximum allowed absolute change in the value between successive iterations (i.e. one period apart) for convergence. Allowed structured are as for max_rel_change. See the documentation of solvers.equilibrate_reaction_diffusion_advection_1D() for more details.
					grid_out		= None,		# optional 1D numeric array, spatial grid for the returned results. This may be used e.g. to request a coarser grid for the outputs, compared to the fine grid used during simulation. If None, this is equal to the BioreactorColumn's internal grid.
					only_final 		= True,		# if True, only the last system state (i.e., at the last time point) will be returned. If False, all intermediate states (one period apart from each other) are also returned.
					plot_basepath 	= None,		# optional string, basepath for output plot files. If provided, the simulation results will be visualized through various contour & profile plots, as multiple PDF files.
					data_basepath	= None,		# optional string, basepath for output data files.
					Nthreads		= 0,
					verbosity 		= 0,
					verbose_prefix 	= ""):
		BCleftType, BCleftTarget, BCrightType, BCrightTarget = self.get_boundary_conditions_for_solver()
		if(isinstance(max_rel_change,dict)):
			max_rel_change = numpy.expand_dims([max_rel_change[metabolite] for metabolite in self.dynamic_metabolites]+[max_rel_change[reaction] for reaction in self.biotic_reactions], axis=0)
		if(isinstance(max_norm_change,dict)):
			max_norm_change = numpy.expand_dims([max_norm_change[metabolite] for metabolite in self.dynamic_metabolites]+[max_norm_change[reaction] for reaction in self.biotic_reactions], axis=0)
		if(isinstance(max_abs_change,dict)):
			max_abs_change = numpy.expand_dims([max_abs_change[metabolite] for metabolite in self.dynamic_metabolites]+[max_abs_change[reaction] for reaction in self.biotic_reactions], axis=0)
		if(period is None): period = max(max_time/10000000, min(max_time/2, min(self.diffusion_time_scale(),self.advection_time_scale()), (inf if (self.chemistry.NRB==0) else 1/numpy.max(self.chemistry.RBdeaths))))
		varnames = list(self.dynamic_metabolites)+list(self.biotic_reactions)

		if(verbosity>=1): print("%sEquilibrating %s, up to time %g %s, with iteration period %g %s.."%(verbose_prefix,self.name,max_time,self.time_unit,period,self.time_unit))
		sim = equilibrate_reaction_diffusion_advection_1D(M				= self.NMD+self.NRB,
														grid			= self.grid,
														advection		= self.advection,
														diffusivity		= self.diffusivity,
														source			= self.get_source_function_for_solver(),
														BCleftType		= BCleftType,
														BCleftTarget	= BCleftTarget,
														BCrightType		= BCrightType,
														BCrightTarget	= BCrightTarget,
														start_value		= self.parse_start_values_for_solver(concentrations0, cell_densities0),
														period			= period,
														max_time		= max_time,
														grid_out		= grid_out,
														only_final 		= only_final,
														rtol			= self.rtol,
														atol			= numpy.expand_dims(numpy.concatenate((self.chemistry.Mepsilon[self.Mdynamic],[self.density_epsilon]*self.NRB)),axis=0),
														max_rel_change	= max_rel_change,
														max_norm_change	= max_norm_change,
														max_abs_change	= max_abs_change,
														include_flux	= True,
														Nthreads		= Nthreads,
														varnames		= varnames,
														verbosity		= max(0,verbosity-1),
														verbose_prefix	= verbose_prefix+"  ")
		if(not sim.success): return IDict(success=False, error=sim.error)
		
		# repackage simulation/equilibration results
		results = IDict(success					= True,
						Niterations				= sim.Niterations,
						grid					= sim.grid,
						final_time 				= sim.final_time,
						final_concentrations	= sim.final_value[...,:self.NMD],
						final_cell_densities	= sim.final_value[...,self.NMD:],
						final_metabolite_fluxes = sim.final_flux[...,:self.NMD],
						final_cell_fluxes		= sim.final_flux[...,self.NMD:])
		
		# also get rates of various processes, not returned by the PDE solver
		final_ROCs = self.chemistry.rates_of_change(concentrations	= self.merge_dynamic_and_fixed_concentrations(results.final_concentrations),
													cell_densities	= results.final_cell_densities,
													include_details	= True,
													temperature		= self.temperature,
													Mfocal			= self.Mdynamic)
		for key,value in final_ROCs.items():
			results["final_"+key] = value
		self.ensure_final_state(results) # ensure some derived final observables are added to the results as needed
		if(not only_final):
			results.concentrations		= sim.value[...,:self.NMD]
			results.cell_densities		= sim.value[...,self.NMD:]
			results.metabolite_fluxes	= sim.flux[...,:self.NMD]
			results.cell_fluxes			= sim.flux[...,self.NMD:]
			results.times				= sim.times
			ROCs = self.chemistry.rates_of_change(concentrations= self.merge_dynamic_and_fixed_concentrations(results.concentrations),
												  cell_densities= results.cell_densities,
												  include_details	= True,
												  temperature	= self.temperature,
												  Mfocal		= self.Mdynamic)
			for key,value in ROCs.items():
				results[key] = value

		if(plot_basepath is not None):
			if(verbosity>=1): print("%sPlotting equilibrium state.."%(verbose_prefix))
			self.plot_final_profiles(sim=results, plot_basepath=plot_basepath+"equilibrium_")
			if(not only_final):
				if(verbosity>=1): print("%sPlotting intermediate states over time.."%(verbose_prefix))
				self.plot_spatiotemporal(sim=results, plot_basepath=plot_basepath)
			if(verbosity>=1): print("%sPlotting convergence progress over time.."%(verbose_prefix))
			first_iteration_shown = max(1,int(sim.Niterations/10)) # omit the first 10% of iterations ("warmup") in the convergence plots
			for k,(change_type,threshold) in enumerate(zip(["relative change","normalized change","absolute change"],[max_rel_change,max_norm_change,max_abs_change])):
				plot_profile(Zgrid			= numpy.arange(first_iteration_shown,sim.Niterations+1),
							 value			= sim.changes[(first_iteration_shown-1):,k,:],
							 plot_file		= plot_basepath+"convergence_"+change_type.replace(" ","_")+"s.pdf",
							 Zlabel			= "iteration",
							 Vlabel			= change_type,
							 value_scale	= "log",
							 verticalZ		= False,
							 value_ref		= (threshold if numpy.isscalar(threshold) else numpy.mean(threshold,axis=0)),
							 plot_title		= "%s between iterations during equilibration"%(change_type),
							 subplot_titles	= varnames,
							 plot_sizeZ		= 3,
							 plot_sizeV		= 2)
				plot_profile(Zgrid			= numpy.arange(first_iteration_shown,sim.Niterations+1),
							 value			= sim.violations[(first_iteration_shown-1):,k,:],
							 plot_file		= plot_basepath+"convergence_violations_in_"+change_type.replace(" ","_")+"s.pdf",
							 Zlabel			= "iteration",
							 Vlabel			= "violation",
							 value_scale	= "log",
							 verticalZ		= False,
							 value_ref		= 1,
							 plot_title		= "Violations in %s between iterations during equilibration\nviolation = %s/threshold"%(change_type,change_type),
							 subplot_titles	= varnames,
							 plot_sizeZ		= 3,
							 plot_sizeV		= 2)
				
		if(data_basepath is not None):
			if(verbosity>=1): print("%sSaving equilibrium state to files.."%(verbose_prefix))
			self.save_final_state(sim=results, profiles_filepath=data_basepath+"profiles.tsv", means_filepath=data_basepath+"averages.tsv")
				
		return results



	# get the values and names of all "rate" model parameters, i.e., whose units involve inverse time and which would thus change if time were to be rescaled
	# Examples include the diffusivity and advection, as well as affinities of reactions under mass action kinetics
	def get_rate_parameters(self,
							average_diffusivity = "none", # if "grid", then diffusivities that depend on space (i.e., can vary between grid points) are first averaged across the grid and only their average is returned.
							average_advection = "none"): # if "grid", then advection speeds that depend on space (i.e., can vary between grid points) are first averaged across the grid and only their average is returned.
		names  = []
		values = []
		# add a transport parameter (e.g., diffusivity or advection) to names[] and values[]
		def aux_add_transport_param(param_name, names, values, average):
			param = getattr(self,param_name)
			if((average=="all") or numpy.isscalar(param)):
				# parameter is either a single scalar applicable to all metabolites or is to be averaged across the whole grid and across all metabolites
				if(not numpy.isscalar(param)):
					param_name = "mean_"+param_name
					param = numpy.mean(param)
				names.append(param_name)
				values.append(param)
			elif((average=="grid") or (param.shape[0]==1)):
				# parameter either does not depend on location or is to be averaged across the grid
				prefix = ("" if (param.shape[0]==1) else "mean_")
				param  = numpy.mean(param,axis=0)
				if(param.size==1):
					names.append(prefix+param_name)
					values.append(param[0])
				else:
					# parameter varies between metabolites and reactions
					names += [(prefix+param_name+"."+metabolite) for metabolite in self.dynamic_metabolites] + [(prefix+param_name+"."+reaction) for reaction in self.biotic_reactions]
					values += list(param)
			else:
				# parameter depends on location and should not be averaged
				number_format = best_grid_number_format(self.grid)
				if(param.shape[1]==1):
					# parameter does not depend on the metabolite or reaction
					names += [(param_name+"."+(number_format%(z))) for z in self.grid]
					values+= list(param.flat)
				else:
					# parameter varies between metabolites and reactions
					names += [(param_name+"."+(number_format%(z))+"."+name) for z in self.grid for name in numpy.concatenate((self.dynamic_metabolites,self.biotic_reactions))]
					values+= list(param.flat)
		aux_add_transport_param("diffusivity", names, values, average=average_diffusivity)
		aux_add_transport_param("advection", names, values, average=average_advection)
		# add reaction-kinetic parameters
		if(self.NR>0):
			names  += ["affinity."+reaction for reaction in self.chemistry.reactions]
			values += list(self.chemistry.Raffinities)
		# add cell death rates
		if(self.NRB>0):
			names  += ["death."+reaction for reaction in self.chemistry.biotic_reactions]
			values += list(self.chemistry.RBdeaths)
		return numpy.asarray(names, dtype=object), numpy.asarray(values, dtype=float)


		
	# modify the values of all "rate" model parameters, i.e., whose units involve inverse time and which would thus change if time were to be rescaled
	# This function should normally be used in combination with self.get_rate_parameters(), since param_values must be in the same layout as returned by self.get_rate_parameters().
	# This function may be used e.g. for local sensitivity analysis, where rate parameters are slightly modified but the overall structure of the model remains the same.
	def set_rate_parameters(self,
							param_values,
							average_diffusivity = "none",
							average_advection = "none"):
		# set the value(s) of a parameter defined on the grid such as diffusivity or advection, and return the number of values taken from param_values[], i.e., the increment needed to move to the next parameter in param_values[]
		# the number of slots used is determined by the data type/shape currently in use for this parameter.
		# The parameter may be a scalar or an array of arbitrary dimensionality whose first axis has size 1 or NG.
		def aux_set_grid_param(param_name, start_index, average):
			param = getattr(self,param_name)
			if(numpy.isscalar(param)):
				setattr(self,param_name,param_values[start_index])
				return 1
			elif(average=="grid"):
				# only modify the parameter's average across the grid (i.e., along axis 0), keeping the proportions of values the same
				NC 				= numpy.prod(param.shape[1:])
				current_means 	= numpy.mean(param,axis=0,keepdims=True)
				new_means		= numpy.reshape(param_values[start_index:(start_index+NC)],(1,)+param.shape[1:])
				rescalings 		= numpy.ones_like(current_means,dtype=float)
				rescalings[current_means!=0] = new_means[current_means!=0]/current_means[current_means!=0]
				setattr(self, param_name, param*rescalings)
				return NC
			elif(average=="all"):
				# only modify the parameter's average across the grid and across all metabolites, keeping the proportions of values the same
				current_mean = numpy.mean(param,keepdims=True)
				if(current_mean==0):
					new_param = param
				else:
					new_mean  = param_values[start_index]
					new_param = param*(new_mean/current_mean)
				setattr(self, param_name, new_param)
				return 1
			else:
				# param is an array and no averaging is to be done, so just take as many slots as needed to fill this array and reshape appropriately
				setattr(self, param_name, numpy.reshape(param_values[start_index:(start_index+param.size)],param.shape))
				return param.size
		start_index = 0
		start_index += aux_set_grid_param("diffusivity", start_index, average=average_diffusivity)
		self.diffusivity = numpy.maximum(self.diffusivity,0)
		start_index += aux_set_grid_param("advection", start_index, average=average_advection)
		# set reaction affinities
		if(self.NR>0):
			self.chemistry.Raffinities = numpy.maximum(0,param_values[start_index:(start_index+self.NR)])
			start_index += self.NR
		# set cell death rates
		if(self.NRB>0):
			self.chemistry.RBdeaths = numpy.maximum(0,param_values[start_index:(start_index+self.NRB)])
			start_index += self.NRB
		# check if anything is left, which would indicate a bug
		if(start_index!=len(param_values)): raise ValueError("param_values has wrong length (%d): Expected %d"%(len(param_values),start_index))


	# perform metabolic control analysis (MCA) of equilibrium rates/fluxes, with respect to all focal parameters of the model (e.g., diffusivities, advection speeds, reaction affinities etc)
	# Focal parameters are in this case all rate-like parameters.
	# The main output of this function are the concentration & flux control coefficients in the form of a matrix,
	#	with the (r,k)-th entry being (X_k/Y_r) · ∂Y_r/∂X_k = |X_k/Y_r| · ∂|Y_r|/∂|X_k|, 
	#	and where X_k are rate parameters and Y_k are reaction rates, fluxes or similar (either averaged over the grid, or evaluated at the grid edges).
	def analyze_metabolic_controls(	self,
									max_time,					# maximum time up until which the PDE should be integrated, before giving up. The greater this value, the more likely it is that convergence is achieved, at the potential expense of proportionally longer computation. There will thus be at most ceil((max_time-start_time)/period) iterations.
									period				= None,		# time period to consider with each iteration, i.e., how far into the future should the PDE be solved at each iteration. If None, this is chosen automatically heuristically.
									concentrations0		= 0,		# start concentrations of dynamic metabolites at time 0. Allowed formats are similar to concentrations0 in self.simulate().
									cell_densities0		= 0,		# start cell densities at time 0. Allowed formats are similar to cell_densities0 in self.simulate().
									average_diffusivity	= "none",	# Either "none", "grid" or "all". If "grid", then for diffusivities that depend on space (i.e., can vary between grid points) only their grid-average is varied for MCA. If "all", then diffusivities are averaged over the entire grid and across all metabolites, thus resulting in a single diffusivity parameter considered for MCA.
									average_advection	= "none",	# Either "none", "grid" or "all". If "grid", then for advection speeds that depend on space (i.e., can vary between grid points) only their grid-average is varied for MCA. If "all", then advection speeds are averaged over the entire grid and across all metabolites, thus resulting in a single diffusivity parameter considered for MCA..
									focal_zones			= None,		# optional sequence of 2-tuples specifying spatial intervals of interest, for example for examining the sensitivity of reaction rates or metabolite concentrations averaged within specific intervals. For example [(10,20), (100,120)] specifies that the focal zones [10,20] and [100,120].
									max_rel_change		= 0.001,	# Maximum allowed relative change in the flux value between successive iterations (i.e. one period apart) for convergence. Either a scalar float or a dictionary mapping dynamic metabolite & biotic reaction names to scalar floats. See the documentation of solvers.equilibrate_reaction_diffusion_advection_1D() for more details.
									max_norm_change		= 0.001,	# Maximum allowed normalized change in the flux value between successive iterations (i.e. one period apart) for convergence; normalization is done by dividing each variable by its mean absolute value across the entire grid. Allowed structured are as for max_rel_change. See the documentation of solvers.equilibrate_reaction_diffusion_advection_1D() for more details.
									max_abs_change		= 0,		# Maximum allowed absolute change in the flux value between successive iterations (i.e. one period apart) for convergence. Allowed structured are as for max_rel_change. See the documentation of solvers.equilibrate_reaction_diffusion_advection_1D() for more details.
									rel_first_step		= 0.01,		# first relative step (finite difference in a parameter) to try for estimating the partial derivatives dflux/dparam. To make this the actual relative step (i.e., no refinements), set max_refinements=0.
									max_refinements		= 5,		# integer, maximum number of iterative refinements for finite difference estimates of derivatives (on which the FCCs are based). If 0, no refinements are done and hence the finite difference is entirely dictated by rel_first_step.
									fcc_rtol			= 0.001,	# positive float, target relative tolerance for the estimated flux control coefficients. This is essentially the error tolerance for the partial derivatives (Jacobian) due to finite difference approximations, assuming that the flux calculations themselves are error-free. Note that this threshold may not always be satisfied, for example if max_refinements is too low or due to numerical instabilities.
									Nthreads			= 0,
									plot_basepath 		= None,		# optional string, basepath for output plot files. If provided, the MCA results will be visualized through various contour & profile plots, as multiple PDF files.
									data_basepath		= None,		# optional string, basepath for output data files. If provided, the MCA results will be saved as multiple TSV files at this basepath.
									verbosity			= 0,
									verbose_prefix		= ""):
		Nthreads = parse_Nthreads(Nthreads)
		model = self.copy()
		if(focal_zones is None): focal_zones = []
		Xnames, X0 = model.get_rate_parameters(average_diffusivity=average_diffusivity, average_advection=average_advection) # focal parameters
		Ynames = numpy.asarray([("mean_"+observable+"."+part) for observable,parts in zip(self.observable_names,self.observable_parts) for part in parts]\
							  + [("mean_%s_%gto%g.%s"%(observable,zone[0],zone[1],part)) for observable,parts in zip(self.observable_names,self.observable_parts) for zone in focal_zones for part in parts]\
							  + [("flux_"+self.side_names[0]+"_boundary."+name) for name in self.dynamic_metabolites]\
							  + [("flux_"+self.side_names[1]+"_boundary."+name) for name in self.dynamic_metabolites]\
							  + [("flux_"+self.side_names[0]+"_boundary."+name) for name in self.biotic_reactions]\
							  + [("flux_"+self.side_names[1]+"_boundary."+name) for name in self.biotic_reactions]) # 1D array listing all names of responses, e.g. reaction_rates.amo, reaction_rates.nxr etc.
							  
		# determine which responses are fluxes
		# Yflux will be a 1D boolean array specifying for each response Y whether it is a "flux" (rather than a "concentration")
		Yflux  = numpy.asarray(	[flux for flux,parts in zip(self.observable_flux,self.observable_parts) for part in parts]\
							  + [flux for flux,parts in zip(self.observable_flux,self.observable_parts) for zone in focal_zones for part in parts]\
							  + [True]*self.NMD*2\
							  + [True]*self.NRB*2, dtype=bool)
		
		# determine which responses are focal
		# Yfocal will be a 1D boolean array specifying for each response Y whether it is focal
		Yfocal = numpy.concatenate([(numpy.zeros(len(parts),dtype=bool) if (focals is None) else indices2booleans(N=len(parts), indices=focals)) for focals,parts in zip(self.observable_focals,self.observable_parts)]\
								 + [(numpy.zeros(len(parts)*len(focal_zones),dtype=bool) if (focals is None) else numpy.tile(indices2booleans(N=len(parts), indices=focals),len(focals))) for focals,parts in zip(self.observable_focals,self.observable_parts)]
								 +([numpy.zeros(self.NMD*2,dtype=bool)] if (self.MDfocals is None) else [indices2booleans(N=self.NMD,indices=self.MDfocals)]*2)\
								 +([numpy.zeros(self.NRB*2,dtype=bool)] if (self.RBfocals is None) else [indices2booleans(N=self.NRB,indices=self.RBfocals)]*2), dtype=bool)


		# auxiliary function for extracting all grid-averaged responses (Y) & boundary fluxes from a previously computed equilibrium
		def aux_extract_responses(sim):
			return numpy.concatenate(tuple(sim["final_mean_"+observable] for observable in self.observable_names)\
									+ tuple(integrate_on_grid(grid=self.grid, values=sim["final_"+observable], axis=0, grid_min=zone[0], grid_max=zone[1], average=True) for observable in self.observable_names for zone in focal_zones)\
									+ (sim.final_metabolite_fluxes[0,:],sim.final_metabolite_fluxes[-1,:],sim.final_cell_fluxes[0,:],sim.final_cell_fluxes[-1,:]))

		# compute equilibrium for default parameter values (i.e., at X0)
		if(verbosity>=1): print("%sEquilibrating %s at X0.."%(verbose_prefix,self.name))
		equilibrium = model.equilibrate(max_time		= max_time, 
										period			= period, 
										concentrations0	= concentrations0,
										cell_densities0	= cell_densities0,
										max_rel_change	= max_rel_change, 
										max_norm_change	= max_norm_change, 
										max_abs_change	= max_abs_change, 
										only_final		= True, 
										plot_basepath 	= plot_basepath+"equilibrium_at_X0/",
										data_basepath	= data_basepath+"equilibrium_at_X0/",
										Nthreads		= Nthreads,
										verbosity		= verbosity-1, 
										verbose_prefix	= verbose_prefix+"  ")
		if(not equilibrium.success): return IDict(success=False, error="Failed to equilibrate at default parameter values: %s"%(equilibrium.error))
		Y0 = aux_extract_responses(equilibrium)

		# function that maps focal model parameters (X) to the resulting equilibrium observables (Y)
		def aux_X2Y(X, verbosity=0, verbose_prefix=""):
			model.set_rate_parameters(param_values=X, average_diffusivity=average_diffusivity, average_advection=average_advection)
			sim = model.equilibrate(max_time		= max_time, 
									period			= period,
									concentrations0	= equilibrium.final_concentrations, # start at the equilibrium established for X0 since it will be close to the new equilibrium, thus saving computing time
									cell_densities0	= equilibrium.final_cell_densities,
									max_rel_change	= max_rel_change, 
									max_norm_change	= max_norm_change, 
									max_abs_change	= max_abs_change, 
									only_final		= True, 
									Nthreads		= max(1,int(math.ceil(Nthreads/max(1,min(len(X0), Nthreads))))),
									verbosity		= verbosity, 
									verbose_prefix	= verbose_prefix)
			return aux_extract_responses(sim)
			
		# estimate control coefficients (CCs) based on Jacobian
		if(verbosity>=1): print("%sEstimating control coefficients for %d temporal params.."%(verbose_prefix,numpy.sum(X0!=0)))
		Jacobian = estimate_Jacobian_FD(F				= aux_X2Y, 
										X0				= X0,
										dX0				= numpy.abs(X0)*rel_first_step,
										Y0				= Y0,
										max_refinements	= max_refinements,
										rtol			= fcc_rtol,
										normalize		= False,
										Xnames			= Xnames,
										verbose_F		= True,
										verbosity		= max(verbosity-1,0),
										verbose_prefix 	= "  ")
		CCs = numpy.expand_dims(X0, axis=0) * divide_nansafe(Jacobian,numpy.expand_dims(Y0, axis=1))
									
		# omit any CCs w.r.t. parameters that completely failed (likely because those parameters are zero)
		validX 	= numpy.where(numpy.any(numpy.isfinite(CCs),axis=0))[0]
		CCs 	= CCs[:,validX]
		Xnames 	= Xnames[validX]
		X0 		= X0[validX]

		# omit any responses that completely failed (likely because those responses are always zero)
		validY 	= numpy.where(numpy.any(numpy.isfinite(CCs),axis=1))[0]
		CCs 	= CCs[validY,:]
		Ynames 	= Ynames[validY]
		Yflux	= Yflux[validY]
		Yfocal	= Yfocal[validY]
		Y0 		= Y0[validY]

		# basic check of the CCs: The sum of flux CCs in each row should be equal to 1, and the sum of concentration CCs in each row should be equal to 0
		if(verbosity>=2): 
			print("%s  Sum of flux CCs across all temporal parameters ranges from %g to %g"%(verbose_prefix,numpy.min(numpy.nansum(CCs[Yflux,:],axis=1)),numpy.max(numpy.nansum(CCs[Yflux,:],axis=1))))
			print("%s  Sum of concentration CCs across all temporal parameters ranges from %g to %g"%(verbose_prefix,numpy.min(numpy.nansum(CCs[~Yflux,:],axis=1)),numpy.max(numpy.nansum(CCs[~Yflux,:],axis=1))))

		# auxiliary function for plotting the control coefficients as a heatmap and saving them to a TSV file
		# Ysubset may be a 1D boolean array of length NY or a 1D integer array with values in 0,..,NY-1, specifying which responses to restrict the output to
		def aux_save_plot_ccs(Ysubset, title, file_suffix, group_by=None):
			if(group_by is not None):
				Xgroups 	= group_identicals(group_by) # sequence of integer lists, where each element Xgroups[g] is an integer list that specifies which X-params are in the g-th group
				local_Xnames= [group_by[group[0]] for group in Xgroups]
				local_CCs	= numpy.asarray([numpy.sum(CCs[numpy.ix_(Ysubset,group)],axis=1) for group in Xgroups]).T
			else:
				local_Xnames= Xnames
				local_CCs 	= CCs[Ysubset,:]
			if(data_basepath is not None):
				if(verbosity>=1): print("%sSaving %s.."%(verbose_prefix,title))
				save_table(	filepath				= data_basepath+file_suffix+".tsv",
							table					= local_CCs,
							row_names				= Ynames[Ysubset],
							column_names			= local_Xnames,
							row_names_column		= "response"+("" if (group_by is None) else "_group"),
							save_column_names_in	= "first_data_line",
							comments				= "%s for %s at equilibrium\nGenerated on: %s"%(title,self.name,get_date_time()))
			if(plot_basepath is not None):
				if(verbosity>=1): print("%sPlotting %s.."%(verbose_prefix,title))
				all_positive = numpy.all(local_CCs>=0)
				plot_heatmap(plot_file		= plot_basepath+file_suffix+".pdf",
							table			= local_CCs,
							row_values		= Ynames[Ysubset],
							column_values	= local_Xnames,
							show_colorbar	= True,
							clabel			= "control coefficient",
							colormap		= ("Blues" if all_positive else "seismic_r"),
							colormap_min	= (0 if all_positive else -numpy.max(numpy.abs(local_CCs))),
							colormap_max	= (None if all_positive else numpy.max(numpy.abs(local_CCs))),
							show_values		= (numpy.abs(local_CCs)>0.01),
							show_grid		= "all",
							cluster			= "none",
							plot_title		= title)
		# plot FCCs
		aux_save_plot_ccs(Ysubset=Yflux, title="Flux control coefficients", file_suffix="FCCs")
		aux_save_plot_ccs(Ysubset=Yflux, title="Flux control coefficients with params grouped by observable", file_suffix="FCCs_grouped_by_observable", group_by=[name.split(".",1)[0] for name in Xnames])
		aux_save_plot_ccs(Ysubset=Yflux, title="Flux control coefficients with params grouped by metabolite or reaction", file_suffix="FCCs_grouped_by_mr", group_by=[(name.split(".",1)[-1] if ("." in name) else "nonspecific") for name in Xnames])
		# plot CCCs
		aux_save_plot_ccs(Ysubset=~Yflux, title="Concentration control coefficients", file_suffix="CCCs")
		aux_save_plot_ccs(Ysubset=~Yflux, title="Concentration control coefficients with params grouped by observable", file_suffix="CCCs_grouped_by_observable", group_by=[name.split(".",1)[0] for name in Xnames])
		aux_save_plot_ccs(Ysubset=~Yflux, title="Concentration control coefficients with params grouped by metabolite or reaction", file_suffix="CCCs_grouped_by_mr", group_by=[(name.split(".",1)[-1] if ("." in name) else "nonspecific") for name in Xnames])
		if(not is_constant(Yfocal[Yflux])):
			# also plot FCCs focusing on focal responses
			aux_save_plot_ccs(Ysubset=(Yflux & Yfocal), title="Flux control coefficients, focals", file_suffix="FCCs_focals")
			aux_save_plot_ccs(Ysubset=(Yflux & Yfocal), title="Flux control coefficients with params grouped by observable, focals", file_suffix="FCCs_grouped_by_observable_focals", group_by=[name.split(".",1)[0] for name in Xnames])
			aux_save_plot_ccs(Ysubset=(Yflux & Yfocal), title="Flux control coefficients with params grouped by metabolite or reaction, focals", file_suffix="FCCs_grouped_by_mr_focals", group_by=[(name.split(".",1)[-1] if ("." in name) else "nonspecific") for name in Xnames])
		if(not is_constant(Yfocal[~Yflux])):			
			# also plot CCCs focusing on focal responses
			aux_save_plot_ccs(Ysubset=((~Yflux) & Yfocal), title="Concentration control coefficients, focals", file_suffix="CCCs_focals")
			aux_save_plot_ccs(Ysubset=((~Yflux) & Yfocal), title="Concentration control coefficients with params grouped by observable, focals", file_suffix="CCCs_grouped_by_observable_focals", group_by=[name.split(".",1)[0] for name in Xnames])
			aux_save_plot_ccs(Ysubset=((~Yflux) & Yfocal), title="Concentration control coefficients with params grouped by metabolite or reaction, focals", file_suffix="CCCs_grouped_by_mr_focals", group_by=[(name.split(".",1)[-1] if ("." in name) else "nonspecific") for name in Xnames])

		return IDict(success = True,
					 CCs	 = CCs,		# 2D numeric array of size NY x NX, listing flux control coefficients. Rows enumerate "fluxes" (e.g., net metabolite production rates, reaction rates or similar) and columnns enumerate temporal parameters (e.g., diffusivity, cell death rates etc)
					 X0		 = X0,		# 1D numeric array of size NX, synchronized with the columns of FCC, default X values at which the FCCs were computed
					 Y0		 = Y0,		# 1D numeric array of size NY, synchronized with the rows of CC, listing the responses (fluxes & concentrations) at X0
					 Xnames	 = Xnames, 	# 1D string array of length NX, synchronized with the columns of CC, listing the parameter names
					 Ynames	 = Ynames, 	# 1D string array of length NY, synchronized with the rows of CC, listing the response names
					 Yflux	 = Yflux)	# 1D boolean array of length NY, synchronized with the rows of CC, specifying whether a response is a flux or a concentration
			
