# Routines for solving PDEs
#
# Stilianos Louca
# Copyright 2024
#
# LICENSE AGREEMENT
# - - - - - - - - -
# All rights reserved.
# Use and redistributions of this code is permitted for commercial and non-commercial purposes,
# under the following conditions:
#
#	* Redistributions must retain the above copyright notice, this list of 
#	  conditions and the following disclaimer in the code itself, as well 
#	  as in documentation and/or other materials provided with the code.
#	* Neither the name of the original author (Stilianos Louca), nor the names 
#	  of its contributors may be used to endorse or promote products derived 
#	  from this code without specific prior written permission.
#	* Proper attribution must be given to the original author, including a 
#     reference to the peer-reviewed publication through which the code was published.
#
# THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# - - - - - - - - -

import string
import sys, os
import numpy
import math
import scipy
import importlib
from numpy import nan, inf


# import other modules
from ReactionTransportModeling.general import *

#####################################################
# MARK PDE solvers


# Solve a multivariate reaction-diffusion-advection initial value problem on a 1D spatial grid with Robin boundary conditions:
#	dU/dt = d(D*dU/dz)/dz - d(vU)/dz + F
# where U(t,x) is a vector-valued function of time and space, D is the diffusivity, v the advection speed and F a source term.
# Boundary conditions are either specified using high-level type descriptions (e.g. "Dirichlet" with fixed concentration 0.001) or in the more versatile Robin form:
#	A*U + B*dU/dz = R
# The A, B & R are henceforth referred to as "Robin" coefficients.
# For example:
#	To fix the value at a boundary (Dirichlet boundary condition) set B=0 and A=1 and R=U(boundary).
# 	To fix the gradient at a boundary (Neumann boundary condition) set A=0 and B=1 and R=dU/dx(boundary).
# 	To fix the flux at a boundary set A=v and B=-D and R=flux(boundary). Note that a positive flux means material flows rightwards, i.e., in the direction of the x-axis.
#	To fix the relative slope (i.e., (1/U)*dU/dz) at a boundary set A=rel_slope and B=-1 and R=0.
# Alternatively, you may specify typical boundary conditions using (BCleftType & BCleftTarget) and similarly (BCrightType & BCrightTarget), for example "Dirichlet" or "flux".
#	In those cases the Robin coefficients A & B & R will be automatically adjusted as appropriate.
#	Either (BCleftType & BCleftTarget) or (BCleftA & BCleftB & BCleftR) must be specified, but not both. Similarly for the right boundary.
# The diffusivity and advection speed can each be time- and space-dependent, and may differ between components.
# The source term can be a function of time, space and U itself. In fact the source term at a given location may depend on the values of U at all grid points and for all components, thus allowing the modeling of interactions and far-range effects.
# The diffusivity may be specified in one of the following ways:
#	scalar: constant and identical diffusivity for all components.
#	2D array of size [1 | (N+1)] x [1 | M]: separate space-dependent time-independent diffusivity for each component. Axes of size 1 are broadcasted, so for example a 2D array of size (N+1) x 1 means that the diffusivity is the same for all components.
#	scalar-valued callable t->R specifying a time-dependent diffusivity that is constant across space and identical across components.
#	2D-array-valued callable mapping any time t to a 2D array of size [1 | (N+1)] x [1 | M], specifying a time- and space-dependent diffusivity that may differ between components. Axes of size 1 are broadcasted, so for example a callable returning a 2D array of size (N+1) x 1 means that the diffusivity is the same for all components but varies across time & space.
# Similar options exist for the advection speed.
#
# Units:
# 	The solver does not assume any specific units for time & distance or for U, as long as units are used consistently.
# 	For example, if m is the unit used for the spatial grid and hr is the unit used for time, then advection must be measured in m/hr and diffusion must be measured in m^2/hr and source must be measured in [U]/hr and fluxes (e.g., for flux-based boundary conditions) must be measured in [U]*m/hr.
def solve_reaction_diffusion_advection_1D(M,						# integer, number of components to model, i.e. the dimensionality of U
										  grid,						# 1D numeric array of size (N+1), specifying spatial coordinates (Z-values) in strictly ascending order. The grid may be regular or irregular. The finer the grid, the more accurate the solution will be.
										  start_value,				# either a scalar or a 2D array of size [1 | (N+1)] x [1 | M], specifying the initial condition at start_time, for each component and at each grid point. Any axes of size 1 will be broadcasted, so for example (N+1) x 1 means that the initial value may differ across space but is the same for all components.
										  times_out,				# 1D numpy array, specifying the times at which the solution is requested in ascending order. These must be equal or greater than start_time.
										  diffusivity,				# diffusivity. See the header for acceptable values.
										  advection,				# directional advection speed, with positive values implying a flux toward the right. See the header for acceptable values.
										  BCleftType	= None,		# either None, or a single string or a length-M sequence of strings, specifying boundary condition types at the left boundary. Choices are "value" ("Dirichlet"), "slope" ("gradient"), "rel_slope" ("rel_gradient"), "flux", "influx" or "outflux". Either (BCleftType & BCleftTarget) or (BCleftA & BCleftB & BCleftR) must be specified, but not both.
										  BCleftTarget	= None,		# values associated with BCleftType. Either None, or a scalar or 1D array of length M or a scalar-valued callable mapping t->R or a vector-valued callable mapping t-->R^M. If a callable, it specifies a time-dependent boundary condition.  The precise interpretation of BCleftTarget depends on the corresponding BCleftType, for example target is a flux if the type is "flux".
										  BCleftA		= None,		# either None, or a scalar or 1D array of length M or a scalar-valued callable mapping t->R or a vector-valued callable mapping t-->R^M. A-coefficient of the Robin boundary condition at the left end. If a callable, it specifies a time-dependent boundary condition.
										  BCleftB		= None,		# either None, or a scalar or 1D array of length M or a scalar-valued callable mapping t->R or a vector-valued callable mapping t-->R^M. B-coefficient of the Robin boundary condition at the left end. If a callable, it specifies a time-dependent boundary condition. Note that BCleftB must either be non-zero at all times (non-Dirichlet condition), or zero at all times (Dirichlet condition).
										  BCleftR		= None,		# either None, or a scalar or 1D array of length M or a scalar-valued callable mapping t->R or a vector-valued callable mapping t-->R^M. R-coefficient of the Robin boundary condition at the left end. If a callable, it specifies a time-dependent boundary condition.
										  BCrightType	= None, 	# similar to BCleftType, but for the right boundary. Either BCrightType or (BCrightA & BCrightB) must be specified, but not both.
										  BCrightTarget	= None, 	# similar to BCleftTarget, but for the right boundary.
										  BCrightA		= None,		# similar to BCleftA, but for the right boundary.
										  BCrightB		= None,		# similar to BCleftB, but for the right boundary.
										  BCrightR		= None,		# similar to NCleftR, but for the right boundary.
										  source		= 0,		# source term, i.e., local rate of change due to processes other than advection & diffusion. Either a scalar (constant source term throughout and for all components) or a 2D array of size [1 | (N+1)] x [1 | M] (separate space-dependent source term for each component), or a callable mapping any pair (t,U) to a 2D array of size [1 | (N+1)] x [1 | M ] specifying an explicitly time- and space- and U-dependent source term. Note that here the input U is a 2D array of size (N+1) x M, specifying the value of U at each grid point and for each component. This may be used to specify "reaction" kinetics and even "far-range" actions. If 0, no source is included. Any axes of size 1 are broadcasted, for example a 2D array of size (N+1) x 1 means the same source applies to all components.
										  start_time 	= 0,		# float, the time at which the start_value is defined. 
										  grid_out 		= None,		# optional 1D numpy array, specifying the Z-grid points for which the solution should be returned. This will typically be coarser or as coarse as grid[]. If None, this is set equal to grid[].
										  rtol			= 1e-3,		# relative tolerance of the ODE solver (unitless), as interpreted by scipy.integrate.solve_ivp. Either a scalar (same rtol for all components) or a 2D array of size [1 | (N+1)] x [1 | M]. Any axes of size 1 are broadcasted
										  atol			= 1e-10,	# absolute tolerance of the ODE solver (in units of U), as interpreted by scipy.integrate.solve_ivp. Either a scalar (same atol for all components) or a 2D array of size [1 | (N+1)] x [1 | M]. Any axes of size 1 are broadcasted
										  include_roc 	= False,	# if True, the rates of change (dU/dt) are included in the returned results, as a 3D array of size times_out.shape x grid_out.shape x M.
										  include_flux= False,	# if True, the spatial fluxes (v*U - D*dU/dz) are included in the returned results, as a 3D array of size times_out.shape x grid_out.shape x M.
										  Nthreads		= 0,		# integer, number of parallel threads to use where applicable. If <=0, this will be set to the number of available cores minus Nthreads.
										  verbosity		= 0,
										  verbose_prefix= ""):
	# basic input checking
	grid		= numpy.asarray(grid, dtype=float)
	times_out	= numpy.asarray(times_out, dtype=float)
	N			= grid.size-1
	grid_steps	= numpy.diff(grid)
	grid_steps2	= grid[2:] - grid[:-2]
	if(numpy.any(numpy.diff(times_out)<=0)): raise ValueError("times_out must be strictly ascending")
	if(numpy.any(times_out<start_time)): raise ValueError("times_out must not be earlier than start_time (%g); found time out as early as %g"%(start_time,numpy.min(times_out)))
	if(not numpy.isscalar(start_value)): start_value = numpy.asarray(start_value, dtype=float)
	if((not numpy.isscalar(start_value)) and ((start_value.ndim!=2) or (start_value.shape[0] not in [1,N+1]) or (start_value.shape[1] not in [1,M]))): raise ValueError("start_value has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(start_value),N+1,M))
	start_value	 = (numpy.full((N+1,M),start_value,dtype=float) if numpy.isscalar(start_value) else numpy.tile(start_value,((N+1)//start_value.shape[0],M//start_value.shape[1]))) # replicate initial values (open accordion) to achieve the same results as broadcasting along the size-1 axes
	advection0	 = (advection(start_time) if callable(advection) else advection)
	diffusivity0 = (diffusivity(start_time) if callable(diffusivity) else diffusivity)
	if((not numpy.isscalar(diffusivity0)) and ((diffusivity0.ndim!=2) or (diffusivity0.shape[0] not in [1,N+1]) or (diffusivity0.shape[1] not in [1,M]))): raise ValueError("Diffusivity at start time has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(diffusivity0),N+1,M))
	if((not numpy.isscalar(advection0)) and ((advection0.ndim!=2) or (advection0.shape[0] not in [1,N+1]) or (advection0.shape[1] not in [1,M]))): raise ValueError("Advection at start time has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(advection0),N+1,M))
	if(((BCleftA is None) != (BCleftB is None)) or ((BCleftA is None) != (BCleftR is None))): raise ValueError("BCleftA, BCleftB & BCleftR must either all be specified or neither of them")
	if(((BCrightA is None) != (BCrightB is None)) or ((BCrightA is None) != (BCrightR is None))): raise ValueError("BCrightA, BCrightB & BCrightR must either all be specified or neither of them")
	if((BCleftType is None) != (BCleftTarget is None)): raise ValueError("BCleftType and BCleftTarget must either both be specified or neither of them")
	if((BCrightType is None) != (BCrightTarget is None)): raise ValueError("BCrightType and BCrightTarget must either both be specified or neither of them")
	if((BCleftType is None) == (BCleftA is None)): raise ValueError("Either (BCleftType & BCleftTarget) or (BCleftA & BCleftB & BCleftR) must be specified, but not both")
	if((BCrightType is None) == (BCrightA is None)): raise ValueError("Either (BCrightType & BCrightTarget) or (BCrightA & BCrightB & BCrightR) must be specified, but not both")
	if((BCleftTarget is not None) and (not callable(BCleftTarget)) and (not numpy.isscalar(BCleftTarget))):   BCleftTarget = numpy.asarray(BCleftTarget, dtype=float)
	if((BCrightTarget is not None) and (not callable(BCrightTarget)) and (not numpy.isscalar(BCrightTarget))): BCrightTarget = numpy.asarray(BCrightTarget, dtype=float)
	source0 = (source(start_time,start_value) if callable(source) else source)
	if((not numpy.isscalar(source0)) and ((source0.ndim!=2) or (source0.shape[0] not in [1,N+1]) or (source0.shape[1] not in [1,M]))): raise ValueError("source at start time has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(source0),N+1,M))
	elif(not numpy.all(numpy.isfinite(source0))): raise ValueError("source at start time generates NaNs")
	if((not numpy.isscalar(rtol)) and ((rtol.ndim!=2) or (rtol.shape[0] not in [1,N+1]) or (rtol.shape[1] not in [1,M]))): raise ValueError("rtol has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(rtol),N+1,M))
	if((not numpy.isscalar(atol)) and ((atol.ndim!=2) or (atol.shape[0] not in [1,N+1]) or (atol.shape[1] not in [1,M]))): raise ValueError("atol has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(atol),N+1,M))
	if(not numpy.isscalar(rtol)): rtol = numpy.tile(rtol,((N+1)//rtol.shape[0],M//rtol.shape[1])).ravel()
	if(not numpy.isscalar(atol)): atol = numpy.tile(atol,((N+1)//atol.shape[0],M//atol.shape[1])).ravel()
	
	# parse boundary condition types if needed, i.e., convert to Robin coefficients A & B if originallly specified in terms of "type"
	def parse_BC_types(side):
		A, B, R	   		= numpy.zeros(M), numpy.zeros(M), numpy.zeros(M)
		flux_based 		= [] # boolean list to be populated in the loop below. Will list the components for which the boundary condition involves the flux, and thus A & B depend on the current advection & diffusivit, respectively.
		flux_signs 		= [] # integer list to be populated in the loop below. Will be synchronized with flux_based[]. Elements will be either +1 or -1, depending on the flux direction referenced in the boundary condition (+1 for right-ward flux)
		A_target_based 	= [] # integer list to be populated in the loop below. Will specify the components for which the boundary condition requires assigning the target value to the Robin coefficient A.
		R_target_based 	= [] # integer list to be populated in the loop below. Will specify the components for which the boundary condition requires assigning the target value to the Robin coefficient R.
		side_index 		= (0 if (side=="left") else -1) # the point on the spatial grid considered by the boundary conditions
		BCtypes	   		= (BCleftType if (side=="left") else BCrightType)
		BCtargets 		= (BCleftTarget if (side=="left") else BCrightTarget)
		if(isinstance(BCtypes,str)): BCtypes = [BCtypes]*M
		for m, BCtype in zip(range(M),BCtypes):
			if((BCtype=="value") or (BCtype=="Dirichlet")):
				A[m] = 1
				R_target_based.append(m)
				if(not callable(BCtargets)): R[m] = slice_broadcasted(BCtargets,m)
			elif((BCtype=="slope") or (BCtype=="gradient")):
				B[m] = 1
				R_target_based.append(m)
				if(not callable(BCtargets)): R[m] = slice_broadcasted(BCtargets,m)
			elif((BCtype=="rel_slope") or (BCtype=="rel_gradient")):
				B[m] = -1
				R[m] = 0
				A_target_based.append(m)
				if(not callable(BCtargets)): A[m] = slice_broadcasted(BCtargets,m)
			elif(BCtype in ["flux", "influx", "outflux"]):
				flux_based.append(m)
				flux_signs.append(1 if ((BCtype=="flux") or ((BCtype=="influx") and (side=="left")) or ((BCtype=="outflux") and (side=="right"))) else -1)
				if(not callable(advection)):   A[m] = flux_signs[-1]*slice_broadcasted(advection0,side_index,m)
				if(not callable(diffusivity)): B[m] = -flux_signs[-1]*slice_broadcasted(diffusivity0,side_index,m)
				R_target_based.append(m)
				if(not callable(BCtargets)): R[m] = slice_broadcasted(BCtargets,m)
			else:
				raise ValueError("Invalid BC type '%s'"%(BCtype))
		flux_signs = numpy.asarray(flux_signs,dtype=int)
		return A, B, R, flux_based, flux_signs, A_target_based, R_target_based

	if(BCleftType is not None):
		BCleftA, BCleftB, BCleftR, BCleft_flux_based, BCleft_flux_signs, BCleft_A_target_based, BCleft_R_target_based = parse_BC_types("left")
	if(BCrightType is not None):
		BCrightA, BCrightB, BCrightR, BCright_flux_based, BCright_flux_signs, BCright_A_target_based, BCright_R_target_based = parse_BC_types("right")
				
	# auxiliary function for getting the Robin boundary condition coefficients A, B & R at a specific time
	def get_current_BC_coefficients(time, diffusivity_now, advection_now):
		BCleftA_now = (BCleftA(time) if (callable(BCleftA)) else BCleftA)
		BCleftB_now = (BCleftB(time) if (callable(BCleftB)) else BCleftB)
		BCleftR_now = (BCleftR(time) if (callable(BCleftR)) else BCleftR)
		if(BCleftType is not None):
			if((len(BCleft_flux_based)>0) and callable(advection)):   BCleftA_now[BCleft_flux_based] = BCleft_flux_signs*slice_broadcasted(advection_now,0,BCleft_flux_based)
			if((len(BCleft_flux_based)>0) and callable(diffusivity)): BCleftB_now[BCleft_flux_based] = - BCleft_flux_signs*slice_broadcasted(diffusivity_now,0,BCleft_flux_based)
			if((len(BCleft_R_target_based)>0) and callable(BCleftTarget)): BCleftR_now[BCleft_R_target_based] = slice_broadcasted(BCleftTarget(time),0,BCleft_R_target_based)
			if((len(BCleft_A_target_based)>0) and callable(BCleftTarget)): BCleftA_now[BCleft_A_target_based] = slice_broadcasted(BCleftTarget(time),0,BCleft_A_target_based)
		BCrightA_now = (BCrightA(time) if (callable(BCrightA)) else BCrightA)
		BCrightB_now = (BCrightB(time) if (callable(BCrightB)) else BCrightB)
		BCrightR_now = (BCrightR(time) if (callable(BCrightR)) else BCrightR)
		if(BCrightType is not None):
			if((len(BCright_flux_based)>0) and callable(advection)):   BCrightA_now[BCright_flux_based] = BCright_flux_signs*slice_broadcasted(advection_now,-1,BCright_flux_based)
			if((len(BCright_flux_based)>0) and callable(diffusivity)): BCrightB_now[BCright_flux_based] = - BCright_flux_signs*slice_broadcasted(diffusivity_now,-1,BCright_flux_based)
			if((len(BCright_R_target_based)>0) and callable(BCrightTarget)): BCrightR_now[BCright_R_target_based] = slice_broadcasted(BCrightTarget(time),-1,BCright_R_target_based)
			if((len(BCright_A_target_based)>0) and callable(BCrightTarget)): BCrightA_now[BCright_A_target_based] = slice_broadcasted(BCrightTarget(time),-1,BCright_A_target_based)
		return BCleftA_now, BCleftB_now, BCleftR_now, BCrightA_now, BCrightB_now, BCrightR_now
		
	# get boundary condition Robin coefficients at start_time
	BCleftA0, BCleftB0, BCleftR0, BCrightA0, BCrightB0, BCrightR0 = get_current_BC_coefficients(start_time, diffusivity0, advection0)
	
	# determine which boundary conditions are of Dirichlet type, as they need to be treated differently later on
	LeftDirichlet  		= (numpy.full(M,(BCleftB0==0),dtype=bool) if numpy.isscalar(BCleftB0) else (BCleftB0==0))		# 1D boolean array of size M, specifying which components have Dirichlet boundary conditions at the left end
	RightDirichlet 		= (numpy.full(M,(BCrightB0==0),dtype=bool) if numpy.isscalar(BCrightB0) else (BCrightB0==0))	# 1D boolean array of size M, specifying which components have Dirichlet boundary conditions at the right end
	LeftNonDirichlet  	= ~LeftDirichlet
	RightNonDirichlet 	= ~RightDirichlet
	LeftAllDirichlet	= numpy.all(LeftDirichlet)
	RightAllDirichlet	= numpy.all(RightDirichlet)
	

	# ensure start values at boundaries are consistent with Dirichlet conditions
	if(numpy.any(LeftDirichlet)):
		start_value[0,LeftDirichlet] = slice_broadcasted(BCleftR0,LeftDirichlet)/slice_broadcasted(BCleftA0,LeftDirichlet)
	if(numpy.any(RightDirichlet)):
		start_value[-1,RightDirichlet] = slice_broadcasted(BCrightR0,RightDirichlet)/slice_broadcasted(BCrightA0,RightDirichlet)
	
	# calculate the rate of change of U at each grid point and for each of the m components
	# The input U must either be a 2D array of size (N+1) x M or a flattened 1D array of size (N+1)*M, specifying the value of U at the given time
	# The function returns the rate of change as a 1D flattened array of size (N+1)*M.
	def get_rates_of_change(time,U):
		U = U.reshape((-1,M))
		if(verbosity>=1): print("%sTime %g.."%(verbose_prefix,time), end="\r")
		
		# get current D, V & boundary condition coefficients
		advection_now 	= (advection(time) if callable(advection) else advection)
		diffusivity_now = (diffusivity(time) if callable(diffusivity) else diffusivity)
		BCleftA_now, BCleftB_now, BCleftR_now, BCrightA_now, BCrightB_now, BCrightR_now = get_current_BC_coefficients(time, diffusivity_now=diffusivity_now, advection_now=advection_now)
		
		# precompute current coefficient ratios for non-Dirichlet boundary conditions
		if(not LeftAllDirichlet):
			AoBleft_now  = slice_broadcasted(BCleftA_now,LeftNonDirichlet)/slice_broadcasted(BCleftB_now,LeftNonDirichlet)
			RoBleft_now  = slice_broadcasted(BCleftR_now,LeftNonDirichlet)/slice_broadcasted(BCleftB_now,LeftNonDirichlet)
		if(not RightAllDirichlet):
			AoBright_now = slice_broadcasted(BCrightA_now,RightNonDirichlet)/slice_broadcasted(BCrightB_now,RightNonDirichlet)
			RoBright_now = slice_broadcasted(BCrightR_now,RightNonDirichlet)/slice_broadcasted(BCrightB_now,RightNonDirichlet)

		roc = numpy.zeros((N+1,M))
		roc += (source(time,U) if callable(source) else source)

		# add contributions of advection
		if(numpy.any(advection_now!=0)):
			# advection operator away from boundaries
			roc[slice(1,N),:] -= (slice_broadcasted(advection_now,slice(2,None))*U[2:,:] - slice_broadcasted(advection_now,slice(0,-2))*U[:-2,:])/numpy.expand_dims(grid_steps2,axis=1)
			# advection operator at left boundary (non-Dirichlet conditions)
			if(not LeftAllDirichlet):
				roc[0,LeftNonDirichlet] -= U[1,LeftNonDirichlet]*(slice_broadcasted(advection_now,1,LeftNonDirichlet)-slice_broadcasted(advection_now,0,LeftNonDirichlet))/(2*grid_steps[0])\
											- U[0,LeftNonDirichlet]*slice_broadcasted(advection_now,0,LeftNonDirichlet)*AoBleft_now\
											+ slice_broadcasted(advection_now,0,LeftNonDirichlet)*RoBleft_now
			# advection operator at right boundary (non-Dirichlet conditions)
			if(not RightAllDirichlet):
				roc[-1,RightNonDirichlet] -= U[N-1,RightNonDirichlet]*(slice_broadcasted(advection_now,N,RightNonDirichlet)-slice_broadcasted(advection_now,N-1,RightNonDirichlet))/(2*grid_steps[N-1])\
											- U[N,RightNonDirichlet]*slice_broadcasted(advection_now,N,RightNonDirichlet)*AoBright_now\
											+ slice_broadcasted(advection_now,N,RightNonDirichlet)*RoBright_now
		# add contributions of diffusion
		if(numpy.any(diffusivity_now!=0)):
			# diffusion operator away from boundaries
			roc[slice(1,N),:] += ((slice_broadcasted(diffusivity_now,slice(1,N))+slice_broadcasted(diffusivity_now,slice(2,None)))*(U[2:,...]-U[1:N,...])/numpy.expand_dims(grid_steps[1:],axis=1) - (slice_broadcasted(diffusivity_now,slice(0,-2))+slice_broadcasted(diffusivity_now,slice(1,N)))*(U[1:N,...]-U[:-2,...])/numpy.expand_dims(grid_steps[:-1],axis=1))/numpy.expand_dims(grid_steps2,axis=1)
			# diffusion operator at left boundary (non-Dirichlet conditions)
			if(not LeftAllDirichlet):
				D0 = slice_broadcasted(diffusivity_now,0,LeftNonDirichlet)
				D1 = slice_broadcasted(diffusivity_now,1,LeftNonDirichlet)
				roc[0,LeftNonDirichlet]   += (1/grid_steps[0]**2) * (U[1,LeftNonDirichlet]*(1.5*D0+0.5*D1) - U[0,LeftNonDirichlet]*(0.5*D1+1.5*D0-2*grid_steps[0]*D0*AoBleft_now)) - 2*D0*RoBleft_now/grid_steps[0]
			# diffusion operator at right boundary (non-Dirichlet conditions)
			if(not RightAllDirichlet):
				DN 	 = slice_broadcasted(diffusivity_now,N,RightNonDirichlet)
				DNm1 = slice_broadcasted(diffusivity_now,N-1,RightNonDirichlet)
				roc[-1,RightNonDirichlet] += (1/grid_steps[-1]**2) * (U[N-1,RightNonDirichlet]*(1.5*DN+0.5*DNm1) - U[N,RightNonDirichlet]*(0.5*DNm1+1.5*DN+2*grid_steps[-1]*DN*AoBright_now)) + 2*DN*RoBright_now/grid_steps[-1]
				
		# ensure rate of change is zero at left & right boundaries under Dirichlet boundary conditions
		roc[0,LeftDirichlet]	= 0 # rate of change is zero at left boundary under Dirichlet boundary conditions
		roc[-1,RightDirichlet]	= 0 # rate of change is zero at right boundary under Dirichlet boundary conditions
		return roc.ravel()

	# solve ODE
	max_diffusivity = (max(numpy.max(diffusivity(time)) for time in times_out) if callable(diffusivity) else numpy.max(diffusivity))
	max_dt = (numpy.min(grid_steps)**2)/(2*max_diffusivity)
	sol = scipy.integrate.solve_ivp(fun		= get_rates_of_change,
									t_span	= (start_time, times_out[-1]),
									y0		= start_value.ravel(),
									method	= "RK45",
									t_eval	= times_out,
									max_step= max_dt,
									atol	= atol,
									rtol	= rtol)
	if(verbosity>=1): print("")
	if(not sol.success): return IDict(success=False, status=sol.status, error=sol.message)

	# parse and package ODE integration results
	results = IDict(success		 = sol.success,
					times		 = times_out,
					Nevaluations = sol.nfev,
					status		 = sol.status)	
	U = sol.y.T.reshape((-1,N+1,M))
	if(grid_out is None):
		results.value = U
		results.grid  = grid
	else:
		results.value = interpolate_time_series(x=grid_out, xp=grid, fp=U, axis=1, left=nan, right=nan, check_nan=False, Nthreads=Nthreads)
		results.grid  = grid_out

	# also return rates of change if requested
	# first compute ROCs on the original grid used for the simulation, then interpolate onto grid_out if needed
	if(include_roc):
		roc = numpy.empty_like(U)
		for t,time in enumerate(results.times):
			roc[t,...] = get_rates_of_change(time,U[t,...]).reshape((-1,M))
		results.roc = (roc if (grid_out is None) else interpolate_time_series(x=grid_out, xp=grid, fp=roc, axis=1, left=nan, right=nan, check_nan=False, Nthreads=Nthreads))
		
	# also include spatial fluxes if requested
	# first compute fluxes on the original grid used for the simulation, then interpolate onto grid_out if needed
	def aux_format_across_times(param, param0):
		if(numpy.isscalar(param)):
			return numpy.reshape(param,(1,1,1))
		elif(callable(param)):
			param = numpy.asarray([diffusivity(time) for time in times_out], dtype=float)
			if(numpy.isscalar(param0)): param = numpy.expand_dims(param, axis=(1,2))
			return param
		else:
			return numpy.expand_dims(param, axis=0)
	if(include_flux):
		flux = get_spatial_flux(M			= M, 
								grid		= grid,
								value		= U,
								diffusivity	= aux_format_across_times(diffusivity, diffusivity0),
								advection	= aux_format_across_times(advection, advection0))
		results.flux = (flux if (grid_out is None) else interpolate_time_series(x=grid_out, xp=grid, fp=flux, axis=1, left=nan, right=nan, check_nan=False, Nthreads=Nthreads))
	return results


# Solve a multivariate reaction-diffusion-advection equation until reaching steady state
# Instead of providing a fixed time span to integrate over, the caller specifies criteria for detecting convergence to steady state
# This only makes sense if the model's parameters (e.g., diffusivity or boundary conditions) are constant in the long-run.
# Internally, this function integrates the PDE over time in small segments using solve_reaction_diffusion_advection_1D(), until the latest profile is close enough to the profile from the previous run. 
# There are three alternative thresholds for establishing convergence: max_rel_change, max_norm_change, max_abs_change
#	If any of the thresholds is met, convergence is accepted. In other words, the three convergence criteria are connected via OR.
#	Specifically, convergence will be accepted if: numpy.all(abs(U-previous_U)<=max_rel_change*abs(previous_U)) OR numpy.all(abs(U-previous_U)<=max_norm_change*mean(abs(previous_U),axis=0,keepdims=True)) OR numpy.all(abs(U-previous_U)<=max_abs_change).
def equilibrate_reaction_diffusion_advection_1D(M,						# number of components to model, i.e. the dimensionality of U
											  grid,						# 1D array, spatial coordinates in ascending order. Similar to solve_reaction_diffusion_advection_1D().
											  start_value,				# value of each component at each grid point at start_time. Similar to solve_reaction_diffusion_advection_1D().
											  diffusivity,				# diffusivity. See the header of solve_reaction_diffusion_advection_1D() for acceptable values.
											  advection,				# directional advection speed. See the header of solve_reaction_diffusion_advection_1D() for acceptable values.
											  period,					# time period to consider with each iteration, i.e., how far into the future should the PDE be solved at each iteration
											  max_time,					# maximum time up until which the PDE should be integrated, before giving up. The greater this value, the more likely it is that convergence is achieved, at the potential expense of proportionally longer computation. There will thus be at most ceil((max_time-start_time)/period) iterations.
											  BCleftType		= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCleftTarget		= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCleftA			= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCleftB			= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCleftR			= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCrightType		= None, 	# Similar to solve_reaction_diffusion_advection_1D().
											  BCrightTarget		= None, 	# Similar to solve_reaction_diffusion_advection_1D().
											  BCrightA			= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCrightB			= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  BCrightR			= None,		# Similar to solve_reaction_diffusion_advection_1D().
											  source			= 0,		# source term, i.e., local rate of change due to processes other than advection & diffusion. Similar to solve_reaction_diffusion_advection_1D().
											  start_time 		= 0,		# float, the time at which the start_value is defined. 
											  grid_out 			= None,		# optional 1D numpy array, specifying the Z-grid points for which the solution should be returned. Similar to solve_reaction_diffusion_advection_1D().
											  only_final 		= True,		# if True, only the final value (i.e., at the last time point) will be returned. If False, all intermediate values (i.e., one period apart from each other) are also returned.
											  rtol				= 0.001,	# relative tolerance of the ODE solver, as interpreted by scipy.integrate.solve_ivp.
											  atol				= 1e-10,	# absolute tolerance of the ODE solver, as interpreted by scipy.integrate.solve_ivp.
											  max_rel_change	= 0.001,	# either a scalar or a 2D array of size [1 | (N+1)] x [1 | M]. Maximum allowed pointwise relative change in the value between successive iterations (i.e. one period apart) for convergence. See the header for a detailed interpretation.
											  max_norm_change	= 0.0001,	# either a scalar or a 2D array of size [1 | (N+1)] x [1 | M]. Maximum allowed pointwise normalized change in the value between successive iterations (i.e. one period apart) for convergence; normalization is done by dividing by the average absolute value of a component across the entire grid. See the header for a detailed interpretation.
											  max_abs_change	= 0,		# similar structure to max_rel_change. Maximum allowed pointwise absolute change in the value between successive iterations (i.e. one period apart) for convergence. See the header for a detailed interpretation.
											  include_flux		= False,	# if True, the spatial fluxes (v*U - D*dU/dz) are included in the returned results.
											  Nthreads			= 0,		# integer, number of parallel threads to use where applicable. If <=0, this will be set to the number of available cores minus Nthreads.
											  varnames			= None,		# optional string sequence of length M, specifying names for the varioys components of U. Only used for printing progress reports, if verbosity>0.
											  verbosity			= 0,
											  verbose_prefix	= ""):
	# basic input checking
	grid = numpy.asarray(grid, dtype=float)
	N	 = grid.size-1
	if(not numpy.isscalar(start_value)): start_value = numpy.asarray(start_value, dtype=float)
	if((not numpy.isscalar(start_value)) and ((start_value.ndim!=2) or (start_value.shape[0] not in [1,N+1]) or (start_value.shape[1] not in [1,M]))): raise ValueError("start_value has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(start_value),N+1,M))
	start_value	 = (numpy.full((N+1,M),start_value) if numpy.isscalar(start_value) else numpy.tile(start_value,((N+1)//start_value.shape[0],M//start_value.shape[1]))) # replicate initial values (open accordion) to achieve the same results as broadcasting along the size-1 axes
	if((not numpy.isscalar(max_rel_change)) and ((max_rel_change.ndim!=2) or (max_rel_change.shape[0] not in [1,N+1]) or (max_rel_change.shape[1] not in [1,M]))): raise ValueError("max_rel_change has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(max_rel_change),N+1,M))
	if((not numpy.isscalar(max_abs_change)) and ((max_abs_change.ndim!=2) or (max_abs_change.shape[0] not in [1,N+1]) or (max_abs_change.shape[1] not in [1,M]))): raise ValueError("max_abs_change has incorrect shape %s: Expected a scalar or a 2D array of size [1 | %d] x [1 | %d]"%(sprint_shape(max_rel_change),N+1,M))
	period = min(period, max_time)
	if(varnames is None): varnames = ["U%d"%(m) for m in range(M)]

	iteration  = 0
	converged  = False
	last_time  = start_time
	last_value = start_value
	changes    = [] # list to be populated in the loop below. Each i-th element will be a 3-tuple of 1D numpy arrays of size M, listing the relative, normalized & absolute changes between iteration i & (i-1) separately for all M components and maximized over the grid. 
	violations = [] # similar to changes[], but listing violations of change thresholds. The violation of a specific threshold is defined as the change divided by the acceptable threshold.
	if(not only_final):
		all_times  = []
		all_values = []
		if(include_flux): all_fluxes = []
	while ((last_time<max_time) and (not converged)):
		iteration += 1
		end_time = min(max_time, last_time + period)
		if(verbosity>=1): print("%sSolving PDE from time %g to %g (iteration #%d).."%(verbose_prefix,last_time,end_time,iteration))
		sol = solve_reaction_diffusion_advection_1D(M				= M,
												    grid			= grid,
												    start_value		= last_value,
												    diffusivity		= diffusivity,
												    advection		= advection,
												    BCleftType		= BCleftType,
												    BCleftTarget	= BCleftTarget,
												    BCleftA			= BCleftA,
												    BCleftB			= BCleftB,
												    BCleftR			= BCleftR,
												    BCrightType		= BCrightType,
												    BCrightTarget	= BCrightTarget,
												    BCrightA		= BCrightA,
												    BCrightB		= BCrightB,
												    BCrightR		= BCrightR,
												    source			= source,
												    start_time 		= last_time,
												    times_out		= [end_time],
												    grid_out		= grid_out,
												    rtol			= rtol,
												    atol			= atol,
												    include_flux	= include_flux,
												    Nthreads		= Nthreads)
		if(not sol.success): return IDict(success=False, error="Integration of PDE failed at iteration #%d (times %g - %g): %s"%(iteration,last_time,end_time,sol.error), last_time=last_time, last_value=last_value)
		# check convergence by comparing the new value to the last one
		new_value 	= sol.value[-1,...]
		rel_changes	= divide_nansafe(numpy.abs(new_value-last_value), numpy.abs(last_value))
		norm_changes= divide_nansafe(numpy.abs(new_value-last_value), numpy.mean(numpy.abs(last_value),axis=0,keepdims=True))
		abs_changes	= numpy.abs(new_value-last_value)
		changes.append((numpy.nanmax(rel_changes,axis=0),numpy.nanmax(norm_changes,axis=0),numpy.nanmax(abs_changes,axis=0)))
		violations.append((numpy.nanmax(rel_changes/max_rel_change,axis=0), numpy.nanmax(norm_changes/max_norm_change,axis=0), numpy.nanmax(abs_changes/max_abs_change,axis=0)))
		if(verbosity>=2):
			print("%s  Largest relative change compared to previous iteration is %g, largest normalized change is %g, largest absolute change is %g"%(verbose_prefix,numpy.nanmax(rel_changes),numpy.nanmax(norm_changes),numpy.nanmax(abs_changes)))
			top_violator, top_violation = argmax2(violations[-1][0], check_nan=True)
			print("%s  Relative change threshold violated most by %s, by a factor of %g"%(verbose_prefix,varnames[top_violator],top_violation))
			top_violator, top_violation = argmax2(violations[-1][1], check_nan=True)
			print("%s  Normalized change threshold violated most by %s, by a factor of %g"%(verbose_prefix,varnames[top_violator],top_violation))
			top_violator, top_violation = argmax2(violations[-1][2], check_nan=True)
			print("%s  Absolute change threshold violated most by %s, by a factor of %g"%(verbose_prefix,varnames[top_violator],top_violation))
		if(numpy.all(numpy.abs(new_value-last_value)<=max_rel_change*numpy.abs(last_value))):
			converged = True
			if(verbosity>=2): print("%s  Reached convergence after %d iterations (time %g), relative change was below threshold for all components & all grid points"%(verbose_prefix,iteration,end_time))
		elif(numpy.all(numpy.abs(new_value-last_value)<=max_norm_change*numpy.mean(numpy.abs(last_value),axis=0,keepdims=True))):
			converged = True
			if(verbosity>=2): print("%s  Reached convergence after %d iterations (time %g), normalized change was below threshold for all components & all grid points"%(verbose_prefix,iteration,end_time))
		elif(numpy.all(numpy.abs(new_value-last_value)<=max_abs_change)):
			converged = True
			if(verbosity>=2): print("%s  Reached convergence after %d iterations (time %g), absolute change was below threshold for all components & all grid points"%(verbose_prefix,iteration,end_time))
		# keep record of the new value
		last_time  = end_time
		last_value = new_value
		if(not only_final):
			all_times.append(last_time)
			all_values.append(last_value)
			if(include_flux): all_fluxes.append(sol.flux[-1,...])
	results = IDict(success		= converged,
					final_time 	= end_time,
					final_value	= last_value,
					Niterations	= iteration,
					grid		= grid,
					changes		= numpy.asarray(changes, dtype=float), # 3D array of size Niterations x 3 x M, listing for each iteration the relative & normalized & absolute change of each component compared to the previous iteration, maximized over the grid.
					violations	= numpy.asarray(violations, dtype=float)) # 3D array of size Niterations x 3 x M, listing for each iteration the violation in relative & normalized & absolute change, relative to the applicable change threshold, maximized over the grid.
	if(include_flux):
		results.final_flux = sol.flux[-1,...]
	if(not only_final): 
		results.value = numpy.asarray(all_values, dtype=float) # 3D array of size Niterations x (N+1) x M, values at the end of each iteration
		results.times = numpy.asarray(all_times, dtype=float)  # 1D array of size Niterations, listing the end time of each iteration
		if(include_flux): results.flux = numpy.asarray(all_fluxes, dtype=float) # 3D array of size Niterations x (N+1) x M, spatial fluxes at the end of each iteration
	if(not converged): results.error = "Convergence failed within the allowed time"
	return results			 
	
	
	
	
# compute the spatial fluxes (v*U - D*dU/dz) along a 1D spatial grid, in the context of a reaction-advection-diffusion model.
# The input values may be an arbitrary-dimensional array, as long as the last two axes correspond to the spatial grid (of size N+1) and to the M components, respectively.
# The function returns an array of the same shape as the input value array.
def get_spatial_flux(M,				# number of components to model, i.e. the dimensionality of U
					 grid,			# 1D array, spatial coordinates in ascending order. Similar to solve_reaction_diffusion_advection_1D().
					 value,			# value of each component at each grid point. May be an arbitrary (n+2) dimensional array of shape N1 x .. x Nn x (N+1) x M, with the last 2 axes representing location (along the grid) and components. For example this may be a simulation result, and hence a 3D array with the first axis being time. The returned fluxes[] will have the same shape as value[].
					 diffusivity,	# diffusivity at each grid point and for each component. Either a scalar or an (n+2)-dimensional array. In the case of an array, it must have the same number of axes as value[], and each axis must either have size 1 or the same size as the corresponding axis in value[]. Axes of size 1 will be broadcasted.
					 advection):	# advection speed at each grid point and for each component. Allowed formats are similar to diffusivity.
	# basic input checking
	N = len(grid)-1
	if(value.ndim<2): raise ValueError("value as incorrect shape %s: Expected at least 2 dimensions"%(sprint_shape(value)))
	elif((value.shape[-2]!=(N+1)) or (value.shape[-1]!=M)): raise ValueError("value has incorrect shape %s: Expected an array whose last two axes have size %d x %d"%(sprint_shape(value),N+1,M))
	if((not numpy.isscalar(diffusivity)) and ((diffusivity.ndim!=value.ndim) or any((diffusivity.shape[k] not in [1,value.shape[k]]) for k in range(value.ndim)))): raise ValueError("diffusivity has incorrect shape %s: Expected a scalar or a %d-dimensional array whose non-singleton axes match the axis sizes of value (shape %s)"%(sprint_shape(diffusivity),value.ndim,sprint_shape(value)))
	if((not numpy.isscalar(advection)) and ((advection.ndim!=value.ndim) or any((advection.shape[k] not in [1,value.shape[k]]) for k in range(value.ndim)))): raise ValueError("advection has incorrect shape %s: Expected a scalar or a %d-dimensional array whose non-singleton axes match the axis sizes of value (shape %s)"%(sprint_shape(advection),value.ndim,sprint_shape(value)))
	
	# start with contribution of advection
	flux = advection*value
	# add contribution of diffusion, away from boundaries
	flux[...,slice(1,N),:] -= slice_along_axis(diffusivity, axis=-2, indices=slice(1,N), broadcasted=True) * (value[...,2:,:]-value[...,:-2,:])/numpy.expand_dims(grid[2:]-grid[:-2],axis=tuple(range(value.ndim-2))+(value.ndim-1,))
	# add contributions of diffusion at the boundaries
	flux[...,0,:]  -= (diffusivity if numpy.isscalar(diffusivity) else diffusivity[...,0,:])*(value[...,1,:]-value[...,0,:])/(grid[1]-grid[0])
	flux[...,-1,:] -= (diffusivity if numpy.isscalar(diffusivity) else diffusivity[...,-1,:])*(value[...,-1,:]-value[...,-2,:])/(grid[-1]-grid[-2])
	return flux
	
	
