#!/usr/bin/env python3
#
# This code simulates and performs metabolic control analysis of various biogeochemical models, at steady state.
# Models are all in the form of 1-dimensional reaction-advection-diffusion partial differential equations.
# To use this code, run it as a standalone python script, specifying which model to examine using the -m argument.
# For example:
#	./workflow.py -m Saanich_GCM
# or:
#	./workflow.py -m Black_Sea_sediments
#
# Tested using Python 3.11, Mac OS 15.4.
#
# Stilianos Louca
# Copyright 2024
#
# LICENSE AGREEMENT
# - - - - - - - - -
# All rights reserved.
# Use and redistributions of this code is permitted for commercial and non-commercial purposes,
# under the following conditions:
#
#	* Redistributions must retain the above copyright notice, this list of 
#	  conditions and the following disclaimer in the code itself, as well 
#	  as in documentation and/or other materials provided with the code.
#	* Neither the name of the original author (Stilianos Louca), nor the names 
#	  of its contributors may be used to endorse or promote products derived 
#	  from this code without specific prior written permission.
#	* Proper attribution must be given to the original author, including a 
#     reference to the peer-reviewed publication through which the code was published.
#
# THIS CODE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS CODE, 
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# - - - - - - - - -

import sys,os
import numpy
import ReactionTransportModeling as RTM
import argparse


# Set some default plot sizes
PLOT_SIZE_Z = 2.5
PLOT_SIZE_V = 2
PLOT_SIZE_T = 4

# parse command line arguments
parser = argparse.ArgumentParser(description="Perform metabolic control analysis of various reactive systems.", epilog="")
parser.add_argument('-m','--model', required=True, type=str, choices=["Saanich_GCM","Beaufort_SMTZ","Black_Sea_sediments","Black_Sea_sediments_slow100000","Saanich_SMTZ"], help="Model to analyze.");
parser.add_argument('-v','--verbosity', default=4, type=int, help="Verbosity. A higher value means more messages will be printed. 0 means silence, i.e., only fatal errors are printed. (default: '%(default)s')");
args = parser.parse_args()

def abort(message, exit_code=1):
	print(message)
	sys.exit(exit_code)
	

if(args.verbosity>=1): print("Analyzing %s.."%(args.model))

if(args.model=="Beaufort_SMTZ"):
	# Metabolic control analysis of sulfate-driven methane oxidation at an SMTZ, in Alaskan Beaufort Sea
	# Boundary conditions and diffusivities based on: Treude et al. (2014) Sulfate reduction and methane oxidation activity below the sulfate-methane transition zone in Alaskan Beaufort Sea continental margin sediments: Implications for deep sulfur cycling.
	# Reaction kinetics based on: Anderson and Murray (1984) A model for coupled sulfate reduction and methane oxidation in the sediments of Saanich Inlet.	
	pH  = 7.8 		# [pH in Saanich Inlet sediments. Nissenbaum1972Early, Table 3, average of the top layers in cores 1-3]
	DIC = 45e-3 	# total inorganic carbon concentration in mol/L [Treude2014Sulfate, Fig. 1, DIC at the SMTZ of core PC13]
	chemistry 		= RTM.ReactionNetwork(Mtable="input/metabolites.tsv", Rtable="input/"+args.model+"_reactions.tsv")
	porosity		= 0.66 # Treude2014Sulfate, porosity in core PC13
	tortuosity		= (1-2*numpy.log(porosity)) # Treude2014Sulfate, based on a porosity of φ=0.66 given for core PC13
	temperature 	= 0.5 # Treude et al. (2014)
	
	system = RTM.BioreactorColumn(chemistry		 	= chemistry,
								  grid				= numpy.linspace(0,200,num=100),
								  concentrations	= {"H":10**(-pH), "H2O":1} | RTM.carbonate_species_concentrations(pH, DIC, temperature=temperature, salinity=35),
								  BCleft			= {"H2S":("value",0), "SO4":("value",0.027), "CH4":("value",0)}, # upper boundary concentrations taken from Treude2014Sulfate, Fig. 1
								  BCright			= {"H2S":("flux",0),  "SO4":("value",0), 	 "CH4":("value",0.010)}, # lower boundary values concentrations from Treude2014Sulfate, Fig. 1 at depth 200 cm.
								  diffusivity		= {"SO4":151/tortuosity, "CH4":250/tortuosity, "H2S":303/tortuosity}, # diffusivities (cm^2/yr) for SO4 & CH4 & H2S taken from Treude2014Sulfate, Table 4.
								  advection 		= 0.283, # cm/yr, Treude2014Sulfate, omega in Table 4
								  temperature		= temperature,
								  distance_unit		= "cm",
								  time_unit			= "yr",
								  name				= args.model,
								  vertical			= True,
								  plot_sizeZ		= PLOT_SIZE_Z,
								  plot_sizeV		= PLOT_SIZE_V,
								  plot_sizeT		= PLOT_SIZE_T)

	# specify how to group (via averaging) the transport parameters for MCA
	average_diffusivity	= "grid"
	average_advection	= "grid"
	
	# define convergence thresholds for equilibrating the system
	eq_max_rel_change 	= 0.0001
	eq_max_norm_change 	= 0.0001
	eq_max_abs_change	= {metabolite:1e-9 for metabolite in system.dynamic_metabolites}
	eq_period			= None
	
	focal_responses = ["mean_reaction_rates.*", "flux_top_boundary.H2S", "flux_bottom_boundary.CH4"]
	focal_zones		= None


elif(args.model=="Saanich_GCM"):
	# Metabolic control analysis of the Saanich Inlet stead-state gene centric model (Louca et al. 2016)
	# Model parameters, including stoichiometries and boundary conditions, are taken from: Louca et al. (2016) Integrating biogeochemistry with multiomic sequence information in a model oxygen minimum zone
	pH  = 8.07
	DIC = 2.141e-3 # total dissolved inorganic carbon
	temperature = 9
	chemistry = RTM.ReactionNetwork(Mtable="input/metabolites.tsv", Rtable="input/"+args.model+"_reactions.tsv")
	
	system = RTM.BioreactorColumn(chemistry		= chemistry,
								  grid			= numpy.linspace(100,200,num=300),
								  concentrations= {"H":10**(-pH), "H2O":1, "POM":"input/Saanich_GCM_POM.tsv"} | RTM.carbonate_species_concentrations(pH, DIC, temperature=temperature, salinity=31),
								  BCleft		= {"NH4":("value",0), 		"O2":("value",77.23e-6), 	"NO3":("value",27.59e-6),	"NO2":("value",0.045e-6),	"N2":("value",4.8e-4), "SO4":("value",0.028), "H2S":("value",0), 			"N2O":("value",24.49e-9),	"AEROM":("rel_slope",0), 	"AAO":("rel_slope",-0.042), "anammox":("value",0), 			"RNO":("rel_slope",0.042), 		"ANO":("rel_slope",0.0324), "PDNO":("value",0)},
								  BCright		= {"NH4":("value",8.67e-6), "O2":("value",0), 			"NO3":("value",0),			"NO2":("value",0),			"N2":("flux",0), 		"SO4":("value",0.028), "H2S":("value",14.07e-6), 	"N2O":("flux",0),			"AEROM":("value",0), 		"AAO":("value",0), 			"anammox":("rel_slope",-0.11),	"RNO":("rel_slope",-0.0071),	"ANO":("value",0), 			"PDNO":("rel_slope",-0.012)},
								  diffusivity	= "input/Saanich_GCM_eddy_diffusivity.tsv",
								  advection		= {metabolite:0 for metabolite in chemistry.metabolites} | {reaction:0.1 for reaction in chemistry.biotic_reactions}, # cell advection as in Louca et al. (2016). Taken from Fennel et al 2006, sinking velocity v=0.1 m/d used in their models for phytoplankton and small detritus
								  temperature	= temperature,   # temperature used in Louca et al. 2016. Below 60 m depth temperature is fairly constant ~9C.
								  distance_unit	= "m",
								  time_unit		= "d",
								  rtol			= 1e-5,
								  name			= args.model,
								  vertical		= True,
								  Mfocals		= ["NH4", "O2", "NO3", "NO2", "H2S", "N2O"],
								  plot_sizeZ	= PLOT_SIZE_Z,
								  plot_sizeV	= PLOT_SIZE_V,
								  plot_sizeT	= PLOT_SIZE_T)

	# specify how to group (via averaging) the transport parameters for MCA
	average_diffusivity	= "grid"
	average_advection	= "grid"
	
	# define convergence thresholds for equilibrating the system
	eq_max_rel_change 	= {"NH4":0.00001, "O2":0.00001, "NO3":0.00001, "NO2":0.00001, "N2":0.0001, "SO4":0.00001, "H2S":0.00001, "N2O":0.00001} | {reaction:0.00001 for reaction in system.biotic_reactions}
	eq_max_norm_change 	= {"NH4":0.00001, "O2":0.00001, "NO3":0.00001, "NO2":0.00001, "N2":0.0001, "SO4":0.00001, "H2S":0.00001, "N2O":0.00001} | {reaction:0.00001 for reaction in system.biotic_reactions}
	eq_max_abs_change	= {metabolite:1e-9 for metabolite in system.dynamic_metabolites} | {reaction:100000 for reaction in system.biotic_reactions}
	eq_period			= None

	focal_responses = {"main":[	"mean_metabolite_net_productions.N2",\
								"mean_reaction_rates.anammox", "mean_reaction_rates.PDNO", "mean_reaction_rates.RNO",\
								"mean_cell_densities.anammox", "mean_cell_densities.PDNO", "mean_cell_densities.RNO"],\
						"boundary":["flux_top_boundary.N2", "flux_top_boundary.N2O", "flux_top_boundary.NO2", "flux_top_boundary.NO3", "flux_bottom_boundary.NH4", "flux_bottom_boundary.N2"],\
						"interface":["mean_metabolite_net_productions_120to140.N2",\
									"mean_reaction_rates_120to140.anammox", "mean_reaction_rates_120to140.PDNO", "mean_reaction_rates_120to140.RNO",\
									"mean_cell_densities_120to140.anammox", "mean_cell_densities_120to140.PDNO", "mean_cell_densities_120to140.RNO"]}
	focal_zones		= [(120,140)]

	

elif(args.model.startswith("Black_Sea_sediments")):
	# Metabolic control analysis of sulfate-driven methane oxidation & iron-driven sulfide oxidation in Black Sea sediments
	# References:
	#	Jørgensen et al. (2004). Anaerobic methane oxidation and a deep H2S sink generate isotopically heavy sulfides in 2 Black Sea sediments. Station 5. DOI:10.1016/j.gca.2003.07.017
	#	Egger et al. (2016). Anaerobic oxidation of methane alters sediment records of sulfur, iron and phosphorus in the Black Sea. Station 4. DOI:10.5194/bg-13-5333-2016
	#	Treude et al. (2014) Sulfate reduction and methane oxidation activity below the sulfate-methane transition zone in Alaskan Beaufort Sea continental margin sediments: Implications for deep sulfur cycling
	# The model is based on measurements by Jorgensen et al. (2004) and Egger et al. (2016).
	# Boundary conditions taken from Egger et al. (2016).
	# Diffusivities were computed based on seawater diffusivities and corrected for the depth-dependent porosity as described by [Egger et al. 2016, Supplemental text 1].	
	chemistry 			= RTM.ReactionNetwork(Mtable="input/metabolites.tsv", Rtable="input/"+args.model+"_reactions.tsv", omit_metabolites="H2")
	pH  	  			= 7.5 # Value assumed by Jørgensen et al. (2004) 
	grid 	  			= numpy.linspace(0,600,num=300)
	data_DIC  			= RTM.load_profile_from_file(filepath="input/Black_Sea_sediments_DIC.tsv", colname="DIC", grid_out=grid, check_nan=True)
	data_FeII 			= RTM.load_profile_from_file(filepath="input/Black_Sea_sediments_FeII.tsv", colname="FeII", grid_out=grid, check_nan=True)
	data_CH4  			= RTM.load_profile_from_file(filepath="input/Black_Sea_sediments_CH4.tsv", colname="CH4", grid_out=grid, check_nan=True)
	porosity_inf		= 0.61 # porosity at infinite depth, assumed by Egger et al. (2016), Table S1.
	porosity			= porosity_inf + (0.97-porosity_inf)*numpy.exp(-grid/95) # fitted depth profile of porosity (φ) in Black Sea sediments, by Egger et al. (2016), Supplemental text 1.
	tortuosity			= 1 - 2*numpy.log(porosity) # Egger et al. (2016)
	sedimentation		= 0.06 	# sedimentation rate (g/cm^2/year), assumed by Egger et al. (2016)
	sediment_density	= 2.31  # sediment density (g/cm^3), assumed by Egger et al. (2016)
	temperature			= 11 	# Egger et al. (2016) mentions that the in situ bottom water temperature was 11 C
	salinity			= 22 	# Egger et al. (2016)
	advection_solids	= sedimentation/(sediment_density*(1-porosity)) # vertical advection assumed by Egger et al. (2016) for solids. See function setup.compaction.1D in the R package ReacTran v1.4.3.1.
	advection_dissolved	= sedimentation*porosity_inf/(sediment_density*(1-porosity_inf)*porosity) # vertical advection assumed by Egger et al. (2016) for dissolved substances. See function setup.compaction.1D in the R package ReacTran v1.4.3.1.
	
	system = RTM.BioreactorColumn(chemistry		 	= chemistry,
								  grid				= grid,
								  concentrations	= {"H":10**(-pH), "H2O":1} | RTM.carbonate_species_concentrations(pH, data_DIC, temperature=temperature, salinity=salinity),
								  BCleft			= {"H2S":("value",0), "SO4":("value",17.49e-3), "CH4":("value",0), 				"FeII":("value",0),				"FeS":("value",0), "FeS2":("value",0)},
								  BCright			= {"H2S":("value",0), "SO4":("value",0), 	 	"CH4":("value",data_CH4[-1]),	"FeII":("value",data_FeII[-1]),	"FeS":("value",0), "FeS2":("value",20e-6)},
								  diffusivity		= {"SO4":224.8/tortuosity, "CH4":388.6/tortuosity, "H2S":377.6/tortuosity, "FeII":150.0/tortuosity, "FeS":1, "FeS2":1}, # seawater diffusivities (cm^2/yr) computed using the marelac R package for temperature 11C and salinity 22, then corrected for porosity as described by Egger et al. (2016). Diffusivities of the solids FeS & FeS2 are set to a tiny non-zero value to avoid numerical instabilities in the PDE solver.
								  advection			= {"SO4":advection_dissolved, "CH4":advection_dissolved, "H2S":advection_dissolved, "FeII":advection_dissolved, "FeS":advection_solids, "FeS2":advection_solids},
								  rtol				= 1e-5,
								  distance_unit		= "cm",
								  time_unit			= "yr",
								  name				= args.model,
								  temperature		= temperature,
								  vertical			= True,
								  plot_sizeZ		= PLOT_SIZE_Z,
								  plot_sizeV		= PLOT_SIZE_V,
								  plot_sizeT		= PLOT_SIZE_T)

	# specify how to group (via averaging) the transport parameters for MCA
	average_diffusivity	= "grid"
	average_advection	= "all"
	
	# define convergence thresholds for equilibrating the system
	eq_max_rel_change 	= 0.0001
	eq_max_norm_change 	= 0.0001
	eq_max_abs_change	= {metabolite:1e-9 for metabolite in system.dynamic_metabolites}
	eq_period			= 1000
	
	focal_responses = ["mean_metabolite_net_productions.CH4", "mean_concentrations.H2S"]
	focal_zones		= None


elif(args.model=="Saanich_SMTZ"):	
	# Metabolic control analysis of sulfate-driven methane oxidation at an SMTZ, in Saanich Inlet sediments.
	# Model adjusted from: Anderson and Murray (1984) A model for coupled sulfate reduction and methane oxidation in the sediments of Saanich Inlet.
	pH  = 7.8 		# [Nissenbaum1972Early, Table 3, average of the top layers in cores 1-3]
	DIC = 24.5e-3 	# total inorganic carbon concentration in mol/L [Nissenbaum1972Early, Table 3, average of the top layers in cores 1-3]
	chemistry = RTM.ReactionNetwork(Mtable="input/metabolites.tsv", Rtable="input/"+args.model+"_reactions.tsv")
	
	system = RTM.BioreactorColumn(chemistry		 	= chemistry,
								  grid				= numpy.linspace(0,0.25,num=100),
								  concentrations	= {"H":10**(-pH), "H2O":1} | RTM.carbonate_species_concentrations(pH, DIC),
								  BCleft			= {"H2S":("value",0), "SO4":("value",0.028), "CH4":("value",0)},
								  BCright			= {"H2S":("flux",0),  "SO4":("value",0), 	 "CH4":("value",0.0048)},
								  diffusivity		= {"SO4":5e-10*3600*24*365.25, "CH4":10e-10*3600*24*365.25, "H2S":8.2e-5*365.25}, # diffusivities (m^2/yr) for SO4 & CH4 taken from Nissenbaum1972Early. H2S diffusivity is as in Louca2022FRT.
								  advection			= 0,
								  time_unit			= "yr",
								  name				= args.model,
								  vertical			= True,
								  plot_sizeZ		= PLOT_SIZE_Z,
								  plot_sizeV		= PLOT_SIZE_V,
								  plot_sizeT		= PLOT_SIZE_T)

	# specify how to group (via averaging) the transport parameters for MCA
	average_diffusivity	= "grid"
	average_advection	= "grid"
	
	# define convergence thresholds for equilibrating the system
	eq_max_rel_change 	= 0.0001
	eq_max_norm_change 	= 0.0001
	eq_max_abs_change	= {metabolite:1e-9 for metabolite in system.dynamic_metabolites}
	eq_period			= None

	focal_responses = ["mean_reaction_rates.*", "flux_top_boundary.H2S", "flux_bottom_boundary.CH4"]
	focal_zones		= None



#############################################

if(args.verbosity>=1): print(system)

# pre-equilibrate the system
if(args.verbosity>=1): print("Equilibrating %s.."%(system.name))
pickle_file = "output/"+system.name+"/equilibrium.pickle"
expected_checksum = system.hash_dynamics()
if((equilibrium:=RTM.load_object(pickle_file, graceful=True, checksum=expected_checksum)) is None):
	equilibrium = system.equilibrate(max_time		= 100000,
									 period			= eq_period,
									 concentrations0= 0,
									 cell_densities0= 1e6,
									 max_rel_change	= eq_max_rel_change,
									 max_norm_change= eq_max_norm_change,
									 max_abs_change	= eq_max_abs_change,
									 only_final		= False,
									 plot_basepath 	= "output/"+system.name+"/equilibrium/",
									 data_basepath	= "output/"+system.name+"/equilibrium/",
									 verbosity 		= args.verbosity-1,
									 verbose_prefix	= "  ")
	RTM.save_object(RTM.amend_object(equilibrium, checksum=expected_checksum), filepath=pickle_file)
else:
	if(args.verbosity>=2): print("  Note: Using previously computed (pickled) equilibrium..")
if(not equilibrium.success): 
	abort("ERROR: Could not equilibrate system: %s"%(equilibrium.error))


# perform MCA around the equilibrium
if(args.verbosity>=1): print("Performing MCA of %s.."%(system.name))
pickle_file = "output/"+system.name+"/MCA.pickle"
expected_checksum = system.hash_dynamics()
if((MCA:=RTM.load_object(pickle_file, graceful=True, checksum=expected_checksum)) is None):
	MCA = system.analyze_metabolic_controls(concentrations0		= equilibrium.final_concentrations,
											cell_densities0		= equilibrium.final_cell_densities,
											max_time			= 100000,
											period				= eq_period,
											average_diffusivity	= average_diffusivity,
											average_advection	= average_advection,
											focal_zones			= focal_zones,
											max_rel_change		= eq_max_rel_change,
											max_norm_change		= eq_max_norm_change,
											max_abs_change		= eq_max_abs_change,
											rel_first_step		= 0.01,
											max_refinements		= 0,
											plot_basepath 		= "output/"+system.name+"/MCA/",
											data_basepath		= "output/"+system.name+"/MCA/",
											verbosity			= args.verbosity-1,
											verbose_prefix		= "  ")
	RTM.save_object(RTM.amend_object(MCA, checksum=expected_checksum), filepath=pickle_file)
else:
	if(args.verbosity>=2): print("Note: Using previously computed (pickled) MCA..")
if(not MCA.success):
	abort("ERROR: Could not perform MCA: %s"%(MCA.error))


###########################################
# Additional refined plots, for publication

if(args.verbosity>=1): print("Plotting barplots of individual control coefficients..")
Xnames = [name.replace("_"," ").replace(".",": ") for name in MCA.Xnames]
for r in range(MCA.CCs.shape[0]):
	RTM.interleaved_barplot(plot_file		= "output/"+system.name+"/MCA/control_coefficients_barplots/"+MCA.Ynames[r]+".pdf",
							values			= MCA.CCs[r,:],
							member_names	= Xnames,
							vertical		= True,
							value_label		= "coefficient",
							value_ref		= 0,
							show_grid		= True,
							plot_width		= 3.2,
							plot_height		= 1,
							plot_title		= ("Flux" if MCA.Yflux[r] else "Concentration")+" control coefficient: "+MCA.Ynames[r])

if(not isinstance(focal_responses,dict)): focal_responses = {"":focal_responses}
for focals_label,focals_wildcards in focal_responses.items():
	for flux,cctype in zip([True,False], ["flux","concentration"]):
		Yfocals = RTM.filter_names(names = MCA.Ynames, start_pool=numpy.where(MCA.Yflux==flux)[0], only_names=focals_wildcards, pattern_type="wildcard")
		if(len(Yfocals)==0): continue
		if(args.verbosity>=1): print("Plotting heatmaps of focal "+cctype+" control coefficients"+("" if (focals_label=="") else " ("+focals_label+")")+"..")
		focal_CCs = MCA.CCs[Yfocals,:]
		focal_CCs[numpy.abs(focal_CCs)<1e-4] = 0
		all_positive = numpy.all(focal_CCs>=0)
		RTM.plot_heatmap(plot_file		= "output/"+system.name+"/MCA/focal_"+cctype+"_control_coefficients"+("" if (focals_label=="") else "_"+focals_label)+".pdf",
						table			= focal_CCs,
						row_values		= MCA.Ynames[Yfocals],
						column_values	= [name.replace("_"," ").replace(".",": ") for name in MCA.Xnames],
						show_colorbar	= True,
						colormap		= ("Blues" if all_positive else "seismic_r"),
						colormap_min	= (0 if all_positive else -numpy.max(numpy.abs(focal_CCs))),
						colormap_max	= (None if all_positive else numpy.max(numpy.abs(focal_CCs))),
						clabel			= "",
						show_values		= True, # (numpy.abs(focal_CCs)>0.01),
						value_format	= "%.3f",
						show_grid		= "all",
						cluster			= "none",
						plot_width		= 0.5*focal_CCs.shape[1],
						plot_height		= 0.5*focal_CCs.shape[0],
						plot_title		= "Focal %s control coefficients, %s"%(cctype,system.name)+("" if (focals_label=="") else "\n"+focals_label))
	
